
/*
   m__ns1.c
   Neimark-Sacker (neutral saddle) --> Neimark-Sacker (neutral saddle) Curve
   using double-bordering biproduct (AoA - I) method 
   December 17, 1998
*/

#include "common.h"
#include "corefunc.h"

#include "stagen.h"

#include "_util.h"
#include "_linalg.h"

#include "m__ns1.h"
#include "_conti.h"	/* continuer's (i.e. generator's) data */


/* Numbers of visible name classes;
   the order must be the same as in /names section */
#define CL_TIME		1
#define CL_PHASE	2
#define CL_PAR		3
#define CL_TEST         4
#define CL_UTEST	5
#define CL_STDFUN	6
#define CL_MULT		7

#define _eps     staData->eps
#define _dx      staData->dx
#define _branch  staData->branch   

#define PI	4.0*atan(1.0)
#define PID2	2.0*atan(1.0)   

Local(StagenDataPtr) stagenData=NULL;	/* global copy of Starter's parameter passed by content */
Local(m__ns_DataPtr) staData=NULL;	/* global copy of our own data area */
Local(StdLinAlgDataPtr) staFunc=NULL;	/* global copy of pointer standard linear algebra struct */
Local(UtilitiesDataPtr) staUtil=NULL;	/* global copy of pointer to utility functions struct */
Local(Int2Ptr) active=NULL;        	/* index of active parameter(s) */
Local(Int2) nappend;			/* number of appended user functions */ 
Local(Int2) npar,nphase;		/* dims of par and phase spaces */
Local(Vector) x0=NULL;			/* approximate first point */
Local(Vector) x1=NULL;			/* current point */
Local(Vector) v0=NULL;			/* approximate tangent vector at first point */
Local(FloatHiPtr) X=NULL;		/* current set of phases used by Decoder */
Local(FloatHiPtr) P=NULL;		/* current set of params used by Decoder */
Local(FloatHiPtr) M=NULL;		/* current set of multipliers */
Local(VAR) zero=0;
Local(Vector) StdFun=NULL;
Local(int) itmap;
Local(FloatHiPtr) F,DF,DFF,D2,DD2,D3F,DD3F;		/* work space for supdiff */

#define _itmap       staData->itmap   	/* number of superpositions */

#include "_supdiff.c"

/* Working space for defining and test functions  */

Local(Matrix) A0,A,B,C,C1,C2,AA,AT,BB,BP,BT,D,DE,BP1,BT1,AH;
Local(Vector) vn,xp,xp1,V,W,F,F1,F2,Vp,L,VV,VB,V2,W2,Q,Qr,Qi,PP1,PP2,Pr,Pi,aa,bb,cc,dd,rr;
Local(Vector) VB1,VB2,V1,W1,OLD,RL,K1,K2,K3,OK1,OK2,PIVOT;
Local(Vector) Re=NULL,Im=NULL,Mod=NULL,Arg=NULL;
Local(VAR) test[8],dx,ddx,dddx;

#define BifVector stagenData->BifDataOutPtr
#define ActiveUTestFuns     staData->ActiveUserTest
#define nUTest      	    stagenData->udFuncNum

/* Initialize function common for Starter and Decoder */
Local(Int2) Init(StagenDataPtr stagenptr) {
Int2 i,m;
  /* Set local pointers to data area and functions array */
  stagenData=stagenptr;
  staData=(m__ns_DataPtr)stagenptr->StaPtr;
  /* Use standard linear algebra */
  staFunc=(StdLinAlgDataPtr)stagenptr->StdLinAlg;
  staUtil=(UtilitiesDataPtr)stagenptr->Utilities;
  /* Get dimensions of Phase and Par spaces */
  npar=*stagenptr->ClassDim[CL_PAR];
  nphase=*stagenptr->ClassDim[CL_PHASE];
  itmap=_itmap;
  if (itmap<=0) {
    staUtil->ErrMessage("Superposition must be positive.");
    return 6;
  }    
  /* Call udf SetInitPoint function, if any */
  if (!stagenData->ContDataGenLen && FuncParActive(staData->SetInitPoint)) {	/* there is UDF to set InitPoint */
    Int2 Index_X,Index_P;
    Index_X=stagenptr->ParIndex(staData,X);
    Index_P=stagenptr->ParIndex(staData,P);
    FuncParCall(staData->SetInitPoint)(staData);
    stagenptr->UpdatePar(Index_X);
    stagenptr->UpdatePar(Index_P);
  }  /* Do all the checks related to active parameters and find them */
  for (i=0,nappend=0; i<nUTest; i++) if (ActiveUTestFuns[i]==UDF_APPEND) nappend++;
  active=MemNew(sizeof(Int2)*(nappend+2));    
  if (staUtil->CheckParameters(staData->Active,npar,active,2+nappend)) {
    active=MemFree(active);
    return 1;
  }
  /* Checks related to analytical/numerical differentiation of RHS */
  if (!stagenData->Der1 && staData->dx <= 0) {
    staUtil->ErrMessage("Analytical Jacoby matrix is undefined.\nRecompile system or set positive 'Increment'");
    active=MemFree(active);
    return 2;
  }
  if ((staData->ActiveSing[0] || staData->ActiveSing[1] || staData->ActiveSing[2]) && 
      !stagenData->Der2 && staData->dx <= 0) {
    staUtil->ErrMessage("Analytical Hessian matrix is undefined.\nRecompile system or set positive 'Increment'");
    active=MemFree(active);
    return 3;
  }
  /* Create a copy parameters */
  m=nphase*(nphase-1)/2;
  Der2num=ind2_(nphase+npar-1,nphase+npar-1)+1;
  Der3num=ind3_(nphase+npar-1,nphase+npar-1,nphase+npar-1)+1;
  if (!P) {
    P=MemNew(sizeof(FloatHi)*npar);
    MemCopy(P,staData->P,sizeof(FloatHi)*npar);
    /* Allocate memory for copy of phase vars used by Decoder */
    X=MemNew(sizeof(FloatHi)*nphase);
    /* Allocate memory for multipliers */
    M=MemNew(sizeof(FloatHi)*4*nphase);
    Re=staFunc->CreateVector(nphase);
    Im=staFunc->CreateVector(nphase);
    Mod=staFunc->CreateVector(nphase);
    Arg=staFunc->CreateVector(nphase); 
    StdFun=staFunc->CreateVector(1+2*nphase); /* L2_NORM, MIN&MAX */       
    /* Allocate memory for working space used in defining and test functions */
    A0=staFunc->CreateMatrix(nphase,nphase+2+nappend);
    A=staFunc->CreateMatrix(nphase,nphase);
    B=staFunc->CreateMatrix(nphase,nphase);
    BP=staFunc->CreateMatrix(m+2,m+2);
    BT=staFunc->CreateMatrix(m+2,m+2);
    BP1=staFunc->CreateMatrix(m+1,m+1);
    BT1=staFunc->CreateMatrix(m+1,m+1);
    
    V1=staFunc->CreateVector(m+2);
    W1=staFunc->CreateVector(m+2);    
    V2=staFunc->CreateVector(m+2);
    W2=staFunc->CreateVector(m+2);
    VB=staFunc->CreateVector(m+2);
    VB1=staFunc->CreateVector(m+2);
    VB2=staFunc->CreateVector(m+2);    
    VV=staFunc->CreateVector(2*nphase);
    Q=staFunc->CreateVector(m);
    PP1=staFunc->CreateVector(m+2);
    PP2=staFunc->CreateVector(m+2);
    OLD=staFunc->CreateVector(m+1);
    RL=staFunc->CreateVector(m+1);
    K1=staFunc->CreateVector(m+2);
    K2=staFunc->CreateVector(m+2);
    K3=staFunc->CreateVector(m+2);
    OK1=staFunc->CreateVector(m+2);
    OK2=staFunc->CreateVector(m+2);
    PIVOT=staFunc->CreateVector(m);

    AA=staFunc->CreateMatrix(nphase,nphase);
    BB=staFunc->CreateMatrix(2*nphase,2*nphase);
    C=staFunc->CreateMatrix(nphase+nappend,nphase+2+nappend);
    C1=staFunc->CreateMatrix(nphase,nphase+2+nappend);
    C2=staFunc->CreateMatrix(1,nphase+2+nappend);
    x0=staFunc->CreateVector(nphase+2+nappend);
    x1=staFunc->CreateVector(nphase+2+nappend);
    v0=staFunc->CreateVector(nphase+2+nappend);
    vn=staFunc->CreateVector(nphase);
    xp=staFunc->CreateVector(nphase+2+nappend);
    xp1=staFunc->CreateVector(nphase+2+nappend);
    V=staFunc->CreateVector(nphase);
    W=staFunc->CreateVector(nphase);
    L=staFunc->CreateVector(nphase);
    Vp=staFunc->CreateVector(nphase+2+nappend);
    Qr=staFunc->CreateVector(nphase);
    Qi=staFunc->CreateVector(nphase);
    BifVector=MemNew(sizeof(FloatHi)*(4*nphase+3));
    F=MemNew(sizeof(FloatHi)*nphase);  
    F1=staFunc->CreateVector(nphase);
    F2=staFunc->CreateVector(nphase);
    DF=MemNew(sizeof(FloatHi)*nphase*(nphase+npar));  
    DFF=MemNew(sizeof(FloatHi)*nphase*(nphase+npar));  
    if (staData->ActiveSing[0]+staData->ActiveSing[1]) { /* Degenerate NS or R1:1 detection ON */
      D2=MemNew(sizeof(FloatHi)*nphase*Der2num);
      DD2=MemNew(sizeof(FloatHi)*nphase*Der2num);
      if (staData->ActiveSing[0]) { /* Degenerate NS detection ON */
	D3F=MemNew(sizeof(FloatHi)*nphase*Der3num);
	DD3F=MemNew(sizeof(FloatHi)*nphase*Der3num);
      }
    }
    if (nphase==1) staData->ActiveSing[0]=FALSE;
    if ((nphase>1) && staData->ActiveSing[0]) {  /* Bautin detection ON*/    
      AT=staFunc->CreateMatrix(nphase,nphase);
      Pr=staFunc->CreateVector(nphase);
      Pi=staFunc->CreateVector(nphase);
      aa=staFunc->CreateVector(nphase);
      bb=staFunc->CreateVector(nphase);
      cc=staFunc->CreateVector(nphase);
      dd=staFunc->CreateVector(nphase);
      rr=staFunc->CreateVector(nphase);
    }
    if (nphase<=3) staData->ActiveSing[4]=FALSE;
    if ((nphase>3) && staData->ActiveSing[4]) {  /* Double Neimark-Sacker detection ON */
      AH=staFunc->CreateMatrix(nphase-2,nphase-2);
      D=staFunc->CreateMatrix(nphase+1+nappend,nphase+2+nappend);
      DE=staFunc->CreateMatrix(nphase+2+nappend,nphase+2+nappend);
    }
  }
  dx=_dx;
  ddx=pow(dx,3.0/4.0);
  dddx=pow(dx,3.0/6.0);

  ((ContDataPtr)stagenptr->GenPtr)->ActiveSing=staData->ActiveSing;
  ((ContDataPtr)stagenptr->GenPtr)->ActiveUserTest=staData->ActiveUserTest;
  ((ContDataPtr)stagenptr->GenPtr)->TypeTestFuns=staData->TypeTestFunc;
  ((ContDataPtr)stagenptr->GenPtr)->ZeroTestFuns=staData->ZeroTestFunc;
  return 0;
}

Local(Int2) JacRHS (Vector x, Matrix J);

Local(void) Term (Boolean OK) {
Int2 m;
Int2 i,j,imax,jmax;
VAR max1,max2,d,a11,a12,a21,a22;

  stagenData->BifDataOutLen=0;
  m=nphase*(nphase-1)/2;
  if (OK&&P&&(stagenData->Reason==RF_CLEAN)) {
/* saving global vectors V1,V2,W1 and W2 used in defining functions */
    stagenData->ContDataStaLen=4*(m+2)*sizeof(FloatHi);
    stagenData->ContDataStaPtr=MemNew(stagenData->ContDataStaLen);
    MemCopy(stagenData->ContDataStaPtr,V1,(m+2)*sizeof(FloatHi));
    MemCopy(stagenData->ContDataStaPtr+(m+2),W1,(m+2)*sizeof(FloatHi));
    MemCopy(stagenData->ContDataStaPtr+2*(m+2),V2,(m+2)*sizeof(FloatHi));
    MemCopy(stagenData->ContDataStaPtr+3*(m+2),W2,(m+2)*sizeof(FloatHi));    

/* reevaluate Jacobian */
    JacRHS (x1,C1);
    staFunc->CopyMatrixBlock(C1,0,0,A,0,0,nphase,nphase);
    
/* find pair in (g11,g12), (g21,g22) with max absolute value */
    max1=(fabs(VB1[m])>fabs(VB2[m]))?VB1[m]:VB2[m];
    max2=(fabs(VB1[m+1])>fabs(VB2[m+1]))?VB1[m+1]:VB2[m+1];
    if (fabs(max1)>fabs(max2))
     imax=m;
    else imax=m+1;    
    for(i=0;i<m;i++) {
      Q[i]=VB2[imax]*VB1[i]-VB1[imax]*VB2[i];
    }
       
/* find a component of Q with max(fabs) */
    imax=1;
    jmax=0;
    d=fabs(Q[0]);
    for (i=0; i<nphase; i++) {
      for (j=0; j<i; j++) {
        if (fabs(Q[i*(i-1)/2+j])>d) {
          imax=i;
	  jmax=j;
	  d=fabs(Q[i*(i-1)/2+j]);
        }
      }
    }
    d=Q[imax*(imax-1)/2+jmax];

/* compute eigenvectors */
    staFunc->ClearVector(V);
    staFunc->ClearVector(W);
    for (i=0; i<nphase; i++) {
      if (i<imax) V[i]=Q[imax*(imax-1)/2+i]/d;
      if (i>imax) V[i]=-Q[i*(i-1)/2+imax]/d;
      if (i<jmax) W[i]=-Q[jmax*(jmax-1)/2+i]/d;
      if (i>jmax) W[i]=Q[i*(i-1)/2+jmax]/d;
    }

/* compute kappa */
    staFunc->MultiplyMatrixVector(A,V,F1);  /* F1=A*V  */
    staFunc->MultiplyMatrixVector(A,W,F2);  /* F2=A*W  */

    a11=staFunc->ScalProd(V,V)*staFunc->ScalProd(W,F2);
    a12=staFunc->ScalProd(W,W)*staFunc->ScalProd(V,F1);
    a21=staFunc->ScalProd(V,W)*staFunc->ScalProd(W,F1);
    a22=staFunc->ScalProd(W,V)*staFunc->ScalProd(V,F2);

    d=staFunc->ScalProd(V,V)*staFunc->ScalProd(W,W)
     -staFunc->ScalProd(V,W)*staFunc->ScalProd(V,W);
  
/* orthogonalize and normalize eigenvectors */
    staFunc->NormalizeVector(V);
    staFunc->AddScaledVector(W,V,-staFunc->ScalProd(V,W),L);
    staFunc->NormalizeVector(L);
  
    stagenData->BifDataOutLen=sizeof(FloatHi)*(2*nphase+2);
    BifVector[0]=itmap;
    MemCopy(BifVector+1,V,sizeof(FloatHi)*nphase);
    MemCopy(BifVector+nphase+1,L,sizeof(FloatHi)*nphase);
    BifVector[2*nphase+1]=(a11+a12-a21-a22)/(2.0*d);
    stagenData->ProcessPoint(stagenData->ProcFuncNum, x1, "Last point");
  } else {
    stagenData->ContDataStaLen=0;
    stagenData->BifDataOutLen=0;
  } 
  x0=staFunc->DestroyVector(x0);
  x1=staFunc->DestroyVector(x1);
  v0=staFunc->DestroyVector(v0);
  active=MemFree(active); 
  if (P) {
    P=MemFree(P);
    X=MemFree(X);
    M=MemFree(M);
    Re=staFunc->DestroyVector(Re);
    Im=staFunc->DestroyVector(Im);
    Mod=staFunc->DestroyVector(Mod);
    Arg=staFunc->DestroyVector(Arg); 
    StdFun=staFunc->DestroyVector(StdFun);              
    A=staFunc->DestroyMatrix(A);
    A0=staFunc->DestroyMatrix(A0);
    B=staFunc->DestroyMatrix(B);
    BP=staFunc->DestroyMatrix(BP);
    BT=staFunc->DestroyMatrix(BT);
    BP1=staFunc->DestroyMatrix(BP1);
    BT1=staFunc->DestroyMatrix(BT1);
    V1=staFunc->DestroyVector(V1);
    W1=staFunc->DestroyVector(W1);    
    V2=staFunc->DestroyVector(V2);
    W2=staFunc->DestroyVector(W2);
    Q=staFunc->DestroyVector(Q);
    PP1=staFunc->DestroyVector(PP1);
    PP2=staFunc->DestroyVector(PP2);
    AA=staFunc->DestroyMatrix(AA);
    BB=staFunc->DestroyMatrix(BB);
    VV=staFunc->DestroyVector(VV);
    VB=staFunc->DestroyVector(VB);
    VB1=staFunc->DestroyVector(VB1);
    VB2=staFunc->DestroyVector(VB2);    
    OLD=staFunc->DestroyVector(OLD);
    RL=staFunc->DestroyVector(RL);    
    K1=staFunc->DestroyVector(K1);
    K2=staFunc->DestroyVector(K2);
    K3=staFunc->DestroyVector(K3);
    OK1=staFunc->DestroyVector(OK1);
    OK2=staFunc->DestroyVector(OK2); 
    PIVOT=staFunc->DestroyVector(PIVOT);
    C=staFunc->DestroyMatrix(C);
    C1=staFunc->DestroyMatrix(C1);
    C2=staFunc->DestroyMatrix(C2);
    vn=staFunc->DestroyVector(vn);
    xp=staFunc->DestroyVector(xp);
    xp1=staFunc->DestroyVector(xp1);
    V=staFunc->DestroyVector(V);
    W=staFunc->DestroyVector(W);
    Vp=staFunc->DestroyVector(Vp);
    L=staFunc->DestroyVector(L);
    Qr=staFunc->DestroyVector(Qr);
    Qi=staFunc->DestroyVector(Qi);
    BifVector=MemFree(BifVector);
    F=MemFree(F);
    F1=staFunc->DestroyVector(F1);
    F2=staFunc->DestroyVector(F2);
    DF=MemFree(DF);
    DFF=MemFree(DFF);
    if (staData->ActiveSing[0]+staData->ActiveSing[1]) {
      D2=MemFree(D2);
      DD2=MemFree(DD2);
      if (staData->ActiveSing[0]) {  
 	D3F=MemFree(D3F);
	DD3F=MemFree(DD3F);
      }
    }    
    if ((nphase>1) && staData->ActiveSing[0]) {  /* Bautin detection ON*/
      AT=staFunc->DestroyMatrix(AT);
      Pr=staFunc->DestroyVector(Pr);
      Pi=staFunc->DestroyVector(Pi);
      aa=staFunc->DestroyVector(aa);
      bb=staFunc->DestroyVector(bb);
      cc=staFunc->DestroyVector(cc);
      dd=staFunc->DestroyVector(dd);
      rr=staFunc->DestroyVector(rr);
    }
    if ((nphase>3) && staData->ActiveSing[4]) {  /* Double Neimark-Sacker detection ON*/
      AH=staFunc->DestroyMatrix(AH);
      D=staFunc->DestroyMatrix(D);
      DE=staFunc->DestroyMatrix(DE);
    }
  }
} 

Local(Int2) ns1_JacDefFixedPoint (Vector x, Matrix J);
Entry(Int2) ns1_JacDefNS (Vector x, Matrix J);
Entry(Int2) ns1_DefNS (Vector x, Vector J);

#ifdef WIN32
static int _USERENTRY ns1_SortMult(const void *f, const void *s) {
#else
Entry(int) ns1_SortMult(const void *f, const void *s) {
#endif
VAR Mod1,Mod2;
  Mod1=((FloatHiPtr)f)[0]*((FloatHiPtr)f)[0]+((FloatHiPtr)f)[1]*((FloatHiPtr)f)[1];
  Mod2=((FloatHiPtr)s)[0]*((FloatHiPtr)s)[0]+((FloatHiPtr)s)[1]*((FloatHiPtr)s)[1];
  if (Mod1==Mod2) {
    return 0;
  }
  return (Mod1<Mod2) ? 1 : -1;
}              

/************/
/* Starters */
/* ns -> ns1 Starter */
Entry(Int2) ns_ns1_Starter(StagenDataPtr stagenptr) {
  Int2 i,j,k,kp1,Ret,m,p,q,r,s,imax;
  Vector v1,v2;
  Matrix D,DE,G;
  VAR max1,max2, alpha, beta, t;
  if (!stagenptr) {	/* this is request to terminate */
    Term(TRUE);
    return 0;
  }
  if (Init(stagenptr)) return 1;	/* initialize local variables */
  m=nphase*(nphase-1)/2;

  /* Create the first point for generator */
  if (stagenData->BifDataInOk) _itmap=stagenData->BifDataInPtr[0];
  if (stagenptr->ContDataGenLen) {       /* extension */
    MemCopy(V1,stagenptr->ContDataStaPtr,sizeof(FloatHi)*(m+2));
    MemCopy(W1,stagenptr->ContDataStaPtr+(m+2),sizeof(FloatHi)*(m+2));
    MemCopy(V2,stagenptr->ContDataStaPtr+2*(m+2),sizeof(FloatHi)*(m+2));    
    MemCopy(W2,stagenptr->ContDataStaPtr+3*(m+2),sizeof(FloatHi)*(m+2));    
    MemFree(stagenptr->ContDataStaPtr);
    stagenptr->ContDataStaLen=0;
    MemCopy(x0,staData->X,sizeof(FloatHi)*nphase);
    for (i=0; i<2+nappend; i++) x0[nphase+i]=staData->P[active[i]];    
    
      /* compute Q : needed for extend of Zero Hopf */
      if (JacRHS(x0,C1)) return 1; 	  /* A,C1,BP,V1,V2,W1,W2 - global */
      staFunc->CopyMatrixBlock(C1,0,0,A,0,0,nphase,nphase);
        
      /* Fill in the bordered bi-product matrix BP */
      staFunc->ClearMatrix(BP);
      i=-1;
      for (p=1; p<nphase; p++) {
        for (q=0; q<p; q++) {
         i++;
         j=-1;
         for (r=1; r<nphase; r++) {
           for (s=0; s<r; s++) {
             j++;
             BP[i][j]=A[p][r]*A[q][s]-A[q][r]*A[p][s];
           }
         }
       }
     }
     for (i=0;i<m;i++) {
       BP[i][i]-=1.0;
       BP[m][i]=V1[i];
       BP[i][m]=W1[i];
     }
     for (i=0;i<m+2;i++) {
       BP[m+1][i]=V2[i];
       BP[i][m+1]=W2[i];
     }
     staFunc->CopyMatrix(BP,BT);
     Ret=staFunc->Decomp(BT);
     if (Ret != 0) {
       return 1;
     }
     staFunc->ClearVector(VB1);
     VB1[m]=1.0;
     staFunc->Solve(BT,VB1);
     staFunc->ClearVector(VB2);
     VB2[m+1]=1.0;
     staFunc->Solve(BT,VB2);                         
   /* find pair in (g11,g12), (g21,g22) with max absolute value */
     max1=(fabs(VB1[m])>fabs(VB2[m]))?VB1[m]:VB2[m];
     max2=(fabs(VB1[m+1])>fabs(VB2[m+1]))?VB1[m+1]:VB2[m+1];
     if (fabs(max1)>fabs(max2))
      imax=m;
     else imax=m+1;    
     for(i=0;i<m;i++) {
       Q[i]=VB2[imax]*VB1[i]-VB1[imax]*VB2[i];
     } 
  } else { /* forward or backward */
    v1=staFunc->CreateVector(nphase+2+nappend);
    v2=staFunc->CreateVector(nphase+2+nappend);
    D=staFunc->CreateMatrix(nphase+1+nappend,nphase+2+nappend);
    DE=staFunc->CreateMatrix(nphase+2+nappend,nphase+2+nappend);

    MemCopy(x0,staData->X,sizeof(FloatHi)*nphase);
    for (i=0; i<2+nappend; i++) x0[nphase+i]=staData->P[active[i]];
    JacRHS(x0,C1);

    if (stagenData->BifDataInOk) {
/* Message(MSG_OK,"Eigenvectors exist"); 			*/
/* bifurcation data exist 					*/
/*  BifVector[1] to [nphase]:          NS continuation vector	*/
/*  BifVector[nphase+1] to [2*nphase]: global vector L 		*/

      MemCopy(V,stagenData->BifDataInPtr+1,sizeof(FloatHi)*nphase);
      MemCopy(L,stagenData->BifDataInPtr+nphase+1,sizeof(FloatHi)*nphase);

    } else {
     /* bifurcation data to be computed */
/*  Message(MSG_OK,"Eigenvectors have to be computed"); 	*/
      Ret=0;
{int i,j,imin,jmin;
 double r,s,d,phi,beta,gamma,lambda,lambda1,lambda2,omega,omega2,theta;
    JacRHS(x0,C1);
    staFunc->CopyMatrixBlock(C1,0,0,AA,0,0,nphase,nphase);
    staFunc->EigenValues(AA,Re,Im);
    for (i=0; i<nphase; i++) {
      M[2*i]=Re[i];
      M[2*i+1]=Im[i];
    }
    qsort(M,nphase,2*sizeof(VAR),ns1_SortMult);
    for (i=0; i<nphase; i++) {
      Re[i]=M[2*i];
      Im[i]=M[2*i+1];
    }    

    /* find the pair of multipliers on the unit circle */
    s=fabs(Re[0]*Re[1]-Im[0]*Im[1]-1.0)+fabs(Re[0]*Im[1]+Re[1]*Im[0]);
    imin=0;
    jmin=1;
    for (i=0; i<nphase; i++) {
      for (j=i+1; j<nphase; j++) {
        d=fabs(Re[i]*Re[j]-Im[i]*Im[j]-1.0)+fabs(Re[i]*Im[j]+Re[j]*Im[i]);
        if (d<s) {
          imin=i;
          jmin=j;
          s=d;
        }
      }
    }
    if (fabs(Im[imin])+fabs(Im[jmin])==0) {
/* Neutral saddle */
      lambda=Re[imin];
      lambda1=Re[jmin];
      /* compute real eigenvectors associated with lambda and lambda1=1/lambda */
      staFunc->CopyMatrixBlock(C1,0,0,AA,0,0,nphase,nphase);
      for (i=0; i<nphase; i++) AA[i][i]-=lambda; 
      if (staFunc->SngSolve(AA,_eps,Qr) != nphase-1) {
        staUtil->ErrMessage("Unable to find eigenvector at the initial point");
        Ret=1;
        goto polufinal;
      } 
      staFunc->CopyMatrixBlock(C1,0,0,AA,0,0,nphase,nphase);
      for (i=0; i<nphase; i++) AA[i][i]-=lambda1;
      if (staFunc->SngSolve(AA,_eps,Qi) != nphase-1) {
        staUtil->ErrMessage("Unable to find eigenvector at the initial point");
        Ret=1;
        goto polufinal;
      } 
      staFunc->NormalizeVector(Qr);   
      staFunc->NormalizeVectorRelative(Qi,Qr);
      staFunc->NormalizeVector(Qi);
      
      staFunc->AddScaledVector(Qr,Qi,1.0,V);
      staFunc->NormalizeVector(V);
      staFunc->AddScaledVector(Qr,Qi,-1.0,L);             
      staFunc->NormalizeVector(L); 
    } else {
/* Neimark-Sacker */
      if (fabs(Re[imin]) > 0.5) {
        theta=atan(fabs(Im[imin]/Re[imin]));
      } else {
        theta=PID2-atan(fabs(Re[imin]/Im[imin]));
      }
      if (Re[imin] < 0) theta=PI-theta;
      if (Im[imin] < 0) theta=-theta;
      lambda=cos(theta);
      omega=sin(theta);
      lambda2=cos(2*theta);
      omega2=sin(2*theta);      
      
      /* compute real and imaginary part of the eigenvector */
      staFunc->CopyMatrixBlock(C1,0,0,BB,0,0,nphase,nphase);
      staFunc->CopyMatrixBlock(C1,0,0,BB,nphase,nphase,nphase,nphase);
      for (i=0; i<nphase; i++) {
        BB[i][i+nphase]=omega;
        BB[i][i]-=lambda;
        BB[i+nphase][i]=-omega;
        BB[i+nphase][i+nphase]-=lambda;
      }
      if (staFunc->SngSolve(BB,_eps,VV) != 2*nphase-2) {
        staUtil->ErrMessage("Unable to find eigenvectors at the initial point");
        Ret=1;
        goto polufinal;
      }
      staFunc->CopyVectorElements(VV,0,Qr,0,nphase);
      staFunc->CopyVectorElements(VV,nphase,Qi,0,nphase);

      /* normalize real and imaginary parts */
      d=staFunc->ScalProd(Qr,Qr);
      s=staFunc->ScalProd(Qi,Qi);
      r=staFunc->ScalProd(Qr,Qi);
      if (r!=0) {
        beta=sqrt(fabs(4.0*r*r+(s-d)*(s-d)));
        if (2.0*fabs(2.0*r) > beta) {
          phi=PID2-atan(fabs((s-d)/(2.0*r)));
        } else {
          phi=atan(fabs(2.0*r/(s-d)));
        }
        if (r < 0) phi=PI-phi;
        if ((s-d) < 0) phi=-phi;
        phi/=2.0;
        for (i=0; i<nphase; i++) {
          VV[i]=Qr[i]*cos(phi)-Qi[i]*sin(phi);
          VV[i+nphase]=Qr[i]*sin(phi)+Qi[i]*cos(phi);
        }
      }

      staFunc->CopyVectorElements(VV,0,Qr,0,nphase);
      d=staFunc->ScalProd(Qr,Qr);
      gamma=1/sqrt(d);
      staFunc->MultiplyVectorScalar(VV,gamma,VV);
      staFunc->CopyVectorElements(VV,0,V,0,nphase);
      staFunc->CopyVectorElements(VV,nphase,L,0,nphase);
      staFunc->NormalizeVector(L);
    }
polufinal:
      if (Ret) {
        v1=staFunc->DestroyVector(v1);
        v2=staFunc->DestroyVector(v2);
        D=staFunc->DestroyMatrix(D);
        DE=staFunc->DestroyMatrix(DE);
        Term(FALSE);
        return 1;
      }
    }
}
  if (JacRHS(x0,C1)) return 1;  /* for safety */
  staFunc->CopyMatrixBlock(C1,0,0,A,0,0,nphase,nphase);
  
/* Fill in the bordered bi-product matrix BP */
  staFunc->ClearMatrix(BP);
  i=-1;
  for (p=1; p<nphase; p++) {
    for (q=0; q<p; q++) {
      i++;
      j=-1;
      for (r=1; r<nphase; r++) {
        for (s=0; s<r; s++) {
          j++;
          BP[i][j]=A[p][r]*A[q][s]-A[q][r]*A[p][s];
        }
      }
    }
  }
  for (i=0; i<m; i++) BP[i][i]-=1.0;
/* searching 2 rows with maximal pivots */  
  if (m >= 2) {
    G=staFunc->CreateMatrix(m+1,m+1);
    staFunc->CopyMatrixBlock(BP,0,0,G,0,0,m,m);
    for (i=0; i<m; i++) {
      G[i][m]=i;
      G[m][i]=i;
    } 
    /* Calculation of PIVOTS - complete pivoting */
    for (k=0;k<m-2;k++) {
      t=0.0;
      for (i=k;i<m;i++) {
	for (j=k;j<m;j++) {
    	  if (fabs(G[i][j]) > t) {
            p = i;
            q = j;
            t = fabs(G[i][j]);
          }           
	}
      }      
      if (k!=q) { /* switch colom k with colom  q */        
        for (i=0;i<=m;i++) {
          t=G[i][k];
          G[i][k]=G[i][q];
          G[i][q]=t;
	}
      }
      if (k!=p) { /* switch row k with row p */       
        for (i=0;i<=m;i++) {
          t=G[k][i];         
          G[k][i]=G[p][i];
          G[p][i]=t;
	}
      }        
      if (fabs(G[k][k]) < 0.00001) {
        staUtil->ErrMessage("Biproduct has rank defect at least three");
        Term(FALSE);
        return 1;
      }
      kp1=k+1;
      for (i=kp1;i<m;i++) { /* perform elimination */
        t=-G[i][k]/G[k][k];
        for (j=kp1;j<m;j++) {
          G[i][j]+=t*G[k][j];
	}
      }
    } 
    p=(Int2)G[m-2][m];
    q=(Int2)G[m-1][m];   
    staFunc->ClearVector(W1);
    staFunc->ClearVector(W2);
    W1[p]=1.0;
    W2[q]=1.0;
    BP[p][m]=1.0;
    BP[q][m+1]=1.0;
    p=(Int2)G[m][m-2];
    q=(Int2)G[m][m-1];   
    staFunc->ClearVector(V1);
    staFunc->ClearVector(V2);
    V1[p]=1.0;
    V2[q]=1.0;
    BP[m][p]=1.0;
    BP[m+1][q]=1.0;    
    G=staFunc->DestroyMatrix(G);
      
/* solve first system to obtain new values for V1 */
    staFunc->ClearVector(K1);
    staFunc->ClearVector(K2);
    staFunc->CopyMatrix(BP,BT);
    K1[m]=1.0;
    K2[m+1]=1.0;
    staFunc->Decomp(BT);
    staFunc->Solve(BT,K1);
    staFunc->Solve(BT,K2);
    staFunc->CopyVector(K1,OK1);
    staFunc->CopyVector(K1,K3);
    staFunc->CopyVector(K2,OK2);
    max1=(fabs(OK1[m])>fabs(OK2[m]))?fabs(OK1[m]):fabs(OK2[m]);
    max2=(fabs(OK1[m+1])>fabs(OK2[m+1]))?fabs(OK1[m+1]):fabs(OK2[m+1]);  
    if (max1 > max2) {
     alpha = OK1[m];
     beta = OK2[m];
    }
    else {
     alpha = OK1[m+1];
     beta = OK2[m+1];
    }
    staFunc->MultiplyVectorScalar(OK1,-beta,OK1);
    staFunc->AddScaledVector(OK1,OK2,alpha,K1);
    staFunc->NormalizeVector(K1);
    alpha=staFunc->ScalProd(V1,K1);
    if (alpha < 0) {
      for (i=0;i<m+2;i++)
        K1[i]=-K1[i];
    }
    staFunc->CopyVector(K1,V1);

/* solve second system to obtain new values for W1 */
    staFunc->TransposeMatrix(BP,BT);
    staFunc->ClearVector(K1);
    staFunc->ClearVector(K2);
    K1[m]=1.0;
    K2[m+1]=1.0;
    staFunc->Decomp(BT);
    staFunc->Solve(BT,K1);
    staFunc->Solve(BT,K2);
    staFunc->CopyVector(K1,OK1);
    staFunc->CopyVector(K1,K3);
    staFunc->CopyVector(K2,OK2); 
    max1=(fabs(OK1[m])>fabs(OK2[m]))?fabs(OK1[m]):fabs(OK2[m]);
    max2=(fabs(OK1[m+1])>fabs(OK2[m+1]))?fabs(OK1[m+1]):fabs(OK2[m+1]);  
    if (max1 > max2) {
     alpha = OK1[m];
     beta = OK2[m];
    }
    else {
     alpha = OK1[m+1];
     beta = OK2[m+1];
    }
    staFunc->MultiplyVectorScalar(OK1,-beta,OK1);
    staFunc->AddScaledVector(OK1,OK2,alpha,K1);
    staFunc->NormalizeVector(K1);
    alpha=staFunc->ScalProd(W1,K1);
    if (alpha < 0) {
     for (i=0;i<m+2;i++)
       K1[i]=-K1[i];
    }
    staFunc->CopyVector(K1,W1);
/* adapt V2 */
    staFunc->ClearMatrix(BP1);
    staFunc->CopyMatrixBlock(BP,0,0,BP1,0,0,m,m);
    for (i=0;i<m;i++) {
     BP1[i][m] = W1[i];
     BP1[m][i] = V1[i];
    }    
    staFunc->CopyVectorElements(W2,0,RL,0,m+1);
    staFunc->Decomp(BP1);
    staFunc->Solve(BP1,RL);
    staFunc->NormalizeVector(RL);
    staFunc->CopyVectorElements(V2,0,OLD,0,m+1);
    alpha=staFunc->ScalProd(OLD,RL);
    if (alpha < 0) {
     for (i=0;i<m+1;i++)
       RL[i]=-RL[i];
    }
    staFunc->ClearVector(V2);
    staFunc->CopyVectorElements(RL,0,V2,0,m+1);
/* adapt W2 */
    staFunc->ClearMatrix(BP1);
    staFunc->CopyMatrixBlock(BP,0,0,BP1,0,0,m,m);
    for (i=0;i<m;i++) {
     BP1[i][m] = W1[i];
     BP1[m][i] = V1[i];
    }      
    staFunc->TransposeMatrix(BP1,BT1);
    staFunc->Decomp(BT1);
    staFunc->ClearVector(RL);
    staFunc->CopyVectorElements(V2,0,RL,0,m+1);
/*    staFunc->ClearVector(RL);
      RL[m]=1.0;   */
    staFunc->Solve(BT1,RL);
    staFunc->NormalizeVector(RL);
    staFunc->CopyVectorElements(W2,0,OLD,0,m+1);
    alpha=staFunc->ScalProd(OLD,RL);
    if (alpha < 0) {
     for (i=0;i<m+1;i++)
       RL[i]=-RL[i];
    }  
    staFunc->ClearVector(W2);
    staFunc->CopyVectorElements(RL,0,W2,0,m+1);
  
    for (i=0;i<m;i++) {
      BP[m][i]=V1[i];
      BP[i][m]=W1[i];
    }
    for (i=0;i<m+2;i++) {
      BP[m+1][i]=V2[i];
      BP[i][m+1]=W2[i];
    }
  } else {
     BP[0][m] = 1.0;
     BP[1][0] = 1.0;
     BP[m+1][m+1] = 1.0;
     staFunc->ClearVector(V1);
     staFunc->ClearVector(V2);
     staFunc->ClearVector(W1);
     staFunc->ClearVector(W2);
     V1[0]=1.0;     
     W1[0]=1.0;
     W2[m+1]=1.0;
    } 
  /* Compute a tangent vector to the NS curve */
    ns1_DefNS(x0,v0);
    Ret=ns1_JacDefNS(x0,D);
    for (i=0; i<nphase+2+nappend; i++) {
      v1[i]=1.0;
      staFunc->AppendMatrixVector(D,v1,DE);
      Ret=staFunc->Decomp(DE);
      if (Ret == 0) {
        v2[nphase+1+nappend]=stagenData->Forward ? 1 : -1;
        staFunc->Solve(DE,v2);
        staFunc->CopyVector(v2,v0);
        break;
      };
      v1[i]=0.0;
    };
    if (Ret) {
       staUtil->ErrMessage("Unable to find a tangent vector to the Neimark-Sacker curve");   
       v1=staFunc->DestroyVector(v1);
       v2=staFunc->DestroyVector(v2);
       D=staFunc->DestroyMatrix(D);
       DE=staFunc->DestroyMatrix(DE);
       Term(FALSE);
       return 1;
    }  
    staFunc->NormalizeVector(v0);
    v1=staFunc->DestroyVector(v1);
    v2=staFunc->DestroyVector(v2);
    D=staFunc->DestroyMatrix(D);
    DE=staFunc->DestroyMatrix(DE);
  }
  /* Initialize continuation */
  ((ContDataPtr)stagenptr->GenPtr)->X0=x0;
  ((ContDataPtr)stagenptr->GenPtr)->V0=v0;
  ((ContDataPtr)stagenptr->GenPtr)->ActiveSing=staData->ActiveSing;
  ((ContDataPtr)stagenptr->GenPtr)->ActiveUserTest=staData->ActiveUserTest;
  ((ContDataPtr)stagenptr->GenPtr)->TypeTestFuns=staData->TypeTestFunc;
  ((ContDataPtr)stagenptr->GenPtr)->ZeroTestFuns=staData->ZeroTestFunc;
  /* That's all */
  return 0;
}

/* Resonance 1:1 -> ns Starter */
Entry(Int2) r1_ns1_Starter(StagenDataPtr stagenptr) {
  Int2 i,j,k,kp1,Ret,m,p,q,r,s,imax;
  Vector v1,v2;
  Matrix D,DE,G;
  VAR max1,max2, alpha, beta, t;

  if (!stagenptr) {	/* this is request to terminate */
    Term(TRUE);
    return 0;
  }
  if (Init(stagenptr)) return 1;	/* initialize local variables */
  m=nphase*(nphase-1)/2;

  /* Create the first point for generator */
  if (stagenData->BifDataInOk) _itmap=stagenData->BifDataInPtr[0];
  if (stagenptr->ContDataGenLen) {       /* extension */
    MemCopy(V1,stagenptr->ContDataStaPtr,sizeof(FloatHi)*(m+2));
    MemCopy(W1,stagenptr->ContDataStaPtr+(m+2),sizeof(FloatHi)*(m+2));
    MemCopy(V2,stagenptr->ContDataStaPtr+2*(m+2),sizeof(FloatHi)*(m+2));    
    MemCopy(W2,stagenptr->ContDataStaPtr+3*(m+2),sizeof(FloatHi)*(m+2));    
    MemFree(stagenptr->ContDataStaPtr);
    stagenptr->ContDataStaLen=0;
    MemCopy(x0,staData->X,sizeof(FloatHi)*nphase);
    for (i=0; i<2+nappend; i++) x0[nphase+i]=staData->P[active[i]];    
    
    /* compute Q : needed for extend of Zero Hopf */
      if (JacRHS(x0,C1)) return 1; 	  /* A,C1,BP,V1,V2,W1,W2 - global */
      staFunc->CopyMatrixBlock(C1,0,0,A,0,0,nphase,nphase);
        
      /* Fill in the bordered bi-product matrix BP */
      staFunc->ClearMatrix(BP);
      i=-1;
      for (p=1; p<nphase; p++) {
        for (q=0; q<p; q++) {
         i++;
         j=-1;
         for (r=1; r<nphase; r++) {
           for (s=0; s<r; s++) {
             j++;
             BP[i][j]=A[p][r]*A[q][s]-A[q][r]*A[p][s];
           }
         }
       }
     }
     for (i=0;i<m;i++) {
       BP[i][i]-=1.0;
       BP[m][i]=V1[i];
       BP[i][m]=W1[i];
     }
     for (i=0;i<m+2;i++) {
       BP[m+1][i]=V2[i];
       BP[i][m+1]=W2[i];
     }
     staFunc->CopyMatrix(BP,BT);
     Ret=staFunc->Decomp(BT);
     if (Ret != 0) {
       return 1;
     }
     staFunc->ClearVector(VB1);
     VB1[m]=1.0;
     staFunc->Solve(BT,VB1);
     staFunc->ClearVector(VB2);
     VB2[m+1]=1.0;
     staFunc->Solve(BT,VB2);                         
   /* find pair in (g11,g12), (g21,g22) with max absolute value */
     max1=(fabs(VB1[m])>fabs(VB2[m]))?VB1[m]:VB2[m];
     max2=(fabs(VB1[m+1])>fabs(VB2[m+1]))?VB1[m+1]:VB2[m+1];
     if (fabs(max1)>fabs(max2))
      imax=m;
     else imax=m+1;    
     for(i=0;i<m;i++) {
       Q[i]=VB2[imax]*VB1[i]-VB1[imax]*VB2[i];
     } 
  } else { /* backward or forward */
    v1=staFunc->CreateVector(nphase+2+nappend);
    v2=staFunc->CreateVector(nphase+2+nappend);
    D=staFunc->CreateMatrix(nphase+1+nappend,nphase+2+nappend);
    DE=staFunc->CreateMatrix(nphase+2+nappend,nphase+2+nappend);

    MemCopy(x0,staData->X,sizeof(FloatHi)*nphase);
    for (i=0; i<2+nappend; i++) x0[nphase+i]=staData->P[active[i]];
    JacRHS(x0,C1);

    if (stagenData->BifDataInOk) {
/* Message(MSG_OK,"Eigenvectors exist"); 			         */
/* bifurcation data exist 					         */
/*  BifVector[0]:                         _itmap                         */
/*  BifVector[1] to [nphase]: 		  eigenvector V (A*V=V, <V,V>=1) */

      MemCopy(V,stagenData->BifDataInPtr+1,sizeof(FloatHi)*nphase);
 
    } else {
     /* bifurcation data to be computed */
     /* Message(MSG_OK,"Eigenvector have to be computed"); */
       Ret=0;
       JacRHS(x0,C1);
       staFunc->CopyMatrixBlock(C1,0,0,AA,0,0,nphase,nphase);
       for (i=0; i<nphase; i++) AA[i][i]-=1.0;
       staFunc->CopyMatrixBlock(AA,0,0,BB,0,0,nphase,nphase);
       staFunc->CopyMatrixBlock(AA,0,0,BB,nphase,nphase,nphase,nphase);
       for (i=0; i<nphase; i++) BB[i+nphase][i]=-1;
       i=staFunc->SngSolve(BB,_eps,VV);
       if (i != 2*nphase-2) {
        staUtil->ErrMessage("Unable to find (generalized)eigenvectors at the initial point");
        Ret=1;
        goto final;
       }
       staFunc->CopyVectorElements(VV,0,V,0,nphase);       /* eigenvector */
       staFunc->NormalizeVector(V);
       staFunc->CopyVectorElements(VV,nphase,W,0,nphase);  /* generalized eigenvector */
       staFunc->AddScaledVector(W,V,-staFunc->ScalProd(V,W),W);
       staFunc->NormalizeVector(W);
       staFunc->CopyVectorElements(W,0,L,0,nphase);      
final:
       if (Ret) {
         v1=staFunc->DestroyVector(v1);
         v2=staFunc->DestroyVector(v2);
         D=staFunc->DestroyMatrix(D);
         DE=staFunc->DestroyMatrix(DE);
         Term(FALSE);
         return 1;
       }
    }
  if (JacRHS(x0,C1)) return 1;  /* for safety */
  staFunc->CopyMatrixBlock(C1,0,0,A,0,0,nphase,nphase);
  
/* Fill in the bordered bi-product matrix BP */
  staFunc->ClearMatrix(BP);
  i=-1;
  for (p=1; p<nphase; p++) {
    for (q=0; q<p; q++) {
      i++;
      j=-1;
      for (r=1; r<nphase; r++) {
        for (s=0; s<r; s++) {
          j++;
          BP[i][j]=A[p][r]*A[q][s]-A[q][r]*A[p][s];
        }
      }
    }
  }
  for (i=0; i<m; i++) BP[i][i]-=1.0;
/* searching 2 rows with maximal pivots */  
  if (m >= 2) {
    G=staFunc->CreateMatrix(m+1,m+1);
    staFunc->CopyMatrixBlock(BP,0,0,G,0,0,m,m);
    for (i=0; i<m; i++) {
      G[i][m]=i;
      G[m][i]=i;
    } 
 /* Calculation of PIVOTS - complete pivoting */
    for (k=0;k<m-2;k++) {
      t=0.0;
      for (i=k;i<m;i++) {
	for (j=k;j<m;j++) {
    	  if (fabs(G[i][j]) > t) {
            p = i;
            q = j;
            t = fabs(G[i][j]);
          }           
	}
      }      
      if (k!=q) { /* switch colom k with colom  q */        
        for (i=0;i<=m;i++) {
          t=G[i][k];
          G[i][k]=G[i][q];
          G[i][q]=t;
	}
      }
      if (k!=p) { /* switch row k with row p */       
        for (i=0;i<=m;i++) {
          t=G[k][i];         
          G[k][i]=G[p][i];
          G[p][i]=t;
	}
      }        
      if (fabs(G[k][k]) == 0.0) {
        staUtil->ErrMessage("Biproduct has rank defect at least three");
        Term(FALSE);
        return 1;
      }
      kp1=k+1;
      for (i=kp1;i<m;i++) { /* perform elimination */
        t=-G[i][k]/G[k][k];
        for (j=kp1;j<m;j++) {
          G[i][j]+=t*G[k][j];
	}
      }
    } 
    p=(Int2)G[m-2][m];
    q=(Int2)G[m-1][m];   
    staFunc->ClearVector(W1);
    staFunc->ClearVector(W2);
    W1[p]=1.0;
    W2[q]=1.0;
    BP[p][m]=1.0;
    BP[q][m+1]=1.0;
    p=(Int2)G[m][m-2];
    q=(Int2)G[m][m-1];   
    staFunc->ClearVector(V1);
    staFunc->ClearVector(V2);
    V1[p]=1.0;
    V2[q]=1.0;
    BP[m][p]=1.0;
    BP[m+1][q]=1.0;    
    G=staFunc->DestroyMatrix(G);
      
/* solve first system to obtain new values for V1 */
    staFunc->ClearVector(K1);
    staFunc->ClearVector(K2);
    staFunc->CopyMatrix(BP,BT);  
    K1[m]=1.0;
    K2[m+1]=1.0;
    staFunc->Decomp(BT);
    staFunc->Solve(BT,K1);
    staFunc->Solve(BT,K2);
    staFunc->CopyVector(K1,OK1);
    staFunc->CopyVector(K1,K3);
    staFunc->CopyVector(K2,OK2);
    max1=(fabs(OK1[m])>fabs(OK2[m]))?fabs(OK1[m]):fabs(OK2[m]);
    max2=(fabs(OK1[m+1])>fabs(OK2[m+1]))?fabs(OK1[m+1]):fabs(OK2[m+1]);  
    if (max1 > max2) {
     alpha = OK1[m];
     beta = OK2[m];
    }
    else {
     alpha = OK1[m+1];
     beta = OK2[m+1];
    }
    staFunc->MultiplyVectorScalar(OK1,-beta,OK1);
    staFunc->AddScaledVector(OK1,OK2,alpha,K1);
    staFunc->NormalizeVector(K1);
    alpha=staFunc->ScalProd(V1,K1);
    if (alpha < 0) {
      for (i=0;i<m+2;i++)
        K1[i]=-K1[i];
    }
    staFunc->CopyVector(K1,V1);
  
/* solve second system to obtain new values for W1 */
    staFunc->TransposeMatrix(BP,BT);
    staFunc->ClearVector(K1);
    staFunc->ClearVector(K2);
    K1[m]=1.0;
    K2[m+1]=1.0;
    staFunc->Decomp(BT);
    staFunc->Solve(BT,K1);
    staFunc->Solve(BT,K2);
    staFunc->CopyVector(K1,OK1);
    staFunc->CopyVector(K1,K3);
    staFunc->CopyVector(K2,OK2); 
    max1=(fabs(OK1[m])>fabs(OK2[m]))?fabs(OK1[m]):fabs(OK2[m]);
    max2=(fabs(OK1[m+1])>fabs(OK2[m+1]))?fabs(OK1[m+1]):fabs(OK2[m+1]);  
    if (max1 > max2) {
     alpha = OK1[m];
     beta = OK2[m];
    }
    else {
     alpha = OK1[m+1];
     beta = OK2[m+1];
    }
    staFunc->MultiplyVectorScalar(OK1,-beta,OK1);
    staFunc->AddScaledVector(OK1,OK2,alpha,K1);
    staFunc->NormalizeVector(K1);
    alpha=staFunc->ScalProd(W1,K1);
    if (alpha < 0) {
     for (i=0;i<m+2;i++)
       K1[i]=-K1[i];
    }
    staFunc->CopyVector(K1,W1);
/* adapt V2 */
    staFunc->ClearMatrix(BP1);
    staFunc->CopyMatrixBlock(BP,0,0,BP1,0,0,m,m);
    for (i=0;i<m;i++) {
     BP1[i][m] = W1[i];
     BP1[m][i] = V1[i];
    }    
    staFunc->CopyVectorElements(W2,0,RL,0,m+1);
    staFunc->Decomp(BP1);
    staFunc->Solve(BP1,RL);
    staFunc->NormalizeVector(RL);
    staFunc->CopyVectorElements(V2,0,OLD,0,m+1);
    alpha=staFunc->ScalProd(OLD,RL);
    if (alpha < 0) {
     for (i=0;i<m+1;i++)
       RL[i]=-RL[i];
    }
    staFunc->ClearVector(V2);
    staFunc->CopyVectorElements(RL,0,V2,0,m+1);
/* adapt W2 */
    staFunc->ClearMatrix(BP1);
    staFunc->CopyMatrixBlock(BP,0,0,BP1,0,0,m,m);
    for (i=0;i<m;i++) {
     BP1[i][m] = W1[i];
     BP1[m][i] = V1[i];
    }      
    staFunc->TransposeMatrix(BP1,BT1);
    staFunc->Decomp(BT1);
    staFunc->ClearVector(RL);
    staFunc->CopyVectorElements(V2,0,RL,0,m+1);
    staFunc->Solve(BT1,RL);
    staFunc->NormalizeVector(RL);
    staFunc->CopyVectorElements(W2,0,OLD,0,m+1);
    alpha=staFunc->ScalProd(OLD,RL);
    if (alpha < 0) {
     for (i=0;i<m+1;i++)
       RL[i]=-RL[i];
    }  
    staFunc->ClearVector(W2);
    staFunc->CopyVectorElements(RL,0,W2,0,m+1);
  
    for (i=0;i<m;i++) {
      BP[m][i]=V1[i];
      BP[i][m]=W1[i];
    }
    for (i=0;i<m+2;i++) {
      BP[m+1][i]=V2[i];
      BP[i][m+1]=W2[i];
    }  
  } else {
     BP[0][m] = 1.0;
     BP[1][0] = 1.0;
     BP[m+1][m+1] = 1.0;
     staFunc->ClearVector(V1);
     staFunc->ClearVector(V2);
     staFunc->ClearVector(W1);
     staFunc->ClearVector(W2);
     V1[0]=1.0;     
     W1[0]=1.0;
     W2[m+1]=1.0;
    } 
  /* Compute a tangent vector to the NS curve */
    ns1_DefNS(x0,v0);
    Ret=ns1_JacDefNS(x0,D);
    for (i=0; i<nphase+2+nappend; i++) {
      v1[i]=1.0;
      staFunc->AppendMatrixVector(D,v1,DE);
      Ret=staFunc->Decomp(DE);
      if (Ret == 0) {
        v2[nphase+1+nappend]=stagenData->Forward ? 1 : -1;
        staFunc->Solve(DE,v2);
        staFunc->CopyVector(v2,v0);
        break;
      };
      v1[i]=0.0;
    };
    if (Ret) {
       staUtil->ErrMessage("Unable to find a tangent vector to the Neimark-Sacker curve");   
       v1=staFunc->DestroyVector(v1);
       v2=staFunc->DestroyVector(v2);
       D=staFunc->DestroyMatrix(D);
       DE=staFunc->DestroyMatrix(DE);
       Term(FALSE);
       return 1;
    }  
    staFunc->NormalizeVector(v0);
    v1=staFunc->DestroyVector(v1);
    v2=staFunc->DestroyVector(v2);
    D=staFunc->DestroyMatrix(D);
    DE=staFunc->DestroyMatrix(DE);
  }
  /* Initialize continuation */
  ((ContDataPtr)stagenptr->GenPtr)->X0=x0;
  ((ContDataPtr)stagenptr->GenPtr)->V0=v0;
  ((ContDataPtr)stagenptr->GenPtr)->ActiveSing=staData->ActiveSing;
  ((ContDataPtr)stagenptr->GenPtr)->ActiveUserTest=staData->ActiveUserTest;
  ((ContDataPtr)stagenptr->GenPtr)->TypeTestFuns=staData->TypeTestFunc;
  ((ContDataPtr)stagenptr->GenPtr)->ZeroTestFuns=staData->ZeroTestFunc;
  /* That's all */
  return 0;
}

/* Resonance 1:2 -> ns Starter */
Entry(Int2) r2_ns1_Starter(StagenDataPtr stagenptr) {
  Int2 i,j,k,kp1,Ret,m,p,q,r,s,imax;
  Vector v1,v2;
  Matrix D,DE,G;
  VAR max1,max2, alpha, beta, t;

  if (!stagenptr) {	/* this is request to terminate */
    Term(TRUE);
    return 0;
  }
  if (Init(stagenptr)) return 1;	/* initialize local variables */
  m=nphase*(nphase-1)/2;

  /* Create the first point for generator */
  if (stagenData->BifDataInOk) _itmap=stagenData->BifDataInPtr[0];
  if (stagenptr->ContDataGenLen) {       /* extension */
    MemCopy(V1,stagenptr->ContDataStaPtr,sizeof(FloatHi)*(m+2));
    MemCopy(W1,stagenptr->ContDataStaPtr+(m+2),sizeof(FloatHi)*(m+2));
    MemCopy(V2,stagenptr->ContDataStaPtr+2*(m+2),sizeof(FloatHi)*(m+2));    
    MemCopy(W2,stagenptr->ContDataStaPtr+3*(m+2),sizeof(FloatHi)*(m+2));    
    MemFree(stagenptr->ContDataStaPtr);
    stagenptr->ContDataStaLen=0;
    MemCopy(x0,staData->X,sizeof(FloatHi)*nphase);
    for (i=0; i<2+nappend; i++) x0[nphase+i]=staData->P[active[i]];    
    
    /* compute Q : needed for extend of Zero Hopf */
      if (JacRHS(x0,C1)) return 1; 	  /* A,C1,BP,V1,V2,W1,W2 - global */
      staFunc->CopyMatrixBlock(C1,0,0,A,0,0,nphase,nphase);
        
      /* Fill in the bordered bi-product matrix BP */
      staFunc->ClearMatrix(BP);
      i=-1;
      for (p=1; p<nphase; p++) {
        for (q=0; q<p; q++) {
         i++;
         j=-1;
         for (r=1; r<nphase; r++) {
           for (s=0; s<r; s++) {
             j++;
             BP[i][j]=A[p][r]*A[q][s]-A[q][r]*A[p][s];
           }
         }
       }
     }
     for (i=0;i<m;i++) {
       BP[i][i]-=1.0;
       BP[m][i]=V1[i];
       BP[i][m]=W1[i];
     }
     for (i=0;i<m+2;i++) {
       BP[m+1][i]=V2[i];
       BP[i][m+1]=W2[i];
     }
     staFunc->CopyMatrix(BP,BT);
     Ret=staFunc->Decomp(BT);
     if (Ret != 0) {
       return 1;
     }
     staFunc->ClearVector(VB1);
     VB1[m]=1.0;
     staFunc->Solve(BT,VB1);
     staFunc->ClearVector(VB2);
     VB2[m+1]=1.0;
     staFunc->Solve(BT,VB2);                         
   /* find pair in (g11,g12), (g21,g22) with max absolute value */
     max1=(fabs(VB1[m])>fabs(VB2[m]))?VB1[m]:VB2[m];
     max2=(fabs(VB1[m+1])>fabs(VB2[m+1]))?VB1[m+1]:VB2[m+1];
     if (fabs(max1)>fabs(max2))
      imax=m;
     else imax=m+1;    
     for(i=0;i<m;i++) {
       Q[i]=VB2[imax]*VB1[i]-VB1[imax]*VB2[i];
     } 
  } else { /* backward or forward */
    v1=staFunc->CreateVector(nphase+2+nappend);
    v2=staFunc->CreateVector(nphase+2+nappend);
    D=staFunc->CreateMatrix(nphase+1+nappend,nphase+2+nappend);
    DE=staFunc->CreateMatrix(nphase+2+nappend,nphase+2+nappend);

    MemCopy(x0,staData->X,sizeof(FloatHi)*nphase);
    for (i=0; i<2+nappend; i++) x0[nphase+i]=staData->P[active[i]];
    JacRHS(x0,C1);

    if (stagenData->BifDataInOk) {
/* Message(MSG_OK,"Eigenvectors exist"); 			         */
/* bifurcation data exist 					         */
/*  BifVector[0]:                         _itmap                         */
/*  BifVector[1] to [nphase]: 		  eigenvector V (A*V=-V, <V,V>=1) */
      MemCopy(V,stagenData->BifDataInPtr+1,sizeof(FloatHi)*nphase);
    } else {
     /* bifurcation data to be computed */
     /* Message(MSG_OK,"Eigenvector have to be computed"); */
       Ret=0;
       JacRHS(x0,C1);
       staFunc->CopyMatrixBlock(C1,0,0,AA,0,0,nphase,nphase);
       for (i=0; i<nphase; i++) AA[i][i]+=1.0;
       staFunc->CopyMatrixBlock(AA,0,0,BB,0,0,nphase,nphase);
       staFunc->CopyMatrixBlock(AA,0,0,BB,nphase,nphase,nphase,nphase);
       for (i=0; i<nphase; i++) BB[i+nphase][i]=-1;
       i=staFunc->SngSolve(BB,_eps,VV);
       if (i != 2*nphase-2) {
        staUtil->ErrMessage("Unable to find (generalized)eigenvectors at the initial point");
        Ret=1;
        goto final;
       }
       staFunc->CopyVectorElements(VV,0,V,0,nphase);       /* eigenvector */
       staFunc->NormalizeVector(V);
       staFunc->CopyVectorElements(VV,nphase,W,0,nphase);  /* generalized eigenvector */
       staFunc->AddScaledVector(W,V,-staFunc->ScalProd(V,W),W);
       staFunc->NormalizeVector(W);
       staFunc->CopyVectorElements(W,0,L,0,nphase);      
final:
       if (Ret) {
         v1=staFunc->DestroyVector(v1);
         v2=staFunc->DestroyVector(v2);
         D=staFunc->DestroyMatrix(D);
         DE=staFunc->DestroyMatrix(DE);
         Term(FALSE);
         return 1;
       }
    }
  if (JacRHS(x0,C1)) return 1;  /* for safety */
  staFunc->CopyMatrixBlock(C1,0,0,A,0,0,nphase,nphase);
  
/* Fill in the bordered bi-product matrix BP */
  staFunc->ClearMatrix(BP);
  i=-1;
  for (p=1; p<nphase; p++) {
    for (q=0; q<p; q++) {
      i++;
      j=-1;
      for (r=1; r<nphase; r++) {
        for (s=0; s<r; s++) {
          j++;
          BP[i][j]=A[p][r]*A[q][s]-A[q][r]*A[p][s];
        }
      }
    }
  }
  for (i=0; i<m; i++) BP[i][i]-=1.0;
/* searching 2 rows with maximal pivots */  
  if (m >= 2) {
    G=staFunc->CreateMatrix(m+1,m+1);
    staFunc->CopyMatrixBlock(BP,0,0,G,0,0,m,m);
    for (i=0; i<m; i++) {
      G[i][m]=i;
      G[m][i]=i;
    } 
 /* Calculation of PIVOTS - complete pivoting */
    for (k=0;k<m-2;k++) {
      t=0.0;
      for (i=k;i<m;i++) {
	for (j=k;j<m;j++) {
    	  if (fabs(G[i][j]) > t) {
            p = i;
            q = j;
            t = fabs(G[i][j]);
          }           
	}
      }      
      if (k!=q) { /* switch colom k with colom  q */        
        for (i=0;i<=m;i++) {
          t=G[i][k];
          G[i][k]=G[i][q];
          G[i][q]=t;
	}
      }
      if (k!=p) { /* switch row k with row p */       
        for (i=0;i<=m;i++) {
          t=G[k][i];         
          G[k][i]=G[p][i];
          G[p][i]=t;
	}
      }        
      if (fabs(G[k][k]) == 0.0) {
        staUtil->ErrMessage("Biproduct has rank defect at least three");
        Term(FALSE);
        return 1;
      }
      kp1=k+1;
      for (i=kp1;i<m;i++) { /* perform elimination */
        t=-G[i][k]/G[k][k];
        for (j=kp1;j<m;j++) {
          G[i][j]+=t*G[k][j];
	}
      }
    } 
    p=(Int2)G[m-2][m];
    q=(Int2)G[m-1][m];   
    staFunc->ClearVector(W1);
    staFunc->ClearVector(W2);
    W1[p]=1.0;
    W2[q]=1.0;
    BP[p][m]=1.0;
    BP[q][m+1]=1.0;
    p=(Int2)G[m][m-2];
    q=(Int2)G[m][m-1];   
    staFunc->ClearVector(V1);
    staFunc->ClearVector(V2);
    V1[p]=1.0;
    V2[q]=1.0;
    BP[m][p]=1.0;
    BP[m+1][q]=1.0;    
    G=staFunc->DestroyMatrix(G);
      
/* solve first system to obtain new values for V1 */
    staFunc->ClearVector(K1);
    staFunc->ClearVector(K2);
    staFunc->CopyMatrix(BP,BT);  
    K1[m]=1.0;
    K2[m+1]=1.0;
    staFunc->Decomp(BT);
    staFunc->Solve(BT,K1);
    staFunc->Solve(BT,K2);
    staFunc->CopyVector(K1,OK1);
    staFunc->CopyVector(K1,K3);
    staFunc->CopyVector(K2,OK2);
    max1=(fabs(OK1[m])>fabs(OK2[m]))?fabs(OK1[m]):fabs(OK2[m]);
    max2=(fabs(OK1[m+1])>fabs(OK2[m+1]))?fabs(OK1[m+1]):fabs(OK2[m+1]);  
    if (max1 > max2) {
     alpha = OK1[m];
     beta = OK2[m];
    }
    else {
     alpha = OK1[m+1];
     beta = OK2[m+1];
    }
    staFunc->MultiplyVectorScalar(OK1,-beta,OK1);
    staFunc->AddScaledVector(OK1,OK2,alpha,K1);
    staFunc->NormalizeVector(K1);
    alpha=staFunc->ScalProd(V1,K1);
    if (alpha < 0) {
      for (i=0;i<m+2;i++)
        K1[i]=-K1[i];
    }
    staFunc->CopyVector(K1,V1);
  
/* solve second system to obtain new values for W1 */
    staFunc->TransposeMatrix(BP,BT);
    staFunc->ClearVector(K1);
    staFunc->ClearVector(K2);
    K1[m]=1.0;
    K2[m+1]=1.0;
    staFunc->Decomp(BT);
    staFunc->Solve(BT,K1);
    staFunc->Solve(BT,K2);
    staFunc->CopyVector(K1,OK1);
    staFunc->CopyVector(K1,K3);
    staFunc->CopyVector(K2,OK2); 
    max1=(fabs(OK1[m])>fabs(OK2[m]))?fabs(OK1[m]):fabs(OK2[m]);
    max2=(fabs(OK1[m+1])>fabs(OK2[m+1]))?fabs(OK1[m+1]):fabs(OK2[m+1]);  
    if (max1 > max2) {
     alpha = OK1[m];
     beta = OK2[m];
    }
    else {
     alpha = OK1[m+1];
     beta = OK2[m+1];
    }
    staFunc->MultiplyVectorScalar(OK1,-beta,OK1);
    staFunc->AddScaledVector(OK1,OK2,alpha,K1);
    staFunc->NormalizeVector(K1);
    alpha=staFunc->ScalProd(W1,K1);
    if (alpha < 0) {
     for (i=0;i<m+2;i++)
       K1[i]=-K1[i];
    }
    staFunc->CopyVector(K1,W1);
/* adapt V2 */
    staFunc->ClearMatrix(BP1);
    staFunc->CopyMatrixBlock(BP,0,0,BP1,0,0,m,m);
    for (i=0;i<m;i++) {
     BP1[i][m] = W1[i];
     BP1[m][i] = V1[i];
    }    
    staFunc->CopyVectorElements(W2,0,RL,0,m+1);
    staFunc->Decomp(BP1);
    staFunc->Solve(BP1,RL);
    staFunc->NormalizeVector(RL);
    staFunc->CopyVectorElements(V2,0,OLD,0,m+1);
    alpha=staFunc->ScalProd(OLD,RL);
    if (alpha < 0) {
     for (i=0;i<m+1;i++)
       RL[i]=-RL[i];
    }
    staFunc->ClearVector(V2);
    staFunc->CopyVectorElements(RL,0,V2,0,m+1);
/* adapt W2 */
    staFunc->ClearMatrix(BP1);
    staFunc->CopyMatrixBlock(BP,0,0,BP1,0,0,m,m);
    for (i=0;i<m;i++) {
     BP1[i][m] = W1[i];
     BP1[m][i] = V1[i];
    }      
    staFunc->TransposeMatrix(BP1,BT1);
    staFunc->Decomp(BT1);
    staFunc->ClearVector(RL);
    staFunc->CopyVectorElements(V2,0,RL,0,m+1);
    staFunc->Solve(BT1,RL);
    staFunc->NormalizeVector(RL);
    staFunc->CopyVectorElements(W2,0,OLD,0,m+1);
    alpha=staFunc->ScalProd(OLD,RL);
    if (alpha < 0) {
     for (i=0;i<m+1;i++)
       RL[i]=-RL[i];
    }  
    staFunc->ClearVector(W2);
    staFunc->CopyVectorElements(RL,0,W2,0,m+1);
  
    for (i=0;i<m;i++) {
      BP[m][i]=V1[i];
      BP[i][m]=W1[i];
    }
    for (i=0;i<m+2;i++) {
      BP[m+1][i]=V2[i];
      BP[i][m+1]=W2[i];
    }  
  } else {
     BP[0][m] = 1.0;
     BP[1][0] = 1.0;
     BP[m+1][m+1] = 1.0;
     staFunc->ClearVector(V1);
     staFunc->ClearVector(V2);
     staFunc->ClearVector(W1);
     staFunc->ClearVector(W2);
     V1[0]=1.0;     
     W1[0]=1.0;
     W2[m+1]=1.0;
    } 
  /* Compute a tangent vector to the NS curve */
    ns1_DefNS(x0,v0);
    Ret=ns1_JacDefNS(x0,D);
    for (i=0; i<nphase+2+nappend; i++) {
      v1[i]=1.0;
      staFunc->AppendMatrixVector(D,v1,DE);
      Ret=staFunc->Decomp(DE);
      if (Ret == 0) {
        v2[nphase+1+nappend]=stagenData->Forward ? 1 : -1;
        staFunc->Solve(DE,v2);
        staFunc->CopyVector(v2,v0);
        break;
      };
      v1[i]=0.0;
    };
    if (Ret) {
       staUtil->ErrMessage("Unable to find a tangent vector to the Neimark-Sacker curve");   
       v1=staFunc->DestroyVector(v1);
       v2=staFunc->DestroyVector(v2);
       D=staFunc->DestroyMatrix(D);
       DE=staFunc->DestroyMatrix(DE);
       Term(FALSE);
       return 1;
    }  
    staFunc->NormalizeVector(v0);
    v1=staFunc->DestroyVector(v1);
    v2=staFunc->DestroyVector(v2);
    D=staFunc->DestroyMatrix(D);
    DE=staFunc->DestroyMatrix(DE);
  }
  /* Initialize continuation */
  ((ContDataPtr)stagenptr->GenPtr)->X0=x0;
  ((ContDataPtr)stagenptr->GenPtr)->V0=v0;
  ((ContDataPtr)stagenptr->GenPtr)->ActiveSing=staData->ActiveSing;
  ((ContDataPtr)stagenptr->GenPtr)->ActiveUserTest=staData->ActiveUserTest;
  ((ContDataPtr)stagenptr->GenPtr)->TypeTestFuns=staData->TypeTestFunc;
  ((ContDataPtr)stagenptr->GenPtr)->ZeroTestFuns=staData->ZeroTestFunc;
  /* That's all */
  return 0;
}

/* Fold (flip)-NS -> ns Starter */
Entry(Int2) f_ns1_Starter(StagenDataPtr stagenptr) {
  Int2 i,j,k,kp1,Ret,m,p,q,r,s,imax;
  Vector v1,v2,v00;
  Matrix D,DE,G;
  VAR max1,max2, alpha, beta, t;
  if (!stagenptr) {	/* this is request to terminate */
    Term(TRUE);
    return 0;
  }
  if (Init(stagenptr)) return 1;	/* initialize local variables */
  m=nphase*(nphase-1)/2;

  /* Create the first point for generator */
  if (stagenData->BifDataInOk) _itmap=stagenData->BifDataInPtr[0];
  if (stagenptr->ContDataGenLen) {       /* "Extension" */
    MemCopy(V1,stagenptr->ContDataStaPtr,sizeof(FloatHi)*(m+2));
    MemCopy(W1,stagenptr->ContDataStaPtr+(m+2),sizeof(FloatHi)*(m+2));
    MemCopy(V2,stagenptr->ContDataStaPtr+2*(m+2),sizeof(FloatHi)*(m+2));    
    MemCopy(W2,stagenptr->ContDataStaPtr+3*(m+2),sizeof(FloatHi)*(m+2));    
    MemFree(stagenptr->ContDataStaPtr);
    stagenptr->ContDataStaLen=0;
    MemCopy(x0,staData->X,sizeof(FloatHi)*nphase);
    for (i=0; i<2+nappend; i++) x0[nphase+i]=staData->P[active[i]];    
    
    /* compute Q : needed for extend of Zero Hopf */
      if (JacRHS(x0,C1)) return 1; 	  /* A,C1,BP,V1,V2,W1,W2 - global */
      staFunc->CopyMatrixBlock(C1,0,0,A,0,0,nphase,nphase);
        
      /* Fill in the bordered bi-product matrix BP */
      staFunc->ClearMatrix(BP);
      i=-1;
      for (p=1; p<nphase; p++) {
        for (q=0; q<p; q++) {
         i++;
         j=-1;
         for (r=1; r<nphase; r++) {
           for (s=0; s<r; s++) {
             j++;
             BP[i][j]=A[p][r]*A[q][s]-A[q][r]*A[p][s];
           }
         }
       }
     }
     for (i=0;i<m;i++) {
       BP[i][i]-=1.0;
       BP[m][i]=V1[i];
       BP[i][m]=W1[i];
     }
     for (i=0;i<m+2;i++) {
       BP[m+1][i]=V2[i];
       BP[i][m+1]=W2[i];
     }
     staFunc->CopyMatrix(BP,BT);
     Ret=staFunc->Decomp(BT);
     if (Ret != 0) {
       return 1;
     }
     staFunc->ClearVector(VB1);
     VB1[m]=1.0;
     staFunc->Solve(BT,VB1);
     staFunc->ClearVector(VB2);
     VB2[m+1]=1.0;
     staFunc->Solve(BT,VB2);                         
   /* find pair in (g11,g12), (g21,g22) with max absolute value */
     max1=(fabs(VB1[m])>fabs(VB2[m]))?VB1[m]:VB2[m];
     max2=(fabs(VB1[m+1])>fabs(VB2[m+1]))?VB1[m+1]:VB2[m+1];
     if (fabs(max1)>fabs(max2))
      imax=m;
     else imax=m+1;    
     for(i=0;i<m;i++) {
       Q[i]=VB2[imax]*VB1[i]-VB1[imax]*VB2[i];
     } 
  } else {
    v1=staFunc->CreateVector(nphase+2+nappend);
    v2=staFunc->CreateVector(nphase+2+nappend);
    D=staFunc->CreateMatrix(nphase+1+nappend,nphase+2+nappend);
    DE=staFunc->CreateMatrix(nphase+2+nappend,nphase+2+nappend);

    MemCopy(x0,staData->X,sizeof(FloatHi)*nphase);
    for (i=0; i<2+nappend; i++) x0[nphase+i]=staData->P[active[i]];
    JacRHS(x0,C1);

    if (stagenData->BifDataInOk) {
/* Message(MSG_OK,"Eigenvectors exist"); 			         */
/* bifurcation data exist 					         */
/*  BifVector[nphase+1] to [2*nphase]:          1-NS continuation vector */
/*  BifVector[2*nphase+1] to [3*nphase]:        global vector L 	 */
      MemCopy(V,stagenData->BifDataInPtr+nphase+1,sizeof(FloatHi)*nphase);
      MemCopy(L,stagenData->BifDataInPtr+2*nphase+1,sizeof(FloatHi)*nphase);
      staFunc->NormalizeVector(L);
    } else {
     /* bifurcation data to be computed */
/*  Message(MSG_OK,"Eigenvectors have to be computed"); 	*/
      Ret=0;
{int i,j,imin,jmin;
 double r,s,d,phi,beta,gamma,lambda,lambda1,lambda2,omega,omega2,theta;
    JacRHS(x0,C1);
    staFunc->CopyMatrixBlock(C1,0,0,AA,0,0,nphase,nphase);
    staFunc->EigenValues(AA,Re,Im);
    for (i=0; i<nphase; i++) {
      M[2*i]=Re[i];
      M[2*i+1]=Im[i];
    }
    qsort(M,nphase,2*sizeof(VAR),ns1_SortMult);
    for (i=0; i<nphase; i++) {
      Re[i]=M[2*i];
      Im[i]=M[2*i+1];
    }    

    /* find the pair of multipliers on the unit circle */
    s=fabs(Re[0]*Re[1]-Im[0]*Im[1]-1.0)+fabs(Re[0]*Im[1]+Re[1]*Im[0]);
    imin=0;
    jmin=1;
    for (i=0; i<nphase; i++) {
      for (j=i+1; j<nphase; j++) {
        d=fabs(Re[i]*Re[j]-Im[i]*Im[j]-1.0)+fabs(Re[i]*Im[j]+Re[j]*Im[i]);
        if (d<s) {
          imin=i;
          jmin=j;
          s=d;
        }
      }
    }
    if (fabs(Im[imin])+fabs(Im[jmin])==0) {
/* Neutral saddle */
      lambda=Re[imin];
      lambda1=Re[jmin];
      /* compute real eigenvectors associated with lambda and lambda1=1/lambda */
      staFunc->CopyMatrixBlock(C1,0,0,AA,0,0,nphase,nphase);
      for (i=0; i<nphase; i++) AA[i][i]-=lambda; 
      if (staFunc->SngSolve(AA,_eps,Qr) != nphase-1) {
        staUtil->ErrMessage("Unable to find eigenvector at the initial point");
        Ret=1;
        goto polufinal;
      } 
      staFunc->CopyMatrixBlock(C1,0,0,AA,0,0,nphase,nphase);
      for (i=0; i<nphase; i++) AA[i][i]-=lambda1;
      if (staFunc->SngSolve(AA,_eps,Qi) != nphase-1) {
        staUtil->ErrMessage("Unable to find eigenvector at the initial point");
        Ret=1;
        goto polufinal;
      } 
      staFunc->NormalizeVector(Qr);   
      staFunc->NormalizeVectorRelative(Qi,Qr);
      staFunc->NormalizeVector(Qi);
      
      staFunc->AddScaledVector(Qr,Qi,1.0,V);
      staFunc->NormalizeVector(V);
      staFunc->AddScaledVector(Qr,Qi,-1.0,L);             
      staFunc->NormalizeVector(L); 
    } else {
/* Neimark-Sacker */
      if (fabs(Re[imin]) > 0.5) {
        theta=atan(fabs(Im[imin]/Re[imin]));
      } else {
        theta=PID2-atan(fabs(Re[imin]/Im[imin]));
      }
      if (Re[imin] < 0) theta=PI-theta;
      if (Im[imin] < 0) theta=-theta;
      lambda=cos(theta);
      omega=sin(theta);
      lambda2=cos(2*theta);
      omega2=sin(2*theta);      
      
      /* compute real and imaginary part of the eigenvector */
      staFunc->CopyMatrixBlock(C1,0,0,BB,0,0,nphase,nphase);
      staFunc->CopyMatrixBlock(C1,0,0,BB,nphase,nphase,nphase,nphase);
      for (i=0; i<nphase; i++) {
        BB[i][i+nphase]=omega;
        BB[i][i]-=lambda;
        BB[i+nphase][i]=-omega;
        BB[i+nphase][i+nphase]-=lambda;
      }
      if (staFunc->SngSolve(BB,_eps,VV) != 2*nphase-2) {
        staUtil->ErrMessage("Unable to find eigenvectors at the initial point");
        Ret=1;
        goto polufinal;
      }
      staFunc->CopyVectorElements(VV,0,Qr,0,nphase);
      staFunc->CopyVectorElements(VV,nphase,Qi,0,nphase);

      /* normalize real and imaginary parts */
      d=staFunc->ScalProd(Qr,Qr);
      s=staFunc->ScalProd(Qi,Qi);
      r=staFunc->ScalProd(Qr,Qi);
      if (r!=0) {
        beta=sqrt(fabs(4.0*r*r+(s-d)*(s-d)));
        if (2.0*fabs(2.0*r) > beta) {
          phi=PID2-atan(fabs((s-d)/(2.0*r)));
        } else {
          phi=atan(fabs(2.0*r/(s-d)));
        }
        if (r < 0) phi=PI-phi;
        if ((s-d) < 0) phi=-phi;
        phi/=2.0;
        for (i=0; i<nphase; i++) {
          VV[i]=Qr[i]*cos(phi)-Qi[i]*sin(phi);
          VV[i+nphase]=Qr[i]*sin(phi)+Qi[i]*cos(phi);
        }
      }

      staFunc->CopyVectorElements(VV,0,Qr,0,nphase);
      d=staFunc->ScalProd(Qr,Qr);
      gamma=1/sqrt(d);
      staFunc->MultiplyVectorScalar(VV,gamma,VV);
      staFunc->CopyVectorElements(VV,0,V,0,nphase);
      staFunc->CopyVectorElements(VV,nphase,L,0,nphase);
      staFunc->NormalizeVector(L);
    }
polufinal:
      if (Ret) {
        v1=staFunc->DestroyVector(v1);
        v2=staFunc->DestroyVector(v2);
        D=staFunc->DestroyMatrix(D);
        DE=staFunc->DestroyMatrix(DE);
        Term(FALSE);
        return 1;
      }
    }
}
  if (JacRHS(x0,C1)) return 1;  /* for safety */
  staFunc->CopyMatrixBlock(C1,0,0,A,0,0,nphase,nphase);
  
/* Fill in the bordered bi-product matrix BP */
  staFunc->ClearMatrix(BP);
  i=-1;
  for (p=1; p<nphase; p++) {
    for (q=0; q<p; q++) {
      i++;
      j=-1;
      for (r=1; r<nphase; r++) {
        for (s=0; s<r; s++) {
          j++;
          BP[i][j]=A[p][r]*A[q][s]-A[q][r]*A[p][s];
        }
      }
    }
  }
  for (i=0; i<m; i++) BP[i][i]-=1.0;
/* searching 2 rows with maximal pivots */  
  if (m >= 2) {
    G=staFunc->CreateMatrix(m+1,m+1);
    staFunc->CopyMatrixBlock(BP,0,0,G,0,0,m,m);
    for (i=0; i<m; i++) {
      G[i][m]=i;
      G[m][i]=i;
    } 
 /* Calculation of PIVOTS - complete pivoting */
    for (k=0;k<m-2;k++) {
      t=0.0;
      for (i=k;i<m;i++) {
	for (j=k;j<m;j++) {
    	  if (fabs(G[i][j]) > t) {
            p = i;
            q = j;
            t = fabs(G[i][j]);
          }           
	}
      }      
      if (k!=q) { /* switch colom k with colom  q */        
        for (i=0;i<=m;i++) {
          t=G[i][k];
          G[i][k]=G[i][q];
          G[i][q]=t;
	}
      }
      if (k!=p) { /* switch row k with row p */       
        for (i=0;i<=m;i++) {
          t=G[k][i];         
          G[k][i]=G[p][i];
          G[p][i]=t;
	}
      }        
      if (fabs(G[k][k]) == 0.0) {
        staUtil->ErrMessage("Biproduct has rank defect at least three");
        Term(FALSE);
        return 1;
      }
      kp1=k+1;
      for (i=kp1;i<m;i++) { /* perform elimination */
        t=-G[i][k]/G[k][k];
        for (j=kp1;j<m;j++) {
          G[i][j]+=t*G[k][j];
	}
      }
    } 
    p=(Int2)G[m-2][m];
    q=(Int2)G[m-1][m];   
    staFunc->ClearVector(W1);
    staFunc->ClearVector(W2);
    W1[p]=1.0;
    W2[q]=1.0;
    BP[p][m]=1.0;
    BP[q][m+1]=1.0;
    p=(Int2)G[m][m-2];
    q=(Int2)G[m][m-1];   
    staFunc->ClearVector(V1);
    staFunc->ClearVector(V2);
    V1[p]=1.0;
    V2[q]=1.0;
    BP[m][p]=1.0;
    BP[m+1][q]=1.0;    
    G=staFunc->DestroyMatrix(G);
      
/* solve first system to obtain new values for V1 */
    staFunc->ClearVector(K1);
    staFunc->ClearVector(K2);
    staFunc->CopyMatrix(BP,BT);  
    K1[m]=1.0;
    K2[m+1]=1.0;
    staFunc->Decomp(BT);
    staFunc->Solve(BT,K1);
    staFunc->Solve(BT,K2);
    staFunc->CopyVector(K1,OK1);
    staFunc->CopyVector(K1,K3);
    staFunc->CopyVector(K2,OK2);
    max1=(fabs(OK1[m])>fabs(OK2[m]))?fabs(OK1[m]):fabs(OK2[m]);
    max2=(fabs(OK1[m+1])>fabs(OK2[m+1]))?fabs(OK1[m+1]):fabs(OK2[m+1]);  
    if (max1 > max2) {
     alpha = OK1[m];
     beta = OK2[m];
    }
    else {
     alpha = OK1[m+1];
     beta = OK2[m+1];
    }
    staFunc->MultiplyVectorScalar(OK1,-beta,OK1);
    staFunc->AddScaledVector(OK1,OK2,alpha,K1);
    staFunc->NormalizeVector(K1);
    alpha=staFunc->ScalProd(V1,K1);
    if (alpha < 0) {
      for (i=0;i<m+2;i++)
        K1[i]=-K1[i];
    }
    staFunc->CopyVector(K1,V1);
  
/* solve second system to obtain new values for W1 */
    staFunc->TransposeMatrix(BP,BT);
    staFunc->ClearVector(K1);
    staFunc->ClearVector(K2);
    K1[m]=1.0;
    K2[m+1]=1.0;
    staFunc->Decomp(BT);
    staFunc->Solve(BT,K1);
    staFunc->Solve(BT,K2);
    staFunc->CopyVector(K1,OK1);
    staFunc->CopyVector(K1,K3);
    staFunc->CopyVector(K2,OK2); 
    max1=(fabs(OK1[m])>fabs(OK2[m]))?fabs(OK1[m]):fabs(OK2[m]);
    max2=(fabs(OK1[m+1])>fabs(OK2[m+1]))?fabs(OK1[m+1]):fabs(OK2[m+1]);  
    if (max1 > max2) {
     alpha = OK1[m];
     beta = OK2[m];
    }
    else {
     alpha = OK1[m+1];
     beta = OK2[m+1];
    }
    staFunc->MultiplyVectorScalar(OK1,-beta,OK1);
    staFunc->AddScaledVector(OK1,OK2,alpha,K1);
    staFunc->NormalizeVector(K1);
    alpha=staFunc->ScalProd(W1,K1);
    if (alpha < 0) {
     for (i=0;i<m+2;i++)
       K1[i]=-K1[i];
    }
    staFunc->CopyVector(K1,W1);
/* adapt V2 */
    staFunc->ClearMatrix(BP1);
    staFunc->CopyMatrixBlock(BP,0,0,BP1,0,0,m,m);
    for (i=0;i<m;i++) {
     BP1[i][m] = W1[i];
     BP1[m][i] = V1[i];
    }    
    staFunc->CopyVectorElements(W2,0,RL,0,m+1);
    staFunc->Decomp(BP1);
    staFunc->Solve(BP1,RL);
    staFunc->NormalizeVector(RL);
    staFunc->CopyVectorElements(V2,0,OLD,0,m+1);
    alpha=staFunc->ScalProd(OLD,RL);
    if (alpha < 0) {
     for (i=0;i<m+1;i++)
       RL[i]=-RL[i];
    }
    staFunc->ClearVector(V2);
    staFunc->CopyVectorElements(RL,0,V2,0,m+1);
/* adapt W2 */
    staFunc->ClearMatrix(BP1);
    staFunc->CopyMatrixBlock(BP,0,0,BP1,0,0,m,m);
    for (i=0;i<m;i++) {
     BP1[i][m] = W1[i];
     BP1[m][i] = V1[i];
    }      
    staFunc->TransposeMatrix(BP1,BT1);
    staFunc->Decomp(BT1);
    staFunc->ClearVector(RL);
    staFunc->CopyVectorElements(V2,0,RL,0,m+1);
    staFunc->Solve(BT1,RL);
    staFunc->NormalizeVector(RL);
    staFunc->CopyVectorElements(W2,0,OLD,0,m+1);
    alpha=staFunc->ScalProd(OLD,RL);
    if (alpha < 0) {
     for (i=0;i<m+1;i++)
       RL[i]=-RL[i];
    }  
    staFunc->ClearVector(W2);
    staFunc->CopyVectorElements(RL,0,W2,0,m+1);
  
    for (i=0;i<m;i++) {
      BP[m][i]=V1[i];
      BP[i][m]=W1[i];
    }
    for (i=0;i<m+2;i++) {
      BP[m+1][i]=V2[i];
      BP[i][m+1]=W2[i];
    }  
  } else {
     BP[0][m] = 1.0;
     BP[1][0] = 1.0;
     BP[m+1][m+1] = 1.0;
     staFunc->ClearVector(V1);
     staFunc->ClearVector(V2);
     staFunc->ClearVector(W1);
     staFunc->ClearVector(W2);
     V1[0]=1.0;     
     W1[0]=1.0;
     W2[m+1]=1.0;
    } 
  /* Compute a tangent vector to the NS curve */
    ns1_DefNS(x0,v0);
    Ret=ns1_JacDefNS(x0,D);
    for (i=0; i<nphase+2+nappend; i++) {
      v1[i]=1.0;
      staFunc->AppendMatrixVector(D,v1,DE);
      Ret=staFunc->Decomp(DE);
      if (Ret == 0) {
        v2[nphase+1+nappend]=stagenData->Forward ? 1 : -1;
        staFunc->Solve(DE,v2);
        staFunc->CopyVector(v2,v0);
        break;
      };
      v1[i]=0.0;
    };
    if (Ret) {
       staUtil->ErrMessage("Unable to find a tangent vector to the Neimark-Sacker curve");   
       v1=staFunc->DestroyVector(v1);
       v2=staFunc->DestroyVector(v2);
       D=staFunc->DestroyMatrix(D);
       DE=staFunc->DestroyMatrix(DE);
       Term(FALSE);
       return 1;
    }  
    staFunc->NormalizeVector(v0);

    v1=staFunc->DestroyVector(v1);
    v2=staFunc->DestroyVector(v2);
    D=staFunc->DestroyMatrix(D);
    DE=staFunc->DestroyMatrix(DE);
  }
  /* Initialize continuation */
  ((ContDataPtr)stagenptr->GenPtr)->X0=x0;
  ((ContDataPtr)stagenptr->GenPtr)->V0=v0;
  ((ContDataPtr)stagenptr->GenPtr)->ActiveSing=staData->ActiveSing;
  ((ContDataPtr)stagenptr->GenPtr)->ActiveUserTest=staData->ActiveUserTest;
  ((ContDataPtr)stagenptr->GenPtr)->TypeTestFuns=staData->TypeTestFunc;
  ((ContDataPtr)stagenptr->GenPtr)->ZeroTestFuns=staData->ZeroTestFunc;
  /* That's all */
  return 0;
}

/*****************************************************************/
/* Defining conditions for the eigenvector: (A*A-2*kappa*A+E)V=0 */
Entry(Int2) ns_DefEig (Vector x, Vector y) {

  if (JacRHS(x,C1)) return 1;	    /* A,C2,V,W,F - global */
  staFunc->CopyMatrixBlock(C1,0,0,A,0,0,nphase,nphase);

  staFunc->CopyVectorElements(x, nphase+2+nappend, V, 0, nphase);
  staFunc->MultiplyMatrixVector(A,V,W);
  staFunc->AddScaledVector(V,W,-2.0*x[2*nphase+2+nappend],V);
  staFunc->MultiplyMatrixVector(A,W,F1);

  staFunc->AddScaledVector(V,F1,1.0,V);
  staFunc->CopyVectorElements(V, 0, y, 0, nphase);

  return 0;
}

/****************************************************/
/* Jacobian of defining conditions for eigenvectors */
Local(Int2) ns_JacDefEig (Vector x, Matrix Jac) {
int i,j,k,l;
double s;

  staFunc->ClearMatrix(Jac);
  if (stagenData->Der2) {
    /* creation of DefEig Jacobian matrix Jac using symbolic second derivatives */
    H=SymHess(x);
    if (JacRHS(x,C)) return 1;
    staFunc->CopyMatrixBlock(C,0,0,A,0,0,nphase,nphase);
    for (i=0; i<nphase; i++) {
      for (l=0; l<nphase+2+nappend; l++) {
        for (s=0.0,j=0; j<nphase; j++) {
          for (k=0; k<nphase; k++)
            s+=(HessianElem(i,l,j)*A[j][k]+HessianElem(j,l,k)*A[i][j])*x[k+nphase+2+nappend]; 
            s-=2.0*x[2*nphase+2+nappend]*HessianElem(i,l,j)*x[j+nphase+2+nappend];                  
        }
        Jac[i][l]=s;
      }
    }
    for (i=0; i<nphase; i++) {
      for (l=0; l<nphase; l++) {
        for (s=0.0,j=0; j<nphase; j++) s+=A[i][j]*A[j][l];
        s-=2.0*x[2*nphase+2+nappend]*A[i][l];
        if (i==l) s+=1.0;
        Jac[i][l+nphase+2+nappend]=s;
      }
      for (j=0; j<nphase; j++) Jac[i][2*nphase+2+nappend]-=2.0*A[i][j]*x[j+nphase+2+nappend];
    }
  } else {
    if (staFunc->Der(ns_DefEig, nphase, x, dx, Jac)) return 1;
  }
  return 0;
}

Local(Int2) ns1_JacDefFixedPoint (Vector x, Matrix Jac);

/***********************************************/
/* Jacobian of defining functions for NS curve */  
Entry(Int2) ns_JacDefNS (Vector x, Matrix C3, Matrix Jac) {
  int i;
  staFunc->ClearMatrix(Jac);
  if (ns1_JacDefFixedPoint(x,C)) return 1;
  staFunc->CopyMatrixBlock(C,0,0,Jac,0,0,nphase+nappend,nphase+2+nappend);
  if (ns_JacDefEig(x,C3)) return 2;
  staFunc->CopyMatrixBlock(C3,0,0,Jac,nphase+nappend,0,nphase,2*nphase+3+nappend);
  for (i=0; i<nphase; i++) {
    Jac[2*nphase+nappend][i+nphase+2+nappend]=2.0*x[i+nphase+2+nappend];
    Jac[2*nphase+1+nappend][i+nphase+2+nappend]=L[i];
  }
  return 0;
}

/* ns -> ns1 Starter */
Entry(Int2) nn_ns1_Starter(StagenDataPtr stagenptr) {
  Int2 i,j,k,kp1,Ret,m,p,q,r1,s1,imax,imin,jmin,imin0,jmin0,imin1,jmin1;
  VAR r,s,d,phi,alpha,beta,gamma,lambda,lambda1,lambda2,omega,omega2,theta;
  Vector xx0,v1,v2;
  Matrix C3,D,DE,G;
  VAR max1,max2, t;
  if (!stagenptr) {	/* this is request to terminate */
    Term(TRUE);
    return 0;
  }
  if (Init(stagenptr)) return 1;	/* initialize local variables */
  m=nphase*(nphase-1)/2;
  /* Create the first point for generator */
  if (stagenData->BifDataInOk) _itmap=stagenData->BifDataInPtr[0];
  if (stagenptr->ContDataGenLen) {       /* extension */
    MemCopy(V1,stagenptr->ContDataStaPtr,sizeof(FloatHi)*(m+2));
    MemCopy(W1,stagenptr->ContDataStaPtr+(m+2),sizeof(FloatHi)*(m+2));
    MemCopy(V2,stagenptr->ContDataStaPtr+2*(m+2),sizeof(FloatHi)*(m+2));    
    MemCopy(W2,stagenptr->ContDataStaPtr+3*(m+2),sizeof(FloatHi)*(m+2));    
    MemFree(stagenptr->ContDataStaPtr);
    stagenptr->ContDataStaLen=0;
    MemCopy(x0,staData->X,sizeof(FloatHi)*nphase);
    for (i=0; i<2+nappend; i++) x0[nphase+i]=staData->P[active[i]];    
    
    /* compute Q : needed for extend of Zero Hopf */
    if (JacRHS(x0,C1)) return 1; 	  /* A,C1,BP,V1,V2,W1,W2 - global */
    staFunc->CopyMatrixBlock(C1,0,0,A,0,0,nphase,nphase);
        
    /* Fill in the bordered bi-product matrix BP */
    staFunc->ClearMatrix(BP);
    i=-1;
    for (p=1; p<nphase; p++) {
      for (q=0; q<p; q++) {
       i++;
       j=-1;
       for (r1=1; r1<nphase; r1++) {
         for (s1=0; s1<r1; s1++) {
           j++;
           BP[i][j]=A[p][r1]*A[q][s1]-A[q][r1]*A[p][s1];
         }
       }
     }
   }
   for (i=0;i<m;i++) {
     BP[i][i]-=1.0;
     BP[m][i]=V1[i];
     BP[i][m]=W1[i];
   }
   for (i=0;i<m+2;i++) {
     BP[m+1][i]=V2[i];
     BP[i][m+1]=W2[i];
   }
   staFunc->CopyMatrix(BP,BT);
   Ret=staFunc->Decomp(BT);
   if (Ret != 0) {
     return 1;
   }
   staFunc->ClearVector(VB1);
   VB1[m]=1.0;
   staFunc->Solve(BT,VB1);
   staFunc->ClearVector(VB2);
   VB2[m+1]=1.0;
   staFunc->Solve(BT,VB2);                         
 /* find pair in (g11,g12), (g21,g22) with max absolute value */
   max1=(fabs(VB1[m])>fabs(VB2[m]))?VB1[m]:VB2[m];
   max2=(fabs(VB1[m+1])>fabs(VB2[m+1]))?VB1[m+1]:VB2[m+1];
   if (fabs(max1)>fabs(max2))
    imax=m;
   else imax=m+1;    
   for(i=0;i<m;i++) {
     Q[i]=VB2[imax]*VB1[i]-VB1[imax]*VB2[i];
   } 
  } else { /* forward or backward */
    xx0=staFunc->CreateVector(2*nphase+3+nappend);
    C3=staFunc->CreateMatrix(nphase,2*nphase+3+nappend);

    v1=staFunc->CreateVector(2*nphase+3+nappend);
    v2=staFunc->CreateVector(2*nphase+3+nappend);
    D=staFunc->CreateMatrix(2*nphase+2+nappend,2*nphase+3+nappend);
    DE=staFunc->CreateMatrix(2*nphase+3+nappend,2*nphase+3+nappend);

    MemCopy(xx0,staData->X,sizeof(FloatHi)*nphase);
    for (i=0; i<2+nappend; i++) xx0[nphase+i]=staData->P[active[i]];
    JacRHS(xx0,C1);

    if (stagenData->BifDataInOk) { /* bifurcation data exist */
/* Primary branch: */
/*  BifVector[0]:                          _itmap */
/*  BifVector[1] to [nphase]:              1-NS continuation vector */
/*  BifVector[nphase+1] to [2*nphase]:     1-global vector L */
/*  BifVector[2*nphase+1]:                 1-kappa */
/* Secondary branch: */
/*  BifVector[2*nphase+2] to [3*nphase+1]: 2-NS continuation vector */
/*  BifVector[3*nphase+2] to [4*nphase+1]: 2-global vector L */
/*  BifVector[4*nphase+2]:                 2-kappa */

      if (_branch == 0) { /* primary branch */
        MemCopy(xx0+nphase+2+nappend,stagenData->BifDataInPtr+1,sizeof(FloatHi)*nphase);
        MemCopy(V,stagenData->BifDataInPtr+1,sizeof(FloatHi)*nphase);
        MemCopy(L,stagenData->BifDataInPtr+nphase+1,sizeof(FloatHi)*nphase);
        xx0[2*nphase+2+nappend]=stagenData->BifDataInPtr[2*nphase+1];
      } else { /* secondary branch */
         MemCopy(xx0+nphase+2+nappend,stagenData->BifDataInPtr+2*nphase+2,sizeof(FloatHi)*nphase);
         MemCopy(V,stagenData->BifDataInPtr+2*nphase+2,sizeof(FloatHi)*nphase);
         MemCopy(L,stagenData->BifDataInPtr+3*nphase+2,sizeof(FloatHi)*nphase);
         xx0[2*nphase+2+nappend]=stagenData->BifDataInPtr[4*nphase+2];
      }
    } else { /* bifurcation data to be computed */
      Ret=0;
      JacRHS(xx0,C1);
      staFunc->CopyMatrixBlock(C1,0,0,AA,0,0,nphase,nphase);
      staFunc->EigenValues(AA,Re,Im);
      /* find two pairs of multipliers with unit product */
      s=fabs(Re[0]*Re[1]-Im[0]*Im[1]-1.0)+fabs(Re[0]*Im[1]+Re[1]*Im[0]);
      imin0=0;
      jmin0=1;
      for (i=0; i<nphase; i++) {
        for (j=i+1; j<nphase; j++) {
          d=fabs(Re[i]*Re[j]-Im[i]*Im[j]-1.0)+fabs(Re[i]*Im[j]+Re[j]*Im[i]);
          if (d<s) {
            imin0=i;
            jmin0=j;
            s=d;
          }
        }
      }
      imin1=imin0+1;
      if (imin1==nphase) imin1=0; 
      jmin1=jmin0+1;
      if (jmin1==nphase) jmin1=0; 
      s=fabs(Re[imin1]*Re[jmin1]-Im[imin1]*Im[jmin1]-1.0)+fabs(Re[imin1]*Im[jmin1]+Re[imin1]*Im[jmin1]);
      for (i=0; i<nphase; i++) {
        if (i!=imin0) {
          for (j=i+1; j<nphase; j++) {
            d=fabs(Re[i]*Re[j]-Im[i]*Im[j]-1.0)+fabs(Re[i]*Im[j]+Re[j]*Im[i]);
            if (d<s) {
              imin1=i;
              jmin1=j;
              s=d;
            }
          }
        }
      }
      if (_branch == 0) { /* primary branch */
        imin=imin0;
        jmin=jmin0;
      } else { /* secondary branch */
        imin=imin1;
        jmin=jmin1;
      }

    if (fabs(Im[imin])+fabs(Im[jmin])==0) {
/* Neutral saddle */
      lambda=Re[imin];
      lambda1=Re[jmin];
      /* compute real eigenvectors associated with lambda and lambda1=1/lambda */
      staFunc->CopyMatrixBlock(C1,0,0,AA,0,0,nphase,nphase);
      for (i=0; i<nphase; i++) AA[i][i]-=lambda; 
      if (staFunc->SngSolve(AA,_eps,Qr) != nphase-1) {
        staUtil->ErrMessage("Unable to find eigenvector at the initial point");
        Ret=1;
        goto polufinal;
      } 
      staFunc->CopyMatrixBlock(C1,0,0,AA,0,0,nphase,nphase);
      for (i=0; i<nphase; i++) AA[i][i]-=lambda1;
      if (staFunc->SngSolve(AA,_eps,Qi) != nphase-1) {
        staUtil->ErrMessage("Unable to find eigenvector at the initial point");
        Ret=1;
        goto polufinal;
      } 
      staFunc->NormalizeVector(Qr);   
      staFunc->NormalizeVectorRelative(Qi,Qr);
      staFunc->NormalizeVector(Qi);
      
      staFunc->AddScaledVector(Qr,Qi,1.0,V);
      staFunc->NormalizeVector(V);
      staFunc->AddScaledVector(Qr,Qi,-1.0,L);             
      staFunc->NormalizeVector(L); 
      staFunc->CopyVectorElements(V,0,xx0,nphase+2+nappend,nphase);
      xx0[2*nphase+2+nappend]=(lambda+lambda1)/2.0;      
    } else {
/* Neimark-Sacker */
      if (fabs(Re[imin]) > 0.5) {
        theta=atan(fabs(Im[imin]/Re[imin]));
      } else {
        theta=PID2-atan(fabs(Re[imin]/Im[imin]));
      }
      if (Re[imin] < 0) theta=PI-theta;
      if (Im[imin] < 0) theta=-theta;
      lambda=cos(theta);
      omega=sin(theta);
      lambda2=cos(2*theta);
      omega2=sin(2*theta);      
      
      /* compute real and imaginary part of the eigenvector */
      staFunc->ClearMatrix(BB);
      staFunc->CopyMatrixBlock(C1,0,0,BB,0,0,nphase,nphase);
      staFunc->CopyMatrixBlock(C1,0,0,BB,nphase,nphase,nphase,nphase);
      for (i=0; i<nphase; i++) {
        BB[i][i+nphase]=omega;
        BB[i][i]-=lambda;
        BB[i+nphase][i]=-omega;
        BB[i+nphase][i+nphase]-=lambda;
      }
      if (staFunc->SngSolve(BB,_eps,VV) != 2*nphase-2) {
        staUtil->ErrMessage("Unable to find eigenvectors at the initial point");
        Ret=1;
        goto polufinal;
      }
      staFunc->CopyVectorElements(VV,0,Qr,0,nphase);
      staFunc->CopyVectorElements(VV,nphase,Qi,0,nphase);

      /* normalize real and imaginary parts */
      d=staFunc->ScalProd(Qr,Qr);
      s=staFunc->ScalProd(Qi,Qi);
      r=staFunc->ScalProd(Qr,Qi);
      if (r!=0) {
        beta=sqrt(fabs(4.0*r*r+(s-d)*(s-d)));
        if (2.0*fabs(2.0*r) > beta) {
          phi=PID2-atan(fabs((s-d)/(2.0*r)));
        } else {
          phi=atan(fabs(2.0*r/(s-d)));
        }
        if (r < 0) phi=PI-phi;
        if ((s-d) < 0) phi=-phi;
        phi/=2.0;
        for (i=0; i<nphase; i++) {
          VV[i]=Qr[i]*cos(phi)-Qi[i]*sin(phi);
          VV[i+nphase]=Qr[i]*sin(phi)+Qi[i]*cos(phi);
        }
      }

      staFunc->CopyVectorElements(VV,0,Qr,0,nphase);
      d=staFunc->ScalProd(Qr,Qr);
      gamma=1/sqrt(d);
      staFunc->MultiplyVectorScalar(VV,gamma,VV);
      staFunc->CopyVectorElements(VV,0,V,0,nphase);
      staFunc->CopyVectorElements(VV,0,xx0,nphase+2+nappend,nphase);
      staFunc->CopyVectorElements(VV,nphase,L,0,nphase);
      staFunc->NormalizeVector(L);
      xx0[2*nphase+2+nappend]=cos(theta);
    }
polufinal:
      if (Ret) {
        xx0=staFunc->DestroyVector(xx0);
	C3=staFunc->DestroyMatrix(C3);
        v1=staFunc->DestroyVector(v1);
        v2=staFunc->DestroyVector(v2);
        D=staFunc->DestroyMatrix(D);
        DE=staFunc->DestroyMatrix(DE);
        Term(FALSE);
        return 1;
      }
    }
/* Compute a tangent vector to the NS curve by projecting the standard augmented
   tangent vector (see __ns.c) */
   Ret=ns_JacDefNS(xx0,C3,D);
   for (i=0; i<2*nphase+3+nappend; i++) {
     v1[i]=1.0;
     staFunc->AppendMatrixVector(D,v1,DE);
     Ret=staFunc->Decomp(DE);
     if (Ret == 0) {
       v2[2*nphase+2+nappend]=stagenData->Forward ? 1 : -1;
       staFunc->Solve(DE,v2);
       break;
     };
     v1[i]=0.0;
   };
   if (Ret) {
      staUtil->ErrMessage("Unable to find a tangent vector to the Neimark-Sacker curve");   
      xx0=staFunc->DestroyVector(xx0);
      C3=staFunc->DestroyMatrix(C3);
      v1=staFunc->DestroyVector(v1);
      v2=staFunc->DestroyVector(v2);
      D=staFunc->DestroyMatrix(D);
      DE=staFunc->DestroyMatrix(DE);
      Term(FALSE);
      return 1;
   }  
/* project the point and the tangent vector */
   staFunc->CopyVectorElements(xx0,0,x0,0,nphase+2+nappend);    
   staFunc->CopyVectorElements(v2,0,v0,0,nphase+2+nappend);    
   staFunc->NormalizeVector(v0);

   xx0=staFunc->DestroyVector(xx0);
   C3=staFunc->DestroyMatrix(C3);
   v1=staFunc->DestroyVector(v1);
   v2=staFunc->DestroyVector(v2);
   D=staFunc->DestroyMatrix(D);
   DE=staFunc->DestroyMatrix(DE);

   if (JacRHS(x0,C1)) return 1;  /* for safety */
   staFunc->CopyMatrixBlock(C1,0,0,A,0,0,nphase,nphase);
  
/* Fill in the bordered bi-product matrix BP=AoA-I */
  staFunc->ClearMatrix(BP);
  i=-1;
  for (p=1; p<nphase; p++) {
    for (q=0; q<p; q++) {
      i++;
      j=-1;
      for (r1=1; r1<nphase; r1++) {
        for (s1=0; s1<r1; s1++) {
          j++;
          BP[i][j]=A[p][r1]*A[q][s1]-A[q][r1]*A[p][s1];
        }
      }
    }
  }
  for (i=0; i<m; i++) BP[i][i]-=1.0;
/* searching 2 rows with maximal pivots */  
  if (m >= 2) {
    G=staFunc->CreateMatrix(m+1,m+1);
    staFunc->CopyMatrixBlock(BP,0,0,G,0,0,m,m);
    for (i=0; i<m; i++) {
      G[i][m]=i;
      G[m][i]=i;
    } 
    /* Calculation of PIVOTS - complete pivoting */
    for (k=0;k<m-2;k++) {
      t=0.0;
      for (i=k;i<m;i++) {
	for (j=k;j<m;j++) {
    	  if (fabs(G[i][j]) > t) {
            p = i;
            q = j;
            t = fabs(G[i][j]);
          }           
	}
      }      
      if (k!=q) { /* switch colom k with colom  q */        
        for (i=0;i<=m;i++) {
          t=G[i][k];
          G[i][k]=G[i][q];
          G[i][q]=t;
	}
      }
      if (k!=p) { /* switch row k with row p */       
        for (i=0;i<=m;i++) {
          t=G[k][i];         
          G[k][i]=G[p][i];
          G[p][i]=t;
	}
      }        
      if (fabs(G[k][k]) < _eps) {
        staUtil->ErrMessage("Biproduct has rank defect at least three");
        Term(FALSE);
        return 1;
      }
      kp1=k+1;
      for (i=kp1;i<m;i++) { /* perform elimination */
        t=-G[i][k]/G[k][k];
        for (j=kp1;j<m;j++) {
          G[i][j]+=t*G[k][j];
	}
      }
    } 
    p=(Int2)G[m-2][m];
    q=(Int2)G[m-1][m];   
    staFunc->ClearVector(W1);
    staFunc->ClearVector(W2);
    W1[p]=1.0;
    W2[q]=1.0;
    BP[p][m]=1.0;
    BP[q][m+1]=1.0;
    p=(Int2)G[m][m-2];
    q=(Int2)G[m][m-1];   
    staFunc->ClearVector(V1);
    staFunc->ClearVector(V2);
    V1[p]=1.0;
    V2[q]=1.0;
    BP[m][p]=1.0;
    BP[m+1][q]=1.0;    
    G=staFunc->DestroyMatrix(G);

/* adapt bordering vector V1 */
   staFunc->ClearVector(K1);
   staFunc->ClearVector(K2);
   staFunc->CopyMatrix(BP,BT);
   K1[m]=1.0;
   K2[m+1]=1.0;
   staFunc->Decomp(BT);
   staFunc->Solve(BT,K1);
   staFunc->Solve(BT,K2);
   staFunc->CopyVector(K1,OK1);
   staFunc->CopyVector(K1,K3);
   staFunc->CopyVector(K2,OK2);
   max1=(fabs(OK1[m])>fabs(OK2[m]))?fabs(OK1[m]):fabs(OK2[m]);
   max2=(fabs(OK1[m+1])>fabs(OK2[m+1]))?fabs(OK1[m+1]):fabs(OK2[m+1]);  
   if (max1 > max2) {
    alpha = OK1[m];
    beta = OK2[m];
   }
   else {
    alpha = OK1[m+1];
    beta = OK2[m+1];
   }
   if ((alpha < 1E-6) && (beta < 1E-6)) { /* alpha and beta are both zero */
     staFunc->NormalizeVector(K1);
     staFunc->CopyVector(K1,V1);
   } 
   else {
     staFunc->MultiplyVectorScalar(OK1,-beta,OK1);
     staFunc->AddScaledVector(OK1,OK2,alpha,K1);
     staFunc->NormalizeVector(K1);
     alpha=staFunc->ScalProd(V1,K1);
     if (alpha < 0) {
       for (i=0;i<m+2;i++)
         K1[i]=-K1[i];
     }
     staFunc->CopyVector(K1,V1);
   }

/* adapt bordering vector W1 */
   staFunc->TransposeMatrix(BP,BT);
   staFunc->ClearVector(K1);
   staFunc->ClearVector(K2);
   K1[m]=1.0;
   K2[m+1]=1.0;
   staFunc->Decomp(BT);
   staFunc->Solve(BT,K1);
   staFunc->Solve(BT,K2);
   staFunc->CopyVector(K1,OK1);
   staFunc->CopyVector(K1,K3);
   staFunc->CopyVector(K2,OK2); 
   max1=(fabs(OK1[m])>fabs(OK2[m]))?fabs(OK1[m]):fabs(OK2[m]);
   max2=(fabs(OK1[m+1])>fabs(OK2[m+1]))?fabs(OK1[m+1]):fabs(OK2[m+1]);  
   if (max1 > max2) {
    alpha = OK1[m];
    beta = OK2[m];
   }
   else {
    alpha = OK1[m+1];
    beta = OK2[m+1];
   }
   if ((alpha < 1E-6) && (beta < 1E-6)) { /* alpha and beta are both zero */
     staFunc->NormalizeVector(K1);
     staFunc->CopyVector(K1,W1);
   } 
   else {
     staFunc->MultiplyVectorScalar(OK1,-beta,OK1);
     staFunc->AddScaledVector(OK1,OK2,alpha,K1);
     staFunc->NormalizeVector(K1);
     alpha=staFunc->ScalProd(W1,K1);
     if (alpha < 0) {
      for (i=0;i<m+2;i++)
        K1[i]=-K1[i];
     }
     staFunc->CopyVector(K1,W1);
   }

/* adapt bordering vector V2 */
   staFunc->ClearMatrix(BP1);
   staFunc->CopyMatrixBlock(BP,0,0,BP1,0,0,m,m);
   for (i=0;i<m;i++) {
    BP1[i][m] = W1[i];
    BP1[m][i] = V1[i];
   }    
   staFunc->CopyVectorElements(W2,0,RL,0,m+1);
   Ret=staFunc->Decomp(BP1);
   if (!Ret) {
     staFunc->Solve(BP1,RL);
     staFunc->NormalizeVector(RL);
     staFunc->CopyVectorElements(V2,0,OLD,0,m+1);
     alpha=staFunc->ScalProd(OLD,RL);
     if (alpha < 0) {
      for (i=0;i<m+1;i++)
        RL[i]=-RL[i];
     }
     staFunc->ClearVector(V2);
     staFunc->CopyVectorElements(RL,0,V2,0,m+1);
   }

/* adapt bordering vector W2 */
   staFunc->ClearMatrix(BP1);
   staFunc->CopyMatrixBlock(BP,0,0,BP1,0,0,m,m);
   for (i=0;i<m;i++) {
    BP1[i][m] = W1[i];
    BP1[m][i] = V1[i];
   }      
   staFunc->TransposeMatrix(BP1,BT1);
   Ret=staFunc->Decomp(BT1);
   if (!Ret) {
     staFunc->ClearVector(RL);
     staFunc->CopyVectorElements(V2,0,RL,0,m+1);
     staFunc->Solve(BT1,RL);
     staFunc->NormalizeVector(RL);
     staFunc->CopyVectorElements(W2,0,OLD,0,m+1);
     alpha=staFunc->ScalProd(OLD,RL);
     if (alpha < 0) {
      for (i=0;i<m+1;i++)
        RL[i]=-RL[i];
     }  
     staFunc->ClearVector(W2);
     staFunc->CopyVectorElements(RL,0,W2,0,m+1);
   }
  
   for (i=0;i<m;i++) {
     BP[m][i]=V1[i];
     BP[i][m]=W1[i];
   }
   for (i=0;i<m+2;i++) {
     BP[m+1][i]=V2[i];
     BP[i][m+1]=W2[i];
   }
 } else {
    BP[0][m] = 1.0;
    BP[1][0] = 1.0;
    BP[m+1][m+1] = 1.0;
    staFunc->ClearVector(V1);
    staFunc->ClearVector(V2);
    staFunc->ClearVector(W1);
    staFunc->ClearVector(W2);
    V1[0]=1.0;     
    W1[0]=1.0;
    W2[m+1]=1.0;
   } 
 }
/* Initialize continuation */
  ((ContDataPtr)stagenptr->GenPtr)->X0=x0;
  ((ContDataPtr)stagenptr->GenPtr)->V0=v0;
  ((ContDataPtr)stagenptr->GenPtr)->ActiveSing=staData->ActiveSing;
  ((ContDataPtr)stagenptr->GenPtr)->ActiveUserTest=staData->ActiveUserTest;
  ((ContDataPtr)stagenptr->GenPtr)->TypeTestFuns=staData->TypeTestFunc;
  ((ContDataPtr)stagenptr->GenPtr)->ZeroTestFuns=staData->ZeroTestFunc;
  /* That's all */
  return 0;
}

Local(Int2) ns1_JacDefFixedPoint (Vector x, Matrix Jac);

/************************/
/* Some common routines */
Local(void) CopyToX(Vector vp) {
  MemCopy(X,vp,sizeof(FloatHi)*nphase);  
}

Local(void) CopyToP(Vector vp) {
Int2 i;
  for (i=0; i<2+nappend; i++) P[active[i]]=vp[nphase+i];
}

Local(void) CopyToM(void) {
  MemCopy(M,Mod,sizeof(VAR)*nphase);
  MemCopy(M+nphase,Arg,sizeof(VAR)*nphase);
  MemCopy(M+2*nphase,Re,sizeof(VAR)*nphase);
  MemCopy(M+3*nphase,Im,sizeof(VAR)*nphase);
}

/***********/
/* Decoder */

Entry(Int2) ns1_Decoder(VoidPtr vpoint, FloatHiPtr PNTR indirect, Int2 oper) {
  switch (oper) {
    case DECODER_INIT:	/* point is a pointer to StagenData structure */
    case DECODER_TERM:
     break;
    case DECODER_DIM:	/* returns number of M-points in given G-point */
      return 1;
    default:		/* extract next M-point from G-point pointed by vpoint */
      /* vpoint is a Vector */
      indirect[CL_TIME]=&zero;
      indirect[CL_PHASE]=X;
      CopyToX(vpoint);
      indirect[CL_PAR]=P;
      CopyToP(vpoint);
      indirect[CL_MULT]=M;
      CopyToM();
      indirect[CL_TEST]=test;
      indirect[CL_STDFUN]=StdFun;      
  }
  return 0;
}
/***************************/
/* User-defined functions  */
Entry(Int2) ns1_DefUser (Vector x, Vector y) {
Int2 i,j;
  for (i=j=0; i<nUTest; i++) 
    if (ActiveUTestFuns[i]==UDF_APPEND) {
      (stagenData->udFunc)(x,i,&y[j]);
      j++;
    }
  return 0;
}

/*******/
/* RHS */
Entry(Int2) ns1_DefRHS (Vector x, Vector y) {
Int2 i;
  CopyToP(x);
  staFunc->CopyVectorElements(x,0,y,0,nphase);
  for (i=1; i<=_itmap; i++) {
     CopyToX(y);
     stagenData->Rhs(X,P,&zero,y);
   }
  return 0;
}

/*************************************/
/* Defining function for fixed point */
Local(Int2) ns1_DefFixedPoint (Vector x, Vector y) {
Int2 i;
  ns1_DefRHS(x,y);
  for (i=0; i<nphase; i++) y[i]-=x[i];  
  if (nappend) ns1_DefUser(x,y+nphase);
  return 0;
}

/********************/
/* Jacobian of RHS  */
Local(Int2) JacRHS (Vector x, Matrix Jac) {
  staFunc->CopyVectorElements(x,0,xp,0,nphase+2+nappend);     
  if (stagenData->Der1) {
    SymJac(xp,Jac);
  } else {
    if (staFunc->Der(ns1_DefRHS, nphase, xp, dx, Jac)) {
      return 1; }
  }
  return 0;
}

/****************************************************/
/* Jacobian of defining conditions for fixed points */
Local(Int2) ns1_JacDefFixedPoint (Vector x, Matrix Jac) {
Int2 i;
  staFunc->ClearMatrix(Jac);
  JacRHS(x, Jac);
  for (i=0; i<nphase; i++) Jac[i][i]-=1.0;
  if (nappend) staFunc->Der(ns1_DefUser, nappend, x, dx, Jac+nphase);
  return 0;
}

/***************************/
/* NS defining condition */
Entry(Int2) ns1_DefBip (Vector x, Vector y) {
Int2 i,j,p,q,r,s,m,imax;
Int2 Ret;
VAR max1,max2,max3,max4;

  m=nphase*(nphase-1)/2;
   
  if (JacRHS(x,C1)) return 1;	  /* A,C1,BP,V1,V2,W1,W2 - global */
  staFunc->CopyMatrixBlock(C1,0,0,A,0,0,nphase,nphase);

/* Fill in the bordered bi-product matrix BP */
  staFunc->ClearMatrix(BP);
  i=-1;
  for (p=1; p<nphase; p++) {
    for (q=0; q<p; q++) {
      i++;
      j=-1;
      for (r=1; r<nphase; r++) {
        for (s=0; s<r; s++) {
          j++;
          BP[i][j]=A[p][r]*A[q][s]-A[q][r]*A[p][s];
        }
      }
    }
  }
  
  for (i=0; i<m; i++) BP[i][i]-=1.0;

  for (i=0; i<m; i++) {
    BP[m][i]=V1[i];
    BP[i][m]=W1[i];
  }
  for (i=0; i<m+2; i++) {  
    BP[m+1][i]=V2[i];
    BP[i][m+1]=W2[i];
  }
/* compute NS defining function */
  staFunc->CopyMatrix(BP,BT);
  Ret=staFunc->Decomp(BT);
  if (Ret) return 1;
  staFunc->ClearVector(VB1);
  VB1[m]=1.0;
  staFunc->Solve(BT,VB1);
  staFunc->ClearVector(VB2);
  VB2[m+1]=1.0;
  staFunc->Solve(BT,VB2);
  y[0]=(VB1[m]*VB2[m+1]-VB1[m+1]*VB2[m]);   
 
/* find pair in (g11,g12), (g21,g22) with max absolute value */
  max1=(fabs(VB1[m])>fabs(VB2[m]))?VB1[m]:VB2[m];
  max2=(fabs(VB1[m+1])>fabs(VB2[m+1]))?VB1[m+1]:VB2[m+1];
  if ((fabs(max1) < 1E-6) && (fabs(max2) < 1E-6)) {
    max3=(fabs(VB1[m])>fabs(VB1[m+1]))?VB1[m]:VB1[m+1];
    max4=(fabs(VB2[m])>fabs(VB2[m+1]))?VB2[m]:VB2[m+1];    
    if (fabs(max3) > fabs(max4)) 
      staFunc->CopyVectorElements(VB1,0,Q,0,m);
    else staFunc->CopyVectorElements(VB2,0,Q,0,m);
  }
  else {
    if (fabs(max1)>fabs(max2))
     imax=m;
    else imax=m+1;    
    for(i=0;i<m;i++) {
      Q[i]=VB2[imax]*VB1[i]-VB1[imax]*VB2[i];
    }
  }
  return 0;
}

/*************************************/
/* Jacobian of NS defining condition */
Local(Int2) ns1_JacDefBip (Vector x, Matrix Jac) {
int i,j,l,m,p,r,s,q;
VAR s11,s12,s21,s22;

  m=nphase*(nphase-1)/2;
  staFunc->ClearMatrix(Jac);
  if (stagenData->Der2) {
/* creation of DefBip Jacobian matrix Jac using symbolic second derivatives */
    H=SymHess(x);
/* take global vectors VB1 and VB2 computed at the computation of DefBip */
/* compute PP1 and PP2 using the transposed bordered bi-product */
    staFunc->TransposeMatrix(BP,BT);
    staFunc->ClearVector(PP1);
    PP1[m]=1.0;
    staFunc->Decomp(BT);
    staFunc->Solve(BT,PP1);
    staFunc->ClearVector(PP2);
    PP2[m+1]=1.0;
    staFunc->Solve(BT,PP2);

/* compute gradient of DefBip */
    for (l=0; l<nphase+2+nappend; l++) {
      staFunc->ClearMatrix(BT);
      i=-1;
      for (p=1; p<nphase; p++) {
        for (q=0; q<p; q++) {
          i++;
          j=-1;
          for (r=1; r<nphase; r++) {
            for (s=0; s<r; s++) {
              j++;
              BT[i][j]=HessianElem(p,r,l)*A[q][s]+A[p][r]*HessianElem(q,s,l);
              BT[i][j]-=HessianElem(q,r,l)*A[p][s]+A[q][r]*HessianElem(p,s,l);
            }
          }
        }
      }
      for (i=0,s11=0.0,s12=0.0,s21=0.0,s22=0.0; i<m; i++) {
        for (j=0; j<m; j++) {
	  s11-=PP1[i]*VB1[j]*BT[i][j];
	  s12-=PP1[i]*VB2[j]*BT[i][j];
	  s21-=PP2[i]*VB1[j]*BT[i][j];
	  s22-=PP2[i]*VB2[j]*BT[i][j];
	}
      }
       Jac[0][l]=VB1[m]*s22+VB2[m+1]*s11-VB2[m]*s21-VB1[m+1]*s12; 
    }
  } else {
    if (staFunc->Der(ns1_DefBip, 1, x, dx, Jac)) return 1;
  }
  
  return(0);
}

/***********************************/
/* Defining functions for NS curve */

Entry(Int2) ns1_DefNS (Vector x, Vector y) {
  if (ns1_DefFixedPoint(x,y)) return 1;
  if (ns1_DefBip(x,y+nphase+nappend)) return 2;
  return 0;
}

/***********************************************/
/* Jacobian of defining functions for NS curve */  
Entry(Int2) ns1_JacDefNS (Vector x, Matrix Jac) {
  staFunc->ClearMatrix(Jac);
  if (ns1_JacDefFixedPoint(x,C)) return 1;
  staFunc->CopyMatrixBlock(C,0,0,Jac,0,0,nphase+nappend,nphase+2+nappend);
  if (ns1_JacDefBip(x,C2)) return 2;
  staFunc->CopyMatrixBlock(C2,0,0,Jac,nphase+nappend,0,1,nphase+2+nappend);

  return 0;
}

/*********************/
/* Default processor */
#define SCALE	45.0/atan(1.0)

Entry(Int2) ns1_ProcDefault(Int2 Ind, Vector x, Vector v1, CharPtr *msg) {
VAR s;
Int2 i;

  if (Ind == -1) {  /* zero of user-defined function */
    BifVector[0]=(VAR)_itmap;
    stagenData->BifDataOutLen=sizeof(VAR);     
    return 0;
  }  
  /* Assert: v1==NULL, msg==NULL, Ind==0 */
  ns1_DefNS (x, xp1);  /* compute global Q for test functions */
  MemCopy(x1,x,sizeof(FloatHi)*(nphase+2+nappend));     /* for the last point */

  /* L2_NORM and Min&Max */
  for (i=0,s=0; i<nphase; i++) {
    s+=x[i]*x[i];
    StdFun[1+i]=x[i];
    StdFun[1+nphase+i]=x[i];
  }
  StdFun[0]=sqrt(s);
  /* Multipliers */
  if (staData->eigcomp) {
    if (JacRHS(x, A0)) return (1);     
    staFunc->CopyMatrixBlock(A0,0,0,AA,0,0,nphase,nphase); 
    staFunc->EigenValues(AA,Re,Im);
    for (i=0; i<nphase; i++) {
      M[2*i]=Re[i];
      M[2*i+1]=Im[i];
    }
    qsort(M,nphase,2*sizeof(VAR),ns1_SortMult);
    for (i=0; i<nphase; i++) {
      Re[i]=M[2*i];
      Im[i]=M[2*i+1];
    }
    for (i=0; i<nphase; i++) {
      Mod[i]=sqrt(Re[i]*Re[i]+Im[i]*Im[i]);
      if (Mod[i]==0.0) {
        Arg[i]=0.0;
      } else {
        if (fabs(Re[i]/Mod[i]) > 0.5) {
          Arg[i]=atan(fabs(Im[i]/Re[i]));
        } else {
          if (Re[i]!=0.0) {
            Arg[i]=PID2-atan(fabs(Re[i]/Im[i]));
	  } else {
	    Arg[i]=PID2;
	  }
        }
      }
      if (Re[i] < 0) Arg[i]=PI-Arg[i];
      if (Im[i] < 0) Arg[i]=-Arg[i];
      Arg[i]*=SCALE;
    }
  }
  return 0;
}

/*********************************/
/* Test and processing functions */
typedef EntryPtr(Int2,TestFuncPtr,(Vector xx, Vector vv, FloatHiPtr TV));
typedef EntryPtr(Int2,ProcFuncPtr,(Int2 Ind, Vector xx, Vector vv, CharPtr *msg));

Local(Char) buf[80];

/* first Lyapunov coefficient */
Entry(Int2) ns1_TestDegenNS(Vector x, Vector v, FloatHiPtr res) {
VAR beta,gamma,phi,lambda,lambda2,omega,omega2,d,s,s1,r,r1,h,a11,a12,a21,a22;
Int2 i,j,k,l,Ret,imax,jmax;  

  *res=1.0;
  if (nphase < 2) goto end;

/* reevaluate Jacobian */
   if (JacRHS (x,C1)) return 1;
   staFunc->CopyMatrixBlock(C1,0,0,A,0,0,nphase,nphase);

/* it is assumed that Q is correct */
/* find a component of Q with max(fabs) */
  imax=1;
  jmax=0;
  d=fabs(Q[0]);
  for (i=0; i<nphase; i++) {
    for (j=0; j<i; j++) {
      if (fabs(Q[i*(i-1)/2+j])>d) {
        imax=i;
	jmax=j;
	d=fabs(Q[i*(i-1)/2+j]);
      }
    }
  }
  d=Q[imax*(imax-1)/2+jmax];

/* compute eigenvectors */
  staFunc->ClearVector(V);
  staFunc->ClearVector(W);
  for (i=0; i<nphase; i++) {
    if (i<imax) V[i]=Q[imax*(imax-1)/2+i]/d;
    if (i>imax) V[i]=-Q[i*(i-1)/2+imax]/d;
    if (i<jmax) W[i]=-Q[jmax*(jmax-1)/2+i]/d;
    if (i>jmax) W[i]=Q[i*(i-1)/2+jmax]/d;
  }

/* compute kappa */
  staFunc->MultiplyMatrixVector(A,V,F1);  /* F1=A*V  */
  staFunc->MultiplyMatrixVector(A,W,F2);  /* F2=A*W  */  

  a11=staFunc->ScalProd(V,V)*staFunc->ScalProd(W,F2);
  a12=staFunc->ScalProd(W,W)*staFunc->ScalProd(V,F1);
  a21=staFunc->ScalProd(V,W)*staFunc->ScalProd(W,F1);
  a22=staFunc->ScalProd(W,V)*staFunc->ScalProd(V,F2);

  d=staFunc->ScalProd(V,V)*staFunc->ScalProd(W,W)
   -staFunc->ScalProd(V,W)*staFunc->ScalProd(V,W);
  d=(a11+a12-a21-a22)/(2.0*d);

  if (fabs(d) >= 1.0) { 	/* neutral saddle */
    *res=777;
  } else {   				/* Neimark-Sacker */
    lambda=d;                           /* cos(theta) = kappa */
    omega=sqrt(1.0-lambda*lambda);      /* sin(theta) */
    lambda2=2*lambda*lambda-1.0	;	/* cos(2*theta) */
    omega2=2*omega*lambda;		/* sin(2*theta) */

    if (JacRHS(x,C1)) return 1;	        /* C1,A,Qr,Qi - global */  
    
    /* compute real(Qr) and imaginary(Qi) part of the eigenvector */
    staFunc->ClearMatrix(BB);
    staFunc->CopyMatrixBlock(C1,0,0,BB,0,0,nphase,nphase);
    staFunc->CopyMatrixBlock(C1,0,0,BB,nphase,nphase,nphase,nphase);
    for (i=0; i<nphase; i++) {
      BB[i][i+nphase]=omega;
      BB[i][i]-=lambda;
      BB[i+nphase][i]=-omega;
      BB[i+nphase][i+nphase]-=lambda;
    }
    if (staFunc->SngSolve(BB,_eps,VV) != 2*nphase-2) {
      *res=1.0;
      goto end;
    }
    staFunc->CopyVectorElements(VV,0,Qr,0,nphase);
    staFunc->CopyVectorElements(VV,nphase,Qi,0,nphase);
   
    staFunc->CopyMatrixBlock(C1,0,0,A,0,0,nphase,nphase);
    for (i=0; i<nphase; i++) A[i][i]-=lambda;
 
    staFunc->MultiplyMatrixVector(A,Qr,Qi);
    staFunc->MultiplyVectorScalar(Qi,-1.0/omega,Qi);
    staFunc->CopyVectorElements(Qr,0,VV,0,nphase);
    staFunc->CopyVectorElements(Qi,0,VV,nphase,nphase);
    
    /* normalize real and imaginary parts <Qr,Qr>=1, <Qi,Qi>=0 */
    d=staFunc->ScalProd(Qr,Qr);
    s=staFunc->ScalProd(Qi,Qi);
    r=staFunc->ScalProd(Qr,Qi);
    if (r!=0) {
       beta=sqrt(fabs(4.0*r*r+(s-d)*(s-d)));
       if (2.0*fabs(2.0*r) > beta) {
         phi=PID2-atan(fabs((s-d)/(2.0*r)));
       } else {
         phi=atan(fabs(2.0*r/(s-d)));
       }
       if (r < 0) phi=PI-phi;
       if ((s-d) < 0) phi=-phi;
       phi/=2.0;
       for (i=0; i<nphase; i++) {
          VV[i]=Qr[i]*cos(phi)-Qi[i]*sin(phi);
          VV[i+nphase]=Qr[i]*sin(phi)+Qi[i]*cos(phi);
       }
    }

    staFunc->CopyVectorElements(VV,0,Qr,0,nphase);
    d=staFunc->ScalProd(Qr,Qr);
    gamma=1/sqrt(d);
    staFunc->MultiplyVectorScalar(VV,gamma,VV);
    staFunc->CopyVectorElements(VV,0,Qr,0,nphase);
    staFunc->CopyVectorElements(VV,nphase,Qi,0,nphase);   

    /* compute real and imaginary part of the adjoint eigenvector */
    staFunc->CopyMatrixBlock(C1,0,0,AA,0,0,nphase,nphase);
    staFunc->TransposeMatrix(AA,AT);
    staFunc->ClearMatrix(BB);
    staFunc->CopyMatrixBlock(AT,0,0,BB,0,0,nphase,nphase);
    staFunc->CopyMatrixBlock(AT,0,0,BB,nphase,nphase,nphase,nphase);
    for (i=0; i<nphase; i++) {
         BB[i][i+nphase]=-omega;
         BB[i][i]-=lambda;
         BB[i+nphase][i]=omega;
         BB[i+nphase][i+nphase]-=lambda;
    }
    if (staFunc->SngSolve(BB,_eps,VV) != 2*nphase-2) {
       *res=1.0;
       test[0]=*res;
       return 1;
    }
    staFunc->CopyVectorElements(VV,0,Pr,0,nphase);
    staFunc->CopyVectorElements(VV,nphase,Pi,0,nphase);

    /* normalize the adjoint vector <Pr,Qr>+<Pi,Qi>=1, <Pr,Qi>-<Pi,Qr>=0 */
    d=staFunc->ScalProd(Pr,Qr);
    s=staFunc->ScalProd(Pi,Qi);
    r=staFunc->ScalProd(Pr,Qi);
    h=staFunc->ScalProd(Pi,Qr);
/*    phi=atan((r-h)/(d+s)); */
    beta=sqrt(fabs((r-h)*(r-h)+(d+s)*(d+s)));
    if (2.0*fabs(d+s) > beta) {
       phi=atan(fabs((r-h)/(d+s)));
    } else {
       phi=PID2-atan(fabs((d+s)/(r-h)));
    }
    if ((r-h) < 0) phi=PI-phi;
    if ((d+s) < 0) phi=-phi;

    for (i=0; i<nphase; i++) {
      VV[i]=Pr[i]*cos(phi)-Pi[i]*sin(phi);
      VV[i+nphase]=Pr[i]*sin(phi)+Pi[i]*cos(phi);
    }

    staFunc->CopyVectorElements(VV,0,Pr,0,nphase);
    staFunc->CopyVectorElements(VV,nphase,Pi,0,nphase);
    r=staFunc->ScalProd(Pr,Qr);
    h=staFunc->ScalProd(Pi,Qi);
    beta=1.0/(r+h);
    staFunc->MultiplyVectorScalar(VV,beta,VV);
    staFunc->MultiplyVectorScalar(Pr,beta,Pr);
    staFunc->MultiplyVectorScalar(Pi,beta,Pi);

   /* compute the cubic normal form coefficient a(0) */
      /* quadratic terms */
      if (stagenData->Der2)  {
        H=SymHess(x);
	for (i=0; i<nphase; i++) {
           for (s=0.0,r=0.0,h=0.0,j=0; j<nphase; j++) {
              for (k=0; k<nphase; k++) {
                 s+=HessianElem(i,j,k)*Qr[j]*Qr[k];
                 r+=HessianElem(i,j,k)*Qi[j]*Qi[k];
                 h+=HessianElem(i,j,k)*Qr[j]*Qi[k];
	      }
           }
	   aa[i]=s;	
	   bb[i]=r;	
	   cc[i]=-2.0*h;	
	}
      } else {
        staFunc->ClearVector(v0);
        staFunc->CopyVectorElements(Qr,0,v0,0,nphase);
        ns1_DefRHS(x,aa);
        staFunc->AddScaledVector(x,v0,ddx,x1);
        ns1_DefRHS(x1,bb);
        staFunc->AddScaledVector(x,v0,-ddx,x1);
        ns1_DefRHS(x1,cc);
	staFunc->AddScaledVector(bb,aa,-2.0,rr);
	staFunc->AddScaledVector(rr,cc,1.0,aa);
	staFunc->MultiplyVectorScalar(aa,1.0/ddx/ddx,aa);

        staFunc->CopyVectorElements(Qi,0,v0,0,nphase);
        ns1_DefRHS(x,rr);
        staFunc->AddScaledVector(x,v0,ddx,x1);
        ns1_DefRHS(x1,bb);
        staFunc->AddScaledVector(x,v0,-ddx,x1);
        ns1_DefRHS(x1,cc);
	staFunc->AddScaledVector(bb,rr,-2.0,rr);
	staFunc->AddScaledVector(rr,cc,1.0,bb);
	staFunc->MultiplyVectorScalar(bb,1.0/ddx/ddx,bb);

        staFunc->AddScaledVector(Qr,Qi,1.0,rr);
        staFunc->CopyVectorElements(rr,0,v0,0,nphase);
        staFunc->AddScaledVector(x,v0,0.5*ddx,x1);
        ns1_DefRHS(x1,rr);
        staFunc->AddScaledVector(x,v0,-0.5*ddx,x1);
        ns1_DefRHS(x1,cc);
	staFunc->AddScaledVector(rr,cc,1.0,dd);

        staFunc->AddScaledVector(Qr,Qi,-1.0,rr);
        staFunc->CopyVectorElements(rr,0,v0,0,nphase);
        staFunc->AddScaledVector(x,v0,0.5*ddx,x1);
        ns1_DefRHS(x1,rr);
        staFunc->AddScaledVector(x,v0,-0.5*ddx,x1);
        ns1_DefRHS(x1,cc);
	staFunc->AddScaledVector(rr,cc,1.0,rr);
        staFunc->AddScaledVector(dd,rr,-1.0,cc);
	staFunc->MultiplyVectorScalar(cc,-2.0/ddx/ddx,cc);   /* cc=-2*c */
      }

      staFunc->CopyMatrixBlock(C1,0,0,AA,0,0,nphase,nphase);
      for (i=0; i<nphase; i++) AA[i][i]-=1.0;
      staFunc->AddScaledVector(aa,bb,1.0,rr);
      Ret=staFunc->Decomp(AA);
      if (Ret != 0) {
        *res=1.0;
        test[0]=*res;
        return 1;
      }
      staFunc->Solve(AA,rr);
      staFunc->MultiplyVectorScalar(rr,-2.0,rr);		/* rr=-2*r */

      staFunc->AddScaledVector(bb,aa,-1.0,dd);
      staFunc->CopyVectorElements(dd,0,VV,0,nphase);
      staFunc->CopyVectorElements(cc,0,VV,nphase,nphase);
      staFunc->ClearMatrix(BB);
      staFunc->CopyMatrixBlock(C1,0,0,BB,0,0,nphase,nphase);
      staFunc->CopyMatrixBlock(C1,0,0,BB,nphase,nphase,nphase,nphase);
      for (i=0; i<nphase; i++) {
        BB[i][i+nphase]=omega2;
        BB[i][i]-=lambda2;
        BB[i+nphase][i]=-omega2;
        BB[i+nphase][i+nphase]-=lambda2;
      }
      Ret=staFunc->Decomp(BB);
      if (Ret != 0) {
        *res=1.0;
        test[0]=*res;
        return 1;
      }
      staFunc->Solve(BB,VV);
      staFunc->CopyVectorElements(VV,0,aa,0,nphase);          	/* aa=Sr */
      staFunc->CopyVectorElements(VV,nphase,bb,0,nphase);	/* bb=Si */

      if (stagenData->Der2)  {
        for (s=0.0,i=0; i<nphase; i++) {
           for (j=0; j<nphase; j++) {
              for (k=0; k<nphase; k++) 
	         s+=HessianElem(i,j,k)*(
		   Pr[i]*(Qr[j]*rr[k]+Qr[j]*aa[k]+Qi[j]*bb[k])+
		   Pi[i]*(Qi[j]*rr[k]+Qr[j]*bb[k]-Qi[j]*aa[k]));
           }
        }
        for (s1=0.0,i=0; i<nphase; i++) {
           for (j=0; j<nphase; j++) {
              for (k=0; k<nphase; k++) 
	         s1+=HessianElem(i,j,k)*(
		   Pr[i]*(Qi[j]*rr[k]+Qr[j]*bb[k]-Qi[j]*aa[k])-
		   Pi[i]*(Qr[j]*rr[k]+Qr[j]*aa[k]+Qi[j]*bb[k]));
           }
        }
      } else {
        staFunc->ClearVector(v0);

/*  -2*<Pr,B(Qr,r)>  */
        staFunc->AddScaledVector(Qr,rr,1.0,dd);
        staFunc->CopyVectorElements(dd,0,v0,0,nphase);
        staFunc->AddScaledVector(x,v0,0.5*ddx,x1);
        ns1_DefRHS(x1,dd);
	s=staFunc->ScalProd(Pr,dd);
        staFunc->AddScaledVector(x,v0,-0.5*ddx,x1);
        ns1_DefRHS(x1,dd);
	s+=staFunc->ScalProd(Pr,dd);

        staFunc->AddScaledVector(Qr,rr,-1.0,dd);
        staFunc->CopyVectorElements(dd,0,v0,0,nphase);
        staFunc->AddScaledVector(x,v0,0.5*ddx,x1);
        ns1_DefRHS(x1,dd);
	s-=staFunc->ScalProd(Pr,dd);
        staFunc->AddScaledVector(x,v0,-0.5*ddx,x1);
        ns1_DefRHS(x1,dd);
	s-=staFunc->ScalProd(Pr,dd);
/*  -2*<Pr,B(Qi,r)>  */
        staFunc->AddScaledVector(Qi,rr,1.0,dd);
        staFunc->CopyVectorElements(dd,0,v0,0,nphase);
        staFunc->AddScaledVector(x,v0,0.5*ddx,x1);
        ns1_DefRHS(x1,dd);
	s1=staFunc->ScalProd(Pr,dd);
        staFunc->AddScaledVector(x,v0,-0.5*ddx,x1);
        ns1_DefRHS(x1,dd);
	s1+=staFunc->ScalProd(Pr,dd);

        staFunc->AddScaledVector(Qi,rr,-1.0,dd);
        staFunc->CopyVectorElements(dd,0,v0,0,nphase);
        staFunc->AddScaledVector(x,v0,0.5*ddx,x1);
        ns1_DefRHS(x1,dd);
	s1-=staFunc->ScalProd(Pr,dd);
        staFunc->AddScaledVector(x,v0,-0.5*ddx,x1);
        ns1_DefRHS(x1,dd);
	s1-=staFunc->ScalProd(Pr,dd);
/*  -2*<Pi,B(Qi,r)>  */
        staFunc->AddScaledVector(Qi,rr,1.0,dd);
        staFunc->CopyVectorElements(dd,0,v0,0,nphase);
        staFunc->AddScaledVector(x,v0,0.5*ddx,x1);
        ns1_DefRHS(x1,dd);
	s+=staFunc->ScalProd(Pi,dd);
        staFunc->AddScaledVector(x,v0,-0.5*ddx,x1);
        ns1_DefRHS(x1,dd);
	s+=staFunc->ScalProd(Pi,dd);

        staFunc->AddScaledVector(Qi,rr,-1.0,dd);
        staFunc->CopyVectorElements(dd,0,v0,0,nphase);
        staFunc->AddScaledVector(x,v0,0.5*ddx,x1);
        ns1_DefRHS(x1,dd);
	s-=staFunc->ScalProd(Pi,dd);
        staFunc->AddScaledVector(x,v0,-0.5*ddx,x1);
        ns1_DefRHS(x1,dd);
	s-=staFunc->ScalProd(Pi,dd);
/*  2*<Pi,B(Qr,r)>  */
        staFunc->AddScaledVector(Qr,rr,1.0,dd);
        staFunc->CopyVectorElements(dd,0,v0,0,nphase);
        staFunc->AddScaledVector(x,v0,0.5*ddx,x1);
        ns1_DefRHS(x1,dd);
	s1-=staFunc->ScalProd(Pi,dd);
        staFunc->AddScaledVector(x,v0,-0.5*ddx,x1);
        ns1_DefRHS(x1,dd);
	s1-=staFunc->ScalProd(Pi,dd);

        staFunc->AddScaledVector(Qr,rr,-1.0,dd);
        staFunc->CopyVectorElements(dd,0,v0,0,nphase);
        staFunc->AddScaledVector(x,v0,0.5*ddx,x1);
        ns1_DefRHS(x1,dd);
	s1+=staFunc->ScalProd(Pi,dd);
        staFunc->AddScaledVector(x,v0,-0.5*ddx,x1);
        ns1_DefRHS(x1,dd);
	s1+=staFunc->ScalProd(Pi,dd);
/*  <Pr,B(Qr,Sr)>  */
        staFunc->AddScaledVector(Qr,aa,1.0,dd);
        staFunc->CopyVectorElements(dd,0,v0,0,nphase);
        staFunc->AddScaledVector(x,v0,0.5*ddx,x1);
        ns1_DefRHS(x1,dd);
	s+=staFunc->ScalProd(Pr,dd);
        staFunc->AddScaledVector(x,v0,-0.5*ddx,x1);
        ns1_DefRHS(x1,dd);
	s+=staFunc->ScalProd(Pr,dd);

        staFunc->AddScaledVector(Qr,aa,-1.0,dd);
        staFunc->CopyVectorElements(dd,0,v0,0,nphase);
        staFunc->AddScaledVector(x,v0,0.5*ddx,x1);
        ns1_DefRHS(x1,dd);
	s-=staFunc->ScalProd(Pr,dd);
        staFunc->AddScaledVector(x,v0,-0.5*ddx,x1);
        ns1_DefRHS(x1,dd);
	s-=staFunc->ScalProd(Pr,dd);
/*  <Pr,B(Qr,Si)>  */
        staFunc->AddScaledVector(Qr,bb,1.0,dd);
        staFunc->CopyVectorElements(dd,0,v0,0,nphase);
        staFunc->AddScaledVector(x,v0,0.5*ddx,x1);
        ns1_DefRHS(x1,dd);
	s1+=staFunc->ScalProd(Pr,dd);
        staFunc->AddScaledVector(x,v0,-0.5*ddx,x1);
        ns1_DefRHS(x1,dd);
	s1+=staFunc->ScalProd(Pr,dd);

        staFunc->AddScaledVector(Qr,bb,-1.0,dd);
        staFunc->CopyVectorElements(dd,0,v0,0,nphase);
        staFunc->AddScaledVector(x,v0,0.5*ddx,x1);
        ns1_DefRHS(x1,dd);
	s1-=staFunc->ScalProd(Pr,dd);
        staFunc->AddScaledVector(x,v0,-0.5*ddx,x1);
        ns1_DefRHS(x1,dd);
	s1-=staFunc->ScalProd(Pr,dd);
/*  <Pr,B(Qi,Si)>  */
        staFunc->AddScaledVector(Qi,bb,1.0,dd);
        staFunc->CopyVectorElements(dd,0,v0,0,nphase);
        staFunc->AddScaledVector(x,v0,0.5*ddx,x1);
        ns1_DefRHS(x1,dd);
	s+=staFunc->ScalProd(Pr,dd);
        staFunc->AddScaledVector(x,v0,-0.5*ddx,x1);
        ns1_DefRHS(x1,dd);
	s+=staFunc->ScalProd(Pr,dd);

        staFunc->AddScaledVector(Qi,bb,-1.0,dd);
        staFunc->CopyVectorElements(dd,0,v0,0,nphase);
        staFunc->AddScaledVector(x,v0,0.5*ddx,x1);
        ns1_DefRHS(x1,dd);
	s-=staFunc->ScalProd(Pr,dd);
        staFunc->AddScaledVector(x,v0,-0.5*ddx,x1);
        ns1_DefRHS(x1,dd);
	s-=staFunc->ScalProd(Pr,dd);
/* -<Pr,B(Qi,Sr)>  */
        staFunc->AddScaledVector(Qi,aa,1.0,dd);
        staFunc->CopyVectorElements(dd,0,v0,0,nphase);
        staFunc->AddScaledVector(x,v0,0.5*ddx,x1);
        ns1_DefRHS(x1,dd);
	s1-=staFunc->ScalProd(Pr,dd);
        staFunc->AddScaledVector(x,v0,-0.5*ddx,x1);
        ns1_DefRHS(x1,dd);
	s1-=staFunc->ScalProd(Pr,dd);

        staFunc->AddScaledVector(Qi,aa,-1.0,dd);
        staFunc->CopyVectorElements(dd,0,v0,0,nphase);
        staFunc->AddScaledVector(x,v0,0.5*ddx,x1);
        ns1_DefRHS(x1,dd);
	s1+=staFunc->ScalProd(Pr,dd);
        staFunc->AddScaledVector(x,v0,-0.5*ddx,x1);
        ns1_DefRHS(x1,dd);
	s1+=staFunc->ScalProd(Pr,dd);
/*  <Pi,B(Qr,Si)>  */
        staFunc->AddScaledVector(Qr,bb,1.0,dd);
        staFunc->CopyVectorElements(dd,0,v0,0,nphase);
        staFunc->AddScaledVector(x,v0,0.5*ddx,x1);
        ns1_DefRHS(x1,dd);
	s+=staFunc->ScalProd(Pi,dd);
        staFunc->AddScaledVector(x,v0,-0.5*ddx,x1);
        ns1_DefRHS(x1,dd);
	s+=staFunc->ScalProd(Pi,dd);

        staFunc->AddScaledVector(Qr,bb,-1.0,dd);
        staFunc->CopyVectorElements(dd,0,v0,0,nphase);
        staFunc->AddScaledVector(x,v0,0.5*ddx,x1);
        ns1_DefRHS(x1,dd);
	s-=staFunc->ScalProd(Pi,dd);
        staFunc->AddScaledVector(x,v0,-0.5*ddx,x1);
        ns1_DefRHS(x1,dd);
	s-=staFunc->ScalProd(Pi,dd);
/*  -<Pi,B(Qr,Sr)>  */
        staFunc->AddScaledVector(Qr,aa,1.0,dd);
        staFunc->CopyVectorElements(dd,0,v0,0,nphase);
        staFunc->AddScaledVector(x,v0,0.5*ddx,x1);
        ns1_DefRHS(x1,dd);
	s1-=staFunc->ScalProd(Pi,dd);
        staFunc->AddScaledVector(x,v0,-0.5*ddx,x1);
        ns1_DefRHS(x1,dd);
	s1-=staFunc->ScalProd(Pi,dd);

        staFunc->AddScaledVector(Qr,aa,-1.0,dd);
        staFunc->CopyVectorElements(dd,0,v0,0,nphase);
        staFunc->AddScaledVector(x,v0,0.5*ddx,x1);
        ns1_DefRHS(x1,dd);
	s1+=staFunc->ScalProd(Pi,dd);
        staFunc->AddScaledVector(x,v0,-0.5*ddx,x1);
        ns1_DefRHS(x1,dd);
	s1+=staFunc->ScalProd(Pi,dd);
/*  -<Pi,B(Qi,Sr)>  */
        staFunc->AddScaledVector(Qi,aa,1.0,dd);
        staFunc->CopyVectorElements(dd,0,v0,0,nphase);
        staFunc->AddScaledVector(x,v0,0.5*ddx,x1);
        ns1_DefRHS(x1,dd);
	s-=staFunc->ScalProd(Pi,dd);
        staFunc->AddScaledVector(x,v0,-0.5*ddx,x1);
        ns1_DefRHS(x1,dd);
	s-=staFunc->ScalProd(Pi,dd);

        staFunc->AddScaledVector(Qi,aa,-1.0,dd);
        staFunc->CopyVectorElements(dd,0,v0,0,nphase);
        staFunc->AddScaledVector(x,v0,0.5*ddx,x1);
        ns1_DefRHS(x1,dd);
	s+=staFunc->ScalProd(Pi,dd);
        staFunc->AddScaledVector(x,v0,-0.5*ddx,x1);
        ns1_DefRHS(x1,dd);
	s+=staFunc->ScalProd(Pi,dd);
	s/=(ddx*ddx); 
/*  -<Pi,B(Qi,Si)>  */
        staFunc->AddScaledVector(Qi,bb,1.0,dd);
        staFunc->CopyVectorElements(dd,0,v0,0,nphase);
        staFunc->AddScaledVector(x,v0,0.5*ddx,x1);
        ns1_DefRHS(x1,dd);
	s1-=staFunc->ScalProd(Pi,dd);
        staFunc->AddScaledVector(x,v0,-0.5*ddx,x1);
        ns1_DefRHS(x1,dd);
	s1-=staFunc->ScalProd(Pi,dd);

        staFunc->AddScaledVector(Qi,bb,-1.0,dd);
        staFunc->CopyVectorElements(dd,0,v0,0,nphase);
        staFunc->AddScaledVector(x,v0,0.5*ddx,x1);
        ns1_DefRHS(x1,dd);
	s1+=staFunc->ScalProd(Pi,dd);
        staFunc->AddScaledVector(x,v0,-0.5*ddx,x1);
        ns1_DefRHS(x1,dd);
	s1+=staFunc->ScalProd(Pi,dd);
	s1/=(ddx*ddx); 
      }
     /* cubic terms */
     if (stagenData->Der3) {
        D3=SymD3(x);
        for (r=0.0,i=0; i<nphase; i++) {
           for (j=0; j<nphase; j++) {
              for (k=0; k<nphase; k++) {
	         for (l=0; l<nphase; l++) {
	           r+=D3Elem(i,j,k,l)*(Pr[i]*Qr[j]*(Qr[k]*Qr[l]+Qi[k]*Qi[l])+
		                       Pi[i]*Qi[l]*(Qi[j]*Qi[k]+Qr[j]*Qr[k]));
	         }
	      }
           }
        }	
        for (r1=0.0,i=0; i<nphase; i++) {
           for (j=0; j<nphase; j++) {
              for (k=0; k<nphase; k++) {
	         for (l=0; l<nphase; l++) {
	           r1+=D3Elem(i,j,k,l)*(Pr[i]*Qi[j]*(Qr[k]*Qr[l]+Qi[k]*Qi[l])-
		                        Pi[i]*Qr[l]*(Qi[j]*Qi[k]+Qr[j]*Qr[k]));
	         }
	      }
           }
        }	
     } else {
        staFunc->ClearVector(v0);

/* (2/3)<Pr,C(Qr,Qr,Qr)> */	
	staFunc->CopyVectorElements(Qr,0,v0,0,nphase);
        staFunc->AddScaledVector(x,v0,3*dddx/2,x1);
        ns1_DefRHS(x1,dd);
	r=(2.0/3.0)*staFunc->ScalProd(Pr,dd);
        staFunc->AddScaledVector(x,v0,dddx/2,x1);
        ns1_DefRHS(x1,dd);
	r-=2.0*staFunc->ScalProd(Pr,dd);
        staFunc->AddScaledVector(x,v0,-dddx/2,x1);
        ns1_DefRHS(x1,dd);
	r+=2.0*staFunc->ScalProd(Pr,dd);
        staFunc->AddScaledVector(x,v0,-3*dddx/2,x1);
        ns1_DefRHS(x1,dd);
	r-=(2.0/3.0)*staFunc->ScalProd(Pr,dd);
/* (2/3)<Pr,C(Qi,Qi,Qi)> */	
	staFunc->CopyVectorElements(Qi,0,v0,0,nphase);
        staFunc->AddScaledVector(x,v0,3*dddx/2,x1);
        ns1_DefRHS(x1,dd);
	r1=(2.0/3.0)*staFunc->ScalProd(Pr,dd);
        staFunc->AddScaledVector(x,v0,dddx/2,x1);
        ns1_DefRHS(x1,dd);
	r1-=2.0*staFunc->ScalProd(Pr,dd);
        staFunc->AddScaledVector(x,v0,-dddx/2,x1);
        ns1_DefRHS(x1,dd);
	r1+=2.0*staFunc->ScalProd(Pr,dd);
        staFunc->AddScaledVector(x,v0,-3*dddx/2,x1);
        ns1_DefRHS(x1,dd);
	r1-=(2.0/3.0)*staFunc->ScalProd(Pr,dd);
	
/* (2/3)<Pi,C(Qi,Qi,Qi)> */
	staFunc->CopyVectorElements(Qi,0,v0,0,nphase);
        staFunc->AddScaledVector(x,v0,3*dddx/2,x1);
        ns1_DefRHS(x1,dd);
	r+=(2.0/3.0)*staFunc->ScalProd(Pi,dd);
        staFunc->AddScaledVector(x,v0,dddx/2,x1);
        ns1_DefRHS(x1,dd);
	r-=2.0*staFunc->ScalProd(Pi,dd);
        staFunc->AddScaledVector(x,v0,-dddx/2,x1);
        ns1_DefRHS(x1,dd);
	r+=2.0*staFunc->ScalProd(Pi,dd);
        staFunc->AddScaledVector(x,v0,-3*dddx/2,x1);
        ns1_DefRHS(x1,dd);
	r-=(2.0/3.0)*staFunc->ScalProd(Pi,dd);
/* -(2/3)<Pi,C(Qr,Qr,Qr)> */
	staFunc->CopyVectorElements(Qr,0,v0,0,nphase);
        staFunc->AddScaledVector(x,v0,3*dddx/2,x1);
        ns1_DefRHS(x1,dd);
	r1-=(2.0/3.0)*staFunc->ScalProd(Pi,dd);
        staFunc->AddScaledVector(x,v0,dddx/2,x1);
        ns1_DefRHS(x1,dd);
	r1+=2.0*staFunc->ScalProd(Pi,dd);
        staFunc->AddScaledVector(x,v0,-dddx/2,x1);
        ns1_DefRHS(x1,dd);
	r1-=2.0*staFunc->ScalProd(Pi,dd);
        staFunc->AddScaledVector(x,v0,-3*dddx/2,x1);
        ns1_DefRHS(x1,dd);
	r1+=(2.0/3.0)*staFunc->ScalProd(Pi,dd);
	
/* (1/6) <Pr+Pi,C(Qr+Qi,Qr+Qi,Qr+Qi)> */
        staFunc->AddScaledVector(Qr,Qi,1.0,aa);
        staFunc->CopyVectorElements(aa,0,v0,0,nphase);
        staFunc->AddScaledVector(Pr,Pi,1.0,bb);
        staFunc->AddScaledVector(x,v0,3*dddx/2,x1);
        ns1_DefRHS(x1,dd);
	r+=staFunc->ScalProd(bb,dd)/6;
        staFunc->AddScaledVector(x,v0,dddx/2,x1);
        ns1_DefRHS(x1,dd);
	r-=staFunc->ScalProd(bb,dd)/2;
        staFunc->AddScaledVector(x,v0,-dddx/2,x1);
        ns1_DefRHS(x1,dd);
	r+=staFunc->ScalProd(bb,dd)/2;
        staFunc->AddScaledVector(x,v0,-3*dddx/2,x1);
        ns1_DefRHS(x1,dd);
	r-=staFunc->ScalProd(bb,dd)/6;
/* (1/6) <Pr-Pi,C(Qr+Qi,Qr+Qi,Qr+Qi)> */
        staFunc->AddScaledVector(Qr,Qi,1.0,aa);
        staFunc->CopyVectorElements(aa,0,v0,0,nphase);
        staFunc->AddScaledVector(Pr,Pi,-1.0,bb);
        staFunc->AddScaledVector(x,v0,3*dddx/2,x1);
        ns1_DefRHS(x1,dd);
	r1+=staFunc->ScalProd(bb,dd)/6;
        staFunc->AddScaledVector(x,v0,dddx/2,x1);
        ns1_DefRHS(x1,dd);
	r1-=staFunc->ScalProd(bb,dd)/2;
        staFunc->AddScaledVector(x,v0,-dddx/2,x1);
        ns1_DefRHS(x1,dd);
	r1+=staFunc->ScalProd(bb,dd)/2;
        staFunc->AddScaledVector(x,v0,-3*dddx/2,x1);
        ns1_DefRHS(x1,dd);
	r1-=staFunc->ScalProd(bb,dd)/6;

/* (1/6) <Pr-Pi,C(Qr-Qi,Qr-Qi,Qr-Qi)> */
        staFunc->AddScaledVector(Qr,Qi,-1.0,aa);
        staFunc->CopyVectorElements(aa,0,v0,0,nphase);
        staFunc->AddScaledVector(Pr,Pi,-1.0,bb);
        staFunc->AddScaledVector(x,v0,3*dddx/2,x1);
        ns1_DefRHS(x1,dd);
	r+=staFunc->ScalProd(bb,dd)/6;
        staFunc->AddScaledVector(x,v0,dddx/2,x1);
        ns1_DefRHS(x1,dd);
	r-=staFunc->ScalProd(bb,dd)/2;
        staFunc->AddScaledVector(x,v0,-dddx/2,x1);
        ns1_DefRHS(x1,dd);
	r+=staFunc->ScalProd(bb,dd)/2;
        staFunc->AddScaledVector(x,v0,-3*dddx/2,x1);
        ns1_DefRHS(x1,dd);
	r-=staFunc->ScalProd(bb,dd)/6;
        r/=(dddx*dddx*dddx);                                         
/* -(1/6) <Pr+Pi,C(Qr-Qi,Qr-Qi,Qr-Qi)> */
        staFunc->AddScaledVector(Qr,Qi,-1.0,aa);
        staFunc->CopyVectorElements(aa,0,v0,0,nphase);
        staFunc->AddScaledVector(Pr,Pi,1.0,bb);
        staFunc->AddScaledVector(x,v0,3*dddx/2,x1);
        ns1_DefRHS(x1,dd);
	r1-=staFunc->ScalProd(bb,dd)/6;
        staFunc->AddScaledVector(x,v0,dddx/2,x1);
        ns1_DefRHS(x1,dd);
	r1+=staFunc->ScalProd(bb,dd)/2;
        staFunc->AddScaledVector(x,v0,-dddx/2,x1);
        ns1_DefRHS(x1,dd);
	r1-=staFunc->ScalProd(bb,dd)/2;
        staFunc->AddScaledVector(x,v0,-3*dddx/2,x1);
        ns1_DefRHS(x1,dd);
	r1+=staFunc->ScalProd(bb,dd)/6;
        r1/=(dddx*dddx*dddx);                                         
    }
    *res=(lambda*(s+r)+omega*(s1+r1))/2/(1.0-lambda);
  }
end:	       
  test[0]=*res;
  return 0;
}

Entry(Int2) ns1_ProcDegenNS(Int2 Ind, Vector x, Vector v, CharPtr *msg) {
Int2 i,j,m,imax,jmax;
VAR max1,max2,a11,a12,a21,a22,d;
  stagenData->BifDataOutLen=0;
  if (Ind) 
    sprintf(buf,"Degenerate Neimark-Sacker (?)");  
  else {
    m=nphase*(nphase-1)/2;
/* reevaluate Jacobian */
    JacRHS (x,C1);
    staFunc->CopyMatrixBlock(C1,0,0,A,0,0,nphase,nphase);
/* find pair in (g11,g12), (g21,g22) with max absolute value */
    max1=(fabs(VB1[m])>fabs(VB2[m]))?VB1[m]:VB2[m];
    max2=(fabs(VB1[m+1])>fabs(VB2[m+1]))?VB1[m+1]:VB2[m+1];
    if (fabs(max1)>fabs(max2))
     imax=m;
    else imax=m+1;    
    for(i=0;i<m;i++) {
      Q[i]=VB2[imax]*VB1[i]-VB1[imax]*VB2[i];
    }
/* find a component of Q with max(fabs) */
    imax=1;
    jmax=0;
    d=fabs(Q[0]);
    for (i=0; i<nphase; i++) {
      for (j=0; j<i; j++) {
        if (fabs(Q[i*(i-1)/2+j])>d) {
          imax=i;
	  jmax=j;
	  d=fabs(Q[i*(i-1)/2+j]);
        }
      }
    }
    d=Q[imax*(imax-1)/2+jmax];

/* compute eigenvectors */
    staFunc->ClearVector(V);
    staFunc->ClearVector(W);
    for (i=0; i<nphase; i++) {
      if (i<imax) V[i]=Q[imax*(imax-1)/2+i]/d;
      if (i>imax) V[i]=-Q[i*(i-1)/2+imax]/d;
      if (i<jmax) W[i]=-Q[jmax*(jmax-1)/2+i]/d;
      if (i>jmax) W[i]=Q[i*(i-1)/2+jmax]/d;
    }

/* compute kappa */
    staFunc->MultiplyMatrixVector(A,V,F1);  /* F1=A*V  */
    staFunc->MultiplyMatrixVector(A,W,F2);  /* F2=A*W  */

    a11=staFunc->ScalProd(V,V)*staFunc->ScalProd(W,F2);
    a12=staFunc->ScalProd(W,W)*staFunc->ScalProd(V,F1);
    a21=staFunc->ScalProd(V,W)*staFunc->ScalProd(W,F1);
    a22=staFunc->ScalProd(W,V)*staFunc->ScalProd(V,F2);

    d=staFunc->ScalProd(V,V)*staFunc->ScalProd(W,W)
     -staFunc->ScalProd(V,W)*staFunc->ScalProd(V,W);

/* orthogonalize and normalize eigenvectors */
    staFunc->NormalizeVector(V);
    staFunc->AddScaledVector(W,V,-staFunc->ScalProd(V,W),L);
    staFunc->NormalizeVector(L);
   
/*  provide bifurcation data for DN -> NS starter */
/*  BifVector[0]:			 _itmap */
/*  BifVector[1] to [nphase]:  	         NS continuation vector*/
/*  BifVector[nphase+1] to [2*nphase]: 	 global vector L */
/*  BifVector[2*nphase+1]:               kappa=cos(theta)  */
    BifVector[0]=(VAR)_itmap;
    MemCopy(BifVector+1,V,sizeof(FloatHi)*nphase);
    MemCopy(BifVector+nphase+1,L,sizeof(FloatHi)*nphase);
    BifVector[2*nphase+1]=(a11+a12-a21-a22)/(2.0*d);
    stagenData->BifDataOutLen=sizeof(FloatHi)*(2*nphase+2);
    sprintf(buf,"Degenerate Neimark-Sacker: cos(theta)=%g",BifVector[2*nphase+1]);    
  }  
  *msg=buf;
  return 0;
}

Entry(Int2) ns1_TestR1(Vector x, Vector v, FloatHiPtr res) {
Int2 i,j,imax,jmax;
VAR d,a11,a12,a21,a22;
  *res=1;
  if (nphase == 1) goto end;

/* reevaluate Jacobian */
   if (JacRHS (x,C1)) return 1;
   staFunc->CopyMatrixBlock(C1,0,0,A,0,0,nphase,nphase);

/* it is assumed that Q is correct */
/* find a component of Q with max(fabs) */
  imax=1;
  jmax=0;
  d=fabs(Q[0]);
  for (i=0; i<nphase; i++) {
    for (j=0; j<i; j++) {
      if (fabs(Q[i*(i-1)/2+j])>d) {
        imax=i;
	jmax=j;
	d=fabs(Q[i*(i-1)/2+j]);
      }
    }
  }
  d=Q[imax*(imax-1)/2+jmax];

/* compute eigenvectors */
  staFunc->ClearVector(V);
  staFunc->ClearVector(W);
  for (i=0; i<nphase; i++) {
    if (i<imax) V[i]=Q[imax*(imax-1)/2+i]/d;
    if (i>imax) V[i]=-Q[i*(i-1)/2+imax]/d;
    if (i<jmax) W[i]=-Q[jmax*(jmax-1)/2+i]/d;
    if (i>jmax) W[i]=Q[i*(i-1)/2+jmax]/d;
  }

/* compute kappa */
  staFunc->MultiplyMatrixVector(A,V,F1);  /* F1=A*V  */
  staFunc->MultiplyMatrixVector(A,W,F2);  /* F2=A*W  */  

  a11=staFunc->ScalProd(V,V)*staFunc->ScalProd(W,F2);
  a12=staFunc->ScalProd(W,W)*staFunc->ScalProd(V,F1);
  a21=staFunc->ScalProd(V,W)*staFunc->ScalProd(W,F1);
  a22=staFunc->ScalProd(W,V)*staFunc->ScalProd(V,F2);

  d=staFunc->ScalProd(V,V)*staFunc->ScalProd(W,W)
   -staFunc->ScalProd(V,W)*staFunc->ScalProd(V,W);
  
  *res=((a11+a12-a21-a22)/(2.0*d))-1.0;
end:
  test[1]=*res;  
  return 0;
}

Entry(Int2) ns1_ProcR1(Int2 Ind, Vector x, Vector v, CharPtr *msg) {
  stagenData->BifDataOutLen=0;
  if (Ind) 
    sprintf(buf,"Resonance 1:1 (?)");
  else {
Vector v1,W0,W1;
Matrix D;
Int2 i,j,k;
double a20,b20,b11,f0,f1,f2,f3,f4,f5,d;
    v1=staFunc->CreateVector(2*nphase);
    W0=staFunc->CreateVector(nphase);
    W1=staFunc->CreateVector(nphase);
    D=staFunc->CreateMatrix(2*nphase,2*nphase);

/*   compute (generalized)eigenvectors   */
    if (JacRHS (x,C1)) return 1;
    staFunc->CopyMatrixBlock(C1,0,0,A,0,0,nphase,nphase);
    for (i=0; i<nphase; i++) A[i][i]-=1.0;
    
    staFunc->CopyMatrixBlock(A,0,0,D,0,0,nphase,nphase);
    staFunc->CopyMatrixBlock(A,0,0,D,nphase,nphase,nphase,nphase);
    for (i=0; i<nphase; i++) D[i+nphase][i]=-1;
    i=staFunc->SngSolve(D,_eps,v1);
    if (i != 2*nphase-2) {
      sprintf(buf,"1:1 Resonance: eigenvectors (?)");
      goto final;
    }
    staFunc->CopyVectorElements(v1,0,V,0,nphase);       /* eigenvector V0 */    
    d=sqrt(staFunc->ScalProd(V,V));
    staFunc->NormalizeVector(V);
    staFunc->CopyVectorElements(v1,nphase,W,0,nphase);  /* generalized eigenvector V1 */
    staFunc->MultiplyVectorScalar(W,1.0/d,W);
    staFunc->AddScaledVector(W,V,-staFunc->ScalProd(V,W),W);

/*  provide bifurcation data for R1->* starters */
/*  BifVector[0]:			 _itmap */
/*  BifVector[1] to [nphase]:            eigenvector V (A*V=V, <V,V>=1)*/
/*  BifVector[nphase+1]:                 kappa=1.0  */
    BifVector[0]=(VAR)_itmap;
    MemCopy(BifVector+1,V,sizeof(FloatHi)*nphase);
    BifVector[nphase+1]=1.0;
    stagenData->BifDataOutLen=sizeof(FloatHi)*(nphase+2);
    
/* compute the adjoint eigenvectors */
    staFunc->CopyMatrixBlock(C1,0,0,A,0,0,nphase,nphase);
    for (i=0; i<nphase; i++) A[i][i]-=1.0;
    staFunc->TransposeMatrix(A,B);
    staFunc->ClearMatrix(D);
    staFunc->CopyMatrixBlock(B,0,0,D,0,0,nphase,nphase);
    staFunc->CopyMatrixBlock(B,0,0,D,nphase,nphase,nphase,nphase);
    for (i=0; i<nphase; i++) D[i+nphase][i]=-1;
    i=staFunc->SngSolve(D,_eps,v1);
    if (i != 2*nphase-2) {
      sprintf(buf,"1:1 Resonance: adjoint eigenvectors (?)");
      goto final;
    }
    staFunc->CopyVectorElements(v1,0,W1,0,nphase);       /* adjoint eigenvector W1 */
    staFunc->CopyVectorElements(v1,nphase,W0,0,nphase);  /* adjoint generalized eigenvector W0 */
    d=staFunc->ScalProd(V,W0);
    staFunc->MultiplyVectorScalar(W1,1.0/d,W1);
    staFunc->AddScaledVector(W0,W1,-staFunc->ScalProd(W0,W),W0);
    staFunc->MultiplyVectorScalar(W0,1.0/d,W0);

/* compute the coefficients of the normal form map: x'=x+y; y'=Y+ax^2+bxy */
/*  a20=<W0,B(V,V)>, b20=<W1,B(V,V)>,  b11=<W1,B(V,W)> */
     if (stagenData->Der2)  {
        H=SymHess(x);
        for (a20=0,b20=0,b11=0,i=0; i<nphase; i++) {
           for (j=0; j<nphase; j++) {
              for (k=0; k<nphase; k++) { 
	         d=HessianElem(i,j,k);
                 a20+=W0[i]*d*V[j]*V[k];
                 b20+=W1[i]*d*V[j]*V[k];
                 b11+=W1[i]*d*V[j]*W[k];
	      }
           }
        }

     } else {
	staFunc->ClearVector(v0);
/*  a20=<W0,B(V,V)>, b20=<W1,B(V,V)> */
	staFunc->CopyVectorElements(V,0,v0,0,nphase);
        ns1_DefRHS(x,F1);
        f0=staFunc->ScalProd(W0,F1);
        f3=staFunc->ScalProd(W1,F1);
        staFunc->AddScaledVector(x,v0,ddx,x1);
        ns1_DefRHS(x1,F1);
        f1=staFunc->ScalProd(W0,F1);
        f4=staFunc->ScalProd(W1,F1);
        staFunc->AddScaledVector(x,v0,-ddx,x1);
        ns1_DefRHS(x1,F1);
        f2=staFunc->ScalProd(W0,F1);
        f5=staFunc->ScalProd(W1,F1);
	a20=(f1-2*f0+f2)/ddx/ddx;
	b20=(f4-2*f3+f5)/ddx/ddx;

/*  b11=<W1,B(V,W)>  */
        staFunc->AddScaledVector(V,W,1.0,F1);
        staFunc->CopyVectorElements(F1,0,v0,0,nphase);
        staFunc->AddScaledVector(x,v0,0.5*ddx,x1);
        ns1_DefRHS(x1,F1);
	b11=staFunc->ScalProd(W1,F1);
        staFunc->AddScaledVector(x,v0,-0.5*ddx,x1);
        ns1_DefRHS(x1,F1);
	b11+=staFunc->ScalProd(W1,F1);
        staFunc->AddScaledVector(V,W,-1.0,F1);
        staFunc->CopyVectorElements(F1,0,v0,0,nphase);
        staFunc->AddScaledVector(x,v0,0.5*ddx,x1);
        ns1_DefRHS(x1,F1);
	b11-=staFunc->ScalProd(W1,F1);
        staFunc->AddScaledVector(x,v0,-0.5*ddx,x1);
        ns1_DefRHS(x1,F1);
	b11-=staFunc->ScalProd(W1,F1);
	b11/=(ddx*ddx); 
     }
    if (b20>=0)
      sprintf(buf,"1:1 Resonance: a=%g, b=%g",b20/2,a20+b11-b20);
    else
      sprintf(buf,"1:1 Resonance: a=%g, b=%g",-b20/2,-(a20+b11-b20));
final:
    D=staFunc->DestroyMatrix(D);
    v1=staFunc->DestroyVector(v1);
    W0=staFunc->DestroyVector(W0);
    W1=staFunc->DestroyVector(W1);
  }
  *msg=buf;
  return 0;
}

/* det of the Jacobian matrix */
Entry(Int2) ns1_TestFoldNS(Vector x, Vector v, FloatHiPtr res) {
Int2 i;
  *res=1.0;
  if (nphase < 2) goto end;
  if (JacRHS (x,C1)) return 1;
  staFunc->CopyMatrixBlock(C1,0,0,A,0,0,nphase,nphase);
  for (i=0; i<nphase; i++) A[i][i]-=1.0;
  if (staFunc->Decomp(A)) {
    *res=0.0;
  } else {
    staFunc->GetMatrixData(A,res,NULL);
  } 
end:
  test[2]=*res; 
  return 0;
}

Entry(Int2) ns1_ProcFoldNS(Int2  Ind, Vector x, Vector v, CharPtr *msg) {
Int2 i,j,imax,jmax;
VAR d,a11,a12,a21,a22;
  stagenData->BifDataOutLen=0;
  if (Ind) 
    sprintf(buf,"Fold - Neimark-Sacker (?)"); 
  else {
/* reevaluate Jacobian */
   if (JacRHS (x,C1)) return 1;
   staFunc->CopyMatrixBlock(C1,0,0,A,0,0,nphase,nphase);

/* it is assumed that Q is correct */
/* find a component of Q with maxfabs */
  imax=1;
  jmax=0;
  d=fabs(Q[0]);
  for (i=0; i<nphase; i++) {
    for (j=0; j<i; j++) {
      if (fabs(Q[i*(i-1)/2+j])>d) {
        imax=i;
	jmax=j;
	d=fabs(Q[i*(i-1)/2+j]);
      }
    }
  }
  d=Q[imax*(imax-1)/2+jmax];

/* compute eigenvectors */
  staFunc->ClearVector(V);
  staFunc->ClearVector(W);
  for (i=0; i<nphase; i++) {
    if (i<imax) V[i]=Q[imax*(imax-1)/2+i]/d;
    if (i>imax) V[i]=-Q[i*(i-1)/2+imax]/d;
    if (i<jmax) W[i]=-Q[jmax*(jmax-1)/2+i]/d;
    if (i>jmax) W[i]=Q[i*(i-1)/2+jmax]/d;
  }

/* compute kappa */
  staFunc->MultiplyMatrixVector(A,V,F1);  /* F1=A*V  */
  staFunc->MultiplyMatrixVector(A,W,F2);  /* F2=A*W  */

  a11=staFunc->ScalProd(V,V)*staFunc->ScalProd(W,F2);
  a12=staFunc->ScalProd(W,W)*staFunc->ScalProd(V,F1);
  a21=staFunc->ScalProd(V,W)*staFunc->ScalProd(W,F1);
  a22=staFunc->ScalProd(W,V)*staFunc->ScalProd(V,F2);

  d=staFunc->ScalProd(V,V)*staFunc->ScalProd(W,W)
   -staFunc->ScalProd(V,W)*staFunc->ScalProd(V,W);
  
/* orthogonalize and normalize eigenvectors */
    staFunc->NormalizeVector(V);
    staFunc->AddScaledVector(W,V,-staFunc->ScalProd(V,W),L);
    staFunc->NormalizeVector(L);
  
/* provide bifurcation data for FN -> NS starters */
    if (JacRHS (x,C1)) return 1;
    staFunc->CopyMatrixBlock(C1,0,0,A,0,0,nphase,nphase);
    for (i=0; i<nphase; i++) A[i][i]-=1.0;
    if (staFunc->SngSolve(A,_eps,W) != nphase-1) {
      sprintf(buf,"Fold + Neimark-Sacker: fold vector (?)"); 
      *msg=buf;
      return 0;
    }   
    staFunc->NormalizeVector(W);
/*  BifVector[0]:                        _itmap */
/*  BifVector[1] to [nphase]:            fold vector (A*W=W, <W,W>=1) */
/*  BifVector[nphase+1] to [2*nphase]:   NS continuation vector */
/*  BifVector[2*nphase+1] to [3*nphase]: global vector L */
/*  BifVector[3*nphase+1]:               kappa  */
    BifVector[0]=(VAR)_itmap;
    MemCopy(BifVector+1,W,sizeof(FloatHi)*nphase);
    MemCopy(BifVector+nphase+1,V,sizeof(FloatHi)*nphase);
    MemCopy(BifVector+2*nphase+1,L,sizeof(FloatHi)*nphase);
    BifVector[3*nphase+1]=(a11+a12-a21-a22)/(2.0*d);
    stagenData->BifDataOutLen=sizeof(FloatHi)*(3*nphase+2);
    
    if (BifVector[3*nphase+1]>0) {
      sprintf(buf,"Fold + Neimark-Sacker: cos(theta)=%g",BifVector[3*nphase+1]);
    } else {
      sprintf(buf,"Fold + neutral saddle: kappa=%g",BifVector[3*nphase+1]);
    }      
  }  
  *msg=buf;
  return 0;
}

/* det(A+I) */
Entry(Int2) ns1_TestFlipNS(Vector x, Vector v, FloatHiPtr res) {
Int2 i;
  *res=1.0;
  if (nphase < 2) goto end;
  if (JacRHS (x,C1)) return 1;
  staFunc->CopyMatrixBlock(C1,0,0,A,0,0,nphase,nphase);
  for (i=0; i<nphase; i++) A[i][i]+=1.0;
  if (staFunc->Decomp(A)) {
    *res=0.0;
  } else {
    staFunc->GetMatrixData(A,res,NULL);
  } 
end:
  test[3]=*res; 
  return 0;
}

Entry(Int2) ns1_ProcFlipNS(Int2  Ind, Vector x, Vector v, CharPtr *msg) {
Int2 i,j,imax,jmax;
VAR d,a11,a12,a21,a22;
  stagenData->BifDataOutLen=0;
  if (Ind) 
    sprintf(buf,"Flip - Neimark-Sacker (?)"); 
  else {
/* reevaluate Jacobian */
   if (JacRHS (x,C1)) return 1;
   staFunc->CopyMatrixBlock(C1,0,0,A,0,0,nphase,nphase);

/* it is assumed that Q is correct */
/* find a component of Q with maxfabs */
  imax=1;
  jmax=0;
  d=fabs(Q[0]);
  for (i=0; i<nphase; i++) {
    for (j=0; j<i; j++) {
      if (fabs(Q[i*(i-1)/2+j])>d) {
        imax=i;
	jmax=j;
	d=fabs(Q[i*(i-1)/2+j]);
      }
    }
  }
  d=Q[imax*(imax-1)/2+jmax];

/* compute eigenvectors */
  staFunc->ClearVector(V);
  staFunc->ClearVector(W);
  for (i=0; i<nphase; i++) {
    if (i<imax) V[i]=Q[imax*(imax-1)/2+i]/d;
    if (i>imax) V[i]=-Q[i*(i-1)/2+imax]/d;
    if (i<jmax) W[i]=-Q[jmax*(jmax-1)/2+i]/d;
    if (i>jmax) W[i]=Q[i*(i-1)/2+jmax]/d;
  }

/* compute kappa */
  staFunc->MultiplyMatrixVector(A,V,F1);  /* F1=A*V  */
  staFunc->MultiplyMatrixVector(A,W,F2);  /* F2=A*W  */

  a11=staFunc->ScalProd(V,V)*staFunc->ScalProd(W,F2);
  a12=staFunc->ScalProd(W,W)*staFunc->ScalProd(V,F1);
  a21=staFunc->ScalProd(V,W)*staFunc->ScalProd(W,F1);
  a22=staFunc->ScalProd(W,V)*staFunc->ScalProd(V,F2);

  d=staFunc->ScalProd(V,V)*staFunc->ScalProd(W,W)
   -staFunc->ScalProd(V,W)*staFunc->ScalProd(V,W);
  
/* orthogonalize and normalize eigenvectors */
    staFunc->NormalizeVector(V);
    staFunc->AddScaledVector(W,V,-staFunc->ScalProd(V,W),L);
    staFunc->NormalizeVector(L);
  
/* provide bifurcation data for PN -> NS starters */
    if (JacRHS (x,C1)) return 1;
    staFunc->CopyMatrixBlock(C1,0,0,A,0,0,nphase,nphase);
    for (i=0; i<nphase; i++) A[i][i]+=1.0;
    if (staFunc->SngSolve(A,_eps,W) != nphase-1) {
      sprintf(buf,"Flip + Neimark-Sacker: flip vector (?)"); 
      *msg=buf;
      return 0;
    }   
    staFunc->NormalizeVector(W);
/*  BifVector[0]:                        _itmap */
/*  BifVector[1] to [nphase]:            fold vector (A*W=W, <W,W>=1) */
/*  BifVector[nphase+1] to [2*nphase]:   NS continuation vector */
/*  BifVector[2*nphase+1] to [3*nphase]: global vector L */
/*  BifVector[3*nphase+1]:               kappa  */
    BifVector[0]=(VAR)_itmap;
    MemCopy(BifVector+1,W,sizeof(FloatHi)*nphase);
    MemCopy(BifVector+nphase+1,V,sizeof(FloatHi)*nphase);
    MemCopy(BifVector+2*nphase+1,L,sizeof(FloatHi)*nphase);
    BifVector[3*nphase+1]=(a11+a12-a21-a22)/(2.0*d);
    stagenData->BifDataOutLen=sizeof(FloatHi)*(3*nphase+2);
    
    if (BifVector[3*nphase+1]>0) {
      sprintf(buf,"Flip + Neimark-Sacker: cos(theta)=%g",BifVector[3*nphase+1]);
    } else {
      sprintf(buf,"Flip + neutral saddle: kappa=%g",BifVector[3*nphase+1]);
    }      
  }  
  *msg=buf;
  return 0;
}  

/*  DH test function */
Entry(Int2) ns1_TestDoubleNS(Vector x, Vector v, FloatHiPtr res) {
Int2 m, Ret;
Matrix BPT;
  m=nphase*(nphase-1)/2;
  BPT=staFunc->CreateMatrix(m+1,m+1);
  staFunc->CopyMatrixBlock(BP,0,0,BPT,0,0,m+1,m+1);
  Ret=staFunc->Decomp(BPT);
  if (Ret) return 1;
  else {
    staFunc->GetMatrixData(BPT,res,NULL);
  }
  BPT=staFunc->DestroyMatrix(BPT);
  test[4]=*res;
  return 0;
}

Entry(Int2) ns1_ProcDoubleNS(Int2 Ind, Vector x, Vector v, CharPtr *msg) {
  stagenData->BifDataOutLen=0;
  if (Ind) 
    sprintf(buf,"Double NS (?)"); 
  else 
{int i,ii,j,jj,imin,jmin,imax,jmax,maxV,imaxV,maxW,imaxW,m;
 double max1,max2,r,s,d,phi,beta,gamma,lambda,lambda1,omega,theta,kappa,a11,a12,a21,a22;
 Vector Re2,Im2;

    m=nphase*(nphase-1)/2;
/* find pair in (g11,g12), (g21,g22) with max absolute value */
    max1=(fabs(VB1[m])>fabs(VB2[m]))?VB1[m]:VB2[m];
    max2=(fabs(VB1[m+1])>fabs(VB2[m+1]))?VB1[m+1]:VB2[m+1];
    if (fabs(max1)>fabs(max2))
     imax=m;
    else imax=m+1;    
    for(i=0;i<m;i++) {
      Q[i]=VB2[imax]*VB1[i]-VB1[imax]*VB2[i];
    }
       
/* find a component of Q with max(fabs) */
    imax=1;
    jmax=0;
    d=fabs(Q[0]);
    for (i=0; i<nphase; i++) {
      for (j=0; j<i; j++) {
        if (fabs(Q[i*(i-1)/2+j])>d) {
          imax=i;
	  jmax=j;
	  d=fabs(Q[i*(i-1)/2+j]);
        }
      }
    }
    d=Q[imax*(imax-1)/2+jmax];

/* compute eigenvectors */
    staFunc->ClearVector(V);
    staFunc->ClearVector(W);
    for (i=0; i<nphase; i++) {
      if (i<imax) V[i]=Q[imax*(imax-1)/2+i]/d;
      if (i>imax) V[i]=-Q[i*(i-1)/2+imax]/d;
      if (i<jmax) W[i]=-Q[jmax*(jmax-1)/2+i]/d;
      if (i>jmax) W[i]=Q[i*(i-1)/2+jmax]/d;
    }
/* compute kappa */
  staFunc->MultiplyMatrixVector(A,V,F1);  /* F1=A*V  */
  staFunc->MultiplyMatrixVector(A,W,F2);  /* F2=A*W  */  

  a11=staFunc->ScalProd(V,V)*staFunc->ScalProd(W,F2);
  a12=staFunc->ScalProd(W,W)*staFunc->ScalProd(V,F1);
  a21=staFunc->ScalProd(V,W)*staFunc->ScalProd(W,F1);
  a22=staFunc->ScalProd(W,V)*staFunc->ScalProd(V,F2);

  d=staFunc->ScalProd(V,V)*staFunc->ScalProd(W,W)
   -staFunc->ScalProd(V,W)*staFunc->ScalProd(V,W);
  kappa=((a11+a12-a21-a22)/(2.0*d));

  staFunc->NormalizeVector(V);
  MemCopy(BifVector+1,V,sizeof(FloatHi)*nphase);
 
  Re2=staFunc->CreateVector(nphase-2); 
  Im2=staFunc->CreateVector(nphase-2); 

/* find two basis vectors with max projection to Span(V,W) */
  imaxV=0;
  maxV=fabs(V[0]);
  imaxW=0;
  maxW=fabs(W[0]);
  for (i=0; i<nphase; i++) {
    if (fabs(V[i])>maxV) {
      imaxV=i;
      maxV=fabs(V[i]);
    }
    if (fabs(W[i])>maxW) {
      imaxW=i;
      maxW=fabs(W[i]);
    }
  }
  if (imaxV==imaxW) {
    imaxW++;
    if (imaxW==nphase) imaxW=0;
  }
 
/* fill in the matrix B with vectors to be orthogonalized as rows */
  staFunc->ClearMatrix(B);
  staFunc->CopyVectorMatrixRow(V,B,0);
  staFunc->CopyVectorMatrixRow(W,B,1);
  for (i=0,ii=2,jj=0; i<nphase; i++,jj++) {
    if (i==imaxV || i==imaxW) continue;
    B[ii++][jj]=1;
  }

/* Gram-Schmidt orthogonalization of rows of B */
  for (i=2; i<nphase; i++) {
    staFunc->CopyMatrixRowVector(B,i,V);
    staFunc->CopyMatrixRowVector(B,i,F1);
    for (j=0; j<i; j++) {
      staFunc->CopyMatrixRowVector(B,j,W);
      s=staFunc->ScalProd(V,W);
      staFunc->AddScaledVector(F1,W,-s,F1);
    }
    staFunc->NormalizeVector(F1);
    staFunc->CopyVectorMatrixRow(F1,B,i);
  }

/* compute elements of the restriction matrix AH */ 
  if (JacRHS(x,C1)) return 1;
  staFunc->CopyMatrixBlock(C1,0,0,AA,0,0,nphase,nphase);
  staFunc->ClearMatrix(AH); 
  for (i=2; i<nphase; i++) {
    staFunc->CopyMatrixRowVector(B,i,V);
    staFunc->MultiplyMatrixVector(AA,V,W);
    for (j=2; j<nphase; j++) {
      staFunc->CopyMatrixRowVector(B,j,V);
      AH[j-2][i-2]=staFunc->ScalProd(V,W);
    }  
  }
  staFunc->EigenValues(AH,Re2,Im2);
/* find the pair of extra multipliers on the unit circle */
  s=fabs(Re2[0]*Re2[1]-Im2[0]*Im2[1]-1.0)+fabs(Re2[0]*Im2[1]+Re2[1]*Im2[0]);
  imin=0;
  jmin=1;
  for (i=0; i<nphase; i++) {
    for (j=i+1; j<nphase; j++) {
      d=fabs(Re2[i]*Re2[j]-Im2[i]*Im2[j]-1.0)+fabs(Re2[i]*Im2[j]+Re2[j]*Im2[i]);
      if (d<s) {
        imin=i;
        jmin=j;
        s=d;
      }
    }
  }
  JacRHS(x,C1);
    if (fabs(Im2[imin])+fabs(Im2[jmin])==0) {
/* Neutral saddle */
      lambda=Re2[imin];
      lambda1=Re2[jmin];
      /* compute real eigenvectors associated with lambda and lambda1=1/lambda */
      staFunc->CopyMatrixBlock(C1,0,0,A,0,0,nphase,nphase);
      for (i=0; i<nphase; i++) A[i][i]-=lambda; 
      if (staFunc->SngSolve(A,_eps,V) != nphase-1) {
        sprintf(buf,"Double NS: lambda=%g, eigenvectors(?)",lambda);
        goto polufinal;
     } 
      staFunc->CopyMatrixBlock(C1,0,0,A,0,0,nphase,nphase);
      for (i=0; i<nphase; i++) A[i][i]-=lambda1;
      if (staFunc->SngSolve(A,_eps,W) != nphase-1) {
        sprintf(buf,"Double NS: lambda1=%g, eigenvectors(?)",lambda1);
        goto polufinal;
      } 
      staFunc->NormalizeVector(V);   
      staFunc->NormalizeVector(W);
      staFunc->AddScaledVector(V,W,-1.0,F1);             
      staFunc->NormalizeVector(F1);
      staFunc->AddScaledVector(V,W,1.0,V);             
      staFunc->NormalizeVector(V);
/* provide bifurcation data for NN -> NS starter */
/*  Secondary branch: */
/*  BifVector[2*nphase+2] to [3*nphase+1]: 2-NS continuation vector*/
/*  BifVector[3*nphase+2] to [4*nphase+1]: 2-global vector L */
/*  BifVector[4*nphase+2]:                 2-kappa  */
      MemCopy(BifVector+2*nphase+2,V,sizeof(FloatHi)*nphase);
      MemCopy(BifVector+3*nphase+2,F1,sizeof(FloatHi)*nphase);
      BifVector[4*nphase+2]=(lambda+lambda1)/2.0;   /* kappa */
      stagenData->BifDataOutLen=sizeof(FloatHi)*(4*nphase+3);
      if (kappa <= 0.0) {    
        sprintf(buf,"Double neutral saddle");
      } else {
        sprintf(buf,"Neimark-Sacker + neutral saddle");
      }
    } else {
/* Neimark-Sacker */
      if (fabs(Re2[imin]) > 0.5) {
        theta=atan(fabs(Im2[imin]/Re2[imin]));
      } else {
        theta=PID2-atan(fabs(Re2[imin]/Im2[imin]));
      }
      if (Re2[imin] < 0) theta=PI-theta;
      if (Im2[imin] < 0) theta=-theta;
      lambda=cos(theta);
      omega=sin(theta);

      /* compute real and imaginary part of the eigenvector */
      staFunc->ClearMatrix(BB);
      staFunc->CopyMatrixBlock(C1,0,0,BB,0,0,nphase,nphase);
      staFunc->CopyMatrixBlock(C1,0,0,BB,nphase,nphase,nphase,nphase);
      for (i=0; i<nphase; i++) {
        BB[i][i+nphase]=omega;
        BB[i][i]-=lambda;
        BB[i+nphase][i]=-omega;
        BB[i+nphase][i+nphase]-=lambda;
      }
      if (staFunc->SngSolve(BB,_eps,VV) != 2*nphase-2) {
        sprintf(buf,"Double NS: sin_2=%g, cos_2=%g, eigenvectors(?)", omega,lambda);
        goto polufinal;
      }
      staFunc->CopyVectorElements(VV,0,Qr,0,nphase);
      staFunc->CopyVectorElements(VV,nphase,Qi,0,nphase);

      /* normalize real and imaginary parts */
      d=staFunc->ScalProd(Qr,Qr);
      s=staFunc->ScalProd(Qi,Qi);
      r=staFunc->ScalProd(Qr,Qi);
      if (r!=0) {
        beta=sqrt(fabs(4.0*r*r+(s-d)*(s-d)));
        if (2.0*fabs(2.0*r) > beta) {
          phi=PID2-atan(fabs((s-d)/(2.0*r)));
        } else {
          phi=atan(fabs(2.0*r/(s-d)));
        }
        if (r < 0) phi=PI-phi;
        if ((s-d) < 0) phi=-phi;
        phi/=2.0;
        for (i=0; i<nphase; i++) {
          VV[i]=Qr[i]*cos(phi)-Qi[i]*sin(phi);
          VV[i+nphase]=Qr[i]*sin(phi)+Qi[i]*cos(phi);
        }
      }
      staFunc->CopyVectorElements(VV,0,Qr,0,nphase);
      d=staFunc->ScalProd(Qr,Qr);
      gamma=1/sqrt(d);
      staFunc->MultiplyVectorScalar(VV,gamma,VV);
      staFunc->CopyVectorElements(VV,nphase,F1,0,nphase);
      staFunc->NormalizeVector(F1); 

/* provide bifurcation data for NN -> NS starter */
/*  Secondary branch: */
/*  BifVector[2*nphase+2] to [3*nphase+1]: 2-NS continuation vector*/
/*  BifVector[3*nphase+2] to [4*nphase+1]: 2-global vector L */
/*  BifVector[4*nphase+2]:                 2-kappa  */
      MemCopy(BifVector+2*nphase+2,VV,sizeof(FloatHi)*nphase);
      MemCopy(BifVector+3*nphase+2,F1,sizeof(FloatHi)*nphase);
      BifVector[4*nphase+2]=cos(theta);
      stagenData->BifDataOutLen=sizeof(FloatHi)*(4*nphase+3);
      if (kappa <= 0.0) {    
        sprintf(buf,"Neutral saddle + Neimark-Sacker");
      } else {
        sprintf(buf,"Double Neimark-Sacker");
      }
    } /* end of Neimark-Sacker case */
/* provide bifurcation data for NN -> NS starter */
/* Primary branch: */
/*  BifVector[0]:			 _itmap */
/*  BifVector[1] to [nphase]:            1-NS continuation vector */
/*  BifVector[nphase+1] to [2*nphase]:   1-global vector L */
/*  BifVector[2*nphase+1]:               1-kappa  */
    BifVector[0]=_itmap;
    MemCopy(BifVector+1+nphase,L,sizeof(FloatHi)*nphase);
    BifVector[2*nphase+1]=kappa;

polufinal:
    Re2=staFunc->DestroyVector(Re2); 
    Im2=staFunc->DestroyVector(Im2);  
}  
  *msg=buf;
  return 0;
}

Entry(Int2) ns1_TestR2(Vector x, Vector v, FloatHiPtr res) {
Int2 i,j,imax,jmax;
VAR d,a11,a12,a21,a22;
  *res=1;
  if (nphase < 2) goto end;

/* reevaluate Jacobian */
   if (JacRHS (x,C1)) return 1;
   staFunc->CopyMatrixBlock(C1,0,0,A,0,0,nphase,nphase);

/* it is assumed that Q is correct */
/* find a component of Q with max(fabs) */
  imax=1;
  jmax=0;
  d=fabs(Q[0]);
  for (i=0; i<nphase; i++) {
    for (j=0; j<i; j++) {
      if (fabs(Q[i*(i-1)/2+j])>d) {
        imax=i;
	jmax=j;
	d=fabs(Q[i*(i-1)/2+j]);
      }
    }
  }
  d=Q[imax*(imax-1)/2+jmax];

/* compute eigenvectors */
  staFunc->ClearVector(V);
  staFunc->ClearVector(W);
  for (i=0; i<nphase; i++) {
    if (i<imax) V[i]=Q[imax*(imax-1)/2+i]/d;
    if (i>imax) V[i]=-Q[i*(i-1)/2+imax]/d;
    if (i<jmax) W[i]=-Q[jmax*(jmax-1)/2+i]/d;
    if (i>jmax) W[i]=Q[i*(i-1)/2+jmax]/d;
  }

/* compute kappa */
  staFunc->MultiplyMatrixVector(A,V,F1);  /* F1=A*V  */
  staFunc->MultiplyMatrixVector(A,W,F2);  /* F2=A*W  */  

  a11=staFunc->ScalProd(V,V)*staFunc->ScalProd(W,F2);
  a12=staFunc->ScalProd(W,W)*staFunc->ScalProd(V,F1);
  a21=staFunc->ScalProd(V,W)*staFunc->ScalProd(W,F1);
  a22=staFunc->ScalProd(W,V)*staFunc->ScalProd(V,F2);

  d=staFunc->ScalProd(V,V)*staFunc->ScalProd(W,W)
   -staFunc->ScalProd(V,W)*staFunc->ScalProd(V,W);
  
  *res=((a11+a12-a21-a22)/(2.0*d))+1.0;
end:
  test[5]=*res;  
  return 0;
}

Entry(Int2) ns1_ProcR2(Int2  Ind, Vector x, Vector v, CharPtr *msg) {
  stagenData->BifDataOutLen=0;
  if (Ind) 
    sprintf(buf,"Resonance 1:2 (?)");
  else {
Vector v1,W0,W1;
Matrix D;
Int2 i,j,k;
double a20,b20,b11,f0,f1,f2,f3,f4,f5,d;
    v1=staFunc->CreateVector(2*nphase);
    W0=staFunc->CreateVector(nphase);
    W1=staFunc->CreateVector(nphase);
    D=staFunc->CreateMatrix(2*nphase,2*nphase);

/*   compute (generalized)eigenvectors   */
    if (JacRHS (x,C1)) return 1;
    staFunc->CopyMatrixBlock(C1,0,0,A,0,0,nphase,nphase);
    for (i=0; i<nphase; i++) A[i][i]+=1.0;
    
    staFunc->CopyMatrixBlock(A,0,0,D,0,0,nphase,nphase);
    staFunc->CopyMatrixBlock(A,0,0,D,nphase,nphase,nphase,nphase);
    for (i=0; i<nphase; i++) D[i+nphase][i]=-1;
    i=staFunc->SngSolve(D,_eps,v1);
    if (i != 2*nphase-2) {
      sprintf(buf,"1:2 Resonance: eigenvectors (?)");
      goto final;
    }
    staFunc->CopyVectorElements(v1,0,V,0,nphase);       /* eigenvector V0 */
    d=sqrt(staFunc->ScalProd(V,V));
    staFunc->NormalizeVector(V);
    staFunc->CopyVectorElements(v1,nphase,W,0,nphase);  /* generalized eigenvector V1 */
    staFunc->MultiplyVectorScalar(W,1.0/d,W);
    staFunc->AddScaledVector(W,V,-staFunc->ScalProd(V,W),W);

/*  provide bifurcation data for R2->* starters */
/*  BifVector[0]:			 _itmap */
/*  BifVector[1] to [nphase]:            eigenvector V (A*V=-V, <V,V>=1)*/
/*  BifVector[nphase+1]:                 kappa */    
    BifVector[0]=(VAR)_itmap;
    MemCopy(BifVector+1,V,sizeof(FloatHi)*nphase);
    BifVector[nphase+1]=-1.0;
    stagenData->BifDataOutLen=sizeof(FloatHi)*(nphase+2);
    
/* compute the adjoint eigenvectors */
    staFunc->CopyMatrixBlock(C1,0,0,A,0,0,nphase,nphase);
    for (i=0; i<nphase; i++) A[i][i]+=1.0;
    staFunc->TransposeMatrix(A,B);
    staFunc->ClearMatrix(D);
    staFunc->CopyMatrixBlock(B,0,0,D,0,0,nphase,nphase);
    staFunc->CopyMatrixBlock(B,0,0,D,nphase,nphase,nphase,nphase);
    for (i=0; i<nphase; i++) D[i+nphase][i]=-1;
    i=staFunc->SngSolve(D,_eps,v1);
    if (i != 2*nphase-2) {
      sprintf(buf,"1:2 Resonance: adjoint eigenvectors (?)");
      goto final;
    }
    staFunc->CopyVectorElements(v1,0,W1,0,nphase);       /* adjoint eigenvector W1 */
    staFunc->CopyVectorElements(v1,nphase,W0,0,nphase);  /* adjoint generalized eigenvector W0 */
    d=staFunc->ScalProd(V,W0);
    staFunc->MultiplyVectorScalar(W1,1.0/d,W1);
    staFunc->AddScaledVector(W0,W1,-staFunc->ScalProd(W0,W),W0);
    staFunc->MultiplyVectorScalar(W0,1.0/d,W0);

    sprintf(buf,"1:2 Resonance");

final:
    D=staFunc->DestroyMatrix(D);
    v1=staFunc->DestroyVector(v1);
    W0=staFunc->DestroyVector(W0);
    W1=staFunc->DestroyVector(W1);
  }
  *msg=buf;
  return 0;
}

Entry(Int2) ns1_TestR3(Vector x, Vector v, FloatHiPtr res) {
Int2 i,j,imax,jmax;
VAR d,a11,a12,a21,a22;
  *res=1;
  if (nphase < 2) goto end;

/* reevaluate Jacobian */
   if (JacRHS (x,C1)) return 1;
   staFunc->CopyMatrixBlock(C1,0,0,A,0,0,nphase,nphase);

/* it is assumed that Q is correct */
/* find a component of Q with max(fabs) */
  imax=1;
  jmax=0;
  d=fabs(Q[0]);
  for (i=0; i<nphase; i++) {
    for (j=0; j<i; j++) {
      if (fabs(Q[i*(i-1)/2+j])>d) {
        imax=i;
	jmax=j;
	d=fabs(Q[i*(i-1)/2+j]);
      }
    }
  }
  d=Q[imax*(imax-1)/2+jmax];

/* compute eigenvectors */
  staFunc->ClearVector(V);
  staFunc->ClearVector(W);
  for (i=0; i<nphase; i++) {
    if (i<imax) V[i]=Q[imax*(imax-1)/2+i]/d;
    if (i>imax) V[i]=-Q[i*(i-1)/2+imax]/d;
    if (i<jmax) W[i]=-Q[jmax*(jmax-1)/2+i]/d;
    if (i>jmax) W[i]=Q[i*(i-1)/2+jmax]/d;
  }

/* compute kappa */
  staFunc->MultiplyMatrixVector(A,V,F1);  /* F1=A*V  */
  staFunc->MultiplyMatrixVector(A,W,F2);  /* F2=A*W  */  

  a11=staFunc->ScalProd(V,V)*staFunc->ScalProd(W,F2);
  a12=staFunc->ScalProd(W,W)*staFunc->ScalProd(V,F1);
  a21=staFunc->ScalProd(V,W)*staFunc->ScalProd(W,F1);
  a22=staFunc->ScalProd(W,V)*staFunc->ScalProd(V,F2);

  d=staFunc->ScalProd(V,V)*staFunc->ScalProd(W,W)
   -staFunc->ScalProd(V,W)*staFunc->ScalProd(V,W);
  
  *res=((a11+a12-a21-a22)/(2.0*d))+0.5;
end:
  test[6]=*res;  
  return 0;
}



Entry(Int2) ns1_ProcR3(Int2  Ind, Vector x, Vector v, CharPtr *msg) {
Int2 i,j,imax,jmax,m;
VAR max1,max2,a11,a12,a21,a22,d;
  stagenData->BifDataOutLen=0;
  if (Ind) {
    stagenData->BifDataOutLen=0;
    sprintf(buf,"Resonance 1:3 (?)"); 
  } else {
    m=nphase*(nphase-1)/2;
/* reevaluate Jacobian */
    JacRHS (x,C1);
    staFunc->CopyMatrixBlock(C1,0,0,A,0,0,nphase,nphase);
/* find pair in (g11,g12), (g21,g22) with max absolute value */
    max1=(fabs(VB1[m])>fabs(VB2[m]))?VB1[m]:VB2[m];
    max2=(fabs(VB1[m+1])>fabs(VB2[m+1]))?VB1[m+1]:VB2[m+1];
    if (fabs(max1)>fabs(max2))
     imax=m;
    else imax=m+1;    
    for(i=0;i<m;i++) {
      Q[i]=VB2[imax]*VB1[i]-VB1[imax]*VB2[i];
    }
/* find a component of Q with max(fabs) */
    imax=1;
    jmax=0;
    d=fabs(Q[0]);
    for (i=0; i<nphase; i++) {
      for (j=0; j<i; j++) {
        if (fabs(Q[i*(i-1)/2+j])>d) {
          imax=i;
	  jmax=j;
	  d=fabs(Q[i*(i-1)/2+j]);
        }
      }
    }
    d=Q[imax*(imax-1)/2+jmax];

/* compute eigenvectors */
    staFunc->ClearVector(V);
    staFunc->ClearVector(W);
    for (i=0; i<nphase; i++) {
      if (i<imax) V[i]=Q[imax*(imax-1)/2+i]/d;
      if (i>imax) V[i]=-Q[i*(i-1)/2+imax]/d;
      if (i<jmax) W[i]=-Q[jmax*(jmax-1)/2+i]/d;
      if (i>jmax) W[i]=Q[i*(i-1)/2+jmax]/d;
    }

/* compute kappa */
    staFunc->MultiplyMatrixVector(A,V,F1);  /* F1=A*V  */
    staFunc->MultiplyMatrixVector(A,W,F2);  /* F2=A*W  */

    a11=staFunc->ScalProd(V,V)*staFunc->ScalProd(W,F2);
    a12=staFunc->ScalProd(W,W)*staFunc->ScalProd(V,F1);
    a21=staFunc->ScalProd(V,W)*staFunc->ScalProd(W,F1);
    a22=staFunc->ScalProd(W,V)*staFunc->ScalProd(V,F2);

    d=staFunc->ScalProd(V,V)*staFunc->ScalProd(W,W)
     -staFunc->ScalProd(V,W)*staFunc->ScalProd(V,W);

/* orthogonalize and normalize eigenvectors */
    staFunc->NormalizeVector(V);
    staFunc->AddScaledVector(W,V,-staFunc->ScalProd(V,W),L);
    staFunc->NormalizeVector(L);
  
    stagenData->BifDataOutLen=sizeof(FloatHi)*(2*nphase+2);
    BifVector[0]=(VAR)_itmap;
    MemCopy(BifVector+1,V,sizeof(FloatHi)*nphase);
    MemCopy(BifVector+nphase+1,L,sizeof(FloatHi)*nphase);
    BifVector[2*nphase+1]=(a11+a12-a21-a22)/(2.0*d);      
    sprintf(buf,"Resonance 1:3");     
  }
  *msg=buf;
  return 0;
}

Entry(Int2) ns1_TestR4(Vector x, Vector v, FloatHiPtr res) {
Int2 i,j,imax,jmax;
VAR d,a11,a12,a21,a22;
  *res=1;
  if (nphase < 2) goto end;

/* reevaluate Jacobian */
   if (JacRHS (x,C1)) return 1;
   staFunc->CopyMatrixBlock(C1,0,0,A,0,0,nphase,nphase);

/* it is assumed that Q is correct */
/* find a component of Q with max(fabs) */
  imax=1;
  jmax=0;
  d=fabs(Q[0]);
  for (i=0; i<nphase; i++) {
    for (j=0; j<i; j++) {
      if (fabs(Q[i*(i-1)/2+j])>d) {
        imax=i;
	jmax=j;
	d=fabs(Q[i*(i-1)/2+j]);
      }
    }
  }
  d=Q[imax*(imax-1)/2+jmax];

/* compute eigenvectors */
  staFunc->ClearVector(V);
  staFunc->ClearVector(W);
  for (i=0; i<nphase; i++) {
    if (i<imax) V[i]=Q[imax*(imax-1)/2+i]/d;
    if (i>imax) V[i]=-Q[i*(i-1)/2+imax]/d;
    if (i<jmax) W[i]=-Q[jmax*(jmax-1)/2+i]/d;
    if (i>jmax) W[i]=Q[i*(i-1)/2+jmax]/d;
  }

/* compute kappa */
  staFunc->MultiplyMatrixVector(A,V,F1);  /* F1=A*V  */
  staFunc->MultiplyMatrixVector(A,W,F2);  /* F2=A*W  */  

  a11=staFunc->ScalProd(V,V)*staFunc->ScalProd(W,F2);
  a12=staFunc->ScalProd(W,W)*staFunc->ScalProd(V,F1);
  a21=staFunc->ScalProd(V,W)*staFunc->ScalProd(W,F1);
  a22=staFunc->ScalProd(W,V)*staFunc->ScalProd(V,F2);

  d=staFunc->ScalProd(V,V)*staFunc->ScalProd(W,W)
   -staFunc->ScalProd(V,W)*staFunc->ScalProd(V,W);
  
  *res=(a11+a12-a21-a22)/(2.0*d);
end:
  test[7]=*res;
  return 0;
}

Entry(Int2) ns1_ProcR4(Int2  Ind, Vector x, Vector v, CharPtr *msg) {
Int2 i,j,imax,jmax,m;
VAR max1,max2,a11,a12,a21,a22,d;
  stagenData->BifDataOutLen=0;
  if (Ind) {
    stagenData->BifDataOutLen=0;
    sprintf(buf,"Resonance 1:4 (?)"); 
  } else {
    m=nphase*(nphase-1)/2;
/* reevaluate Jacobian */
    JacRHS (x,C1);
    staFunc->CopyMatrixBlock(C1,0,0,A,0,0,nphase,nphase);
/* find pair in (g11,g12), (g21,g22) with max absolute value */
    max1=(fabs(VB1[m])>fabs(VB2[m]))?VB1[m]:VB2[m];
    max2=(fabs(VB1[m+1])>fabs(VB2[m+1]))?VB1[m+1]:VB2[m+1];
    if (fabs(max1)>fabs(max2))
     imax=m;
    else imax=m+1;    
    for(i=0;i<m;i++) {
      Q[i]=VB2[imax]*VB1[i]-VB1[imax]*VB2[i];
    }
/* find a component of Q with max(fabs) */
    imax=1;
    jmax=0;
    d=fabs(Q[0]);
    for (i=0; i<nphase; i++) {
      for (j=0; j<i; j++) {
        if (fabs(Q[i*(i-1)/2+j])>d) {
          imax=i;
	  jmax=j;
	  d=fabs(Q[i*(i-1)/2+j]);
        }
      }
    }
    d=Q[imax*(imax-1)/2+jmax];

/* compute eigenvectors */
    staFunc->ClearVector(V);
    staFunc->ClearVector(W);
    for (i=0; i<nphase; i++) {
      if (i<imax) V[i]=Q[imax*(imax-1)/2+i]/d;
      if (i>imax) V[i]=-Q[i*(i-1)/2+imax]/d;
      if (i<jmax) W[i]=-Q[jmax*(jmax-1)/2+i]/d;
      if (i>jmax) W[i]=Q[i*(i-1)/2+jmax]/d;
    }

/* compute kappa */
    staFunc->MultiplyMatrixVector(A,V,F1);  /* F1=A*V  */
    staFunc->MultiplyMatrixVector(A,W,F2);  /* F2=A*W  */

    a11=staFunc->ScalProd(V,V)*staFunc->ScalProd(W,F2);
    a12=staFunc->ScalProd(W,W)*staFunc->ScalProd(V,F1);
    a21=staFunc->ScalProd(V,W)*staFunc->ScalProd(W,F1);
    a22=staFunc->ScalProd(W,V)*staFunc->ScalProd(V,F2);

    d=staFunc->ScalProd(V,V)*staFunc->ScalProd(W,W)
     -staFunc->ScalProd(V,W)*staFunc->ScalProd(V,W);

/* orthogonalize and normalize eigenvectors */
    staFunc->NormalizeVector(V);
    staFunc->AddScaledVector(W,V,-staFunc->ScalProd(V,W),L);
    staFunc->NormalizeVector(L);
  
    stagenData->BifDataOutLen=sizeof(FloatHi)*(2*nphase+2);
    BifVector[0]=(VAR)_itmap;
    MemCopy(BifVector+1,V,sizeof(FloatHi)*nphase);
    MemCopy(BifVector+nphase+1,L,sizeof(FloatHi)*nphase);
    BifVector[2*nphase+1]=(a11+a12-a21-a22)/(2.0*d);      
    sprintf(buf,"Resonance 1:4");     
  }
  *msg=buf;
  return 0;
}

/**************************************************/
/*  Adaptation of the bordering vectors V1,V2,W1,W2 */
Entry(Int2) ns1_Adapter(Vector x, Vector v) {
int i,m,Ret;
VAR alpha,beta,max1,max2;
  m=nphase*(nphase-1)/2;
/* adapt only in case m >= 2 ; for m = 1 no adaptation is needed */
  if (m >= 2) {
/* solve first system to obtain new values for V1 */
    staFunc->ClearVector(K1);
    staFunc->ClearVector(K2);
    staFunc->CopyMatrix(BP,BT);  
    K1[m]=1.0;
    K2[m+1]=1.0;
    staFunc->Decomp(BT);
    staFunc->Solve(BT,K1);
    staFunc->Solve(BT,K2);
    staFunc->CopyVector(K1,OK1);
    staFunc->CopyVector(K1,K3);
    staFunc->CopyVector(K2,OK2);
    max1=(fabs(OK1[m])>fabs(OK2[m]))?fabs(OK1[m]):fabs(OK2[m]);
    max2=(fabs(OK1[m+1])>fabs(OK2[m+1]))?fabs(OK1[m+1]):fabs(OK2[m+1]);  
    if (max1 > max2) {
     alpha = OK1[m];
     beta = OK2[m];
    }
    else {
     alpha = OK1[m+1];
     beta = OK2[m+1];
    }
    staFunc->MultiplyVectorScalar(OK1,-beta,OK1);
    staFunc->AddScaledVector(OK1,OK2,alpha,K1);
    staFunc->NormalizeVector(K1);
    alpha=staFunc->ScalProd(V1,K1);
    if (alpha < 0) {
      for (i=0;i<m+2;i++)
        K1[i]=-K1[i];
    }
    staFunc->CopyVector(K1,V1);

/* solve second system to obtain new values for W1 */
    staFunc->TransposeMatrix(BP,BT);
    staFunc->ClearVector(K1);
    staFunc->ClearVector(K2);
    K1[m]=1.0;
    K2[m+1]=1.0;
    staFunc->Decomp(BT);
    staFunc->Solve(BT,K1);
    staFunc->Solve(BT,K2);
    staFunc->CopyVector(K1,OK1);
    staFunc->CopyVector(K1,K3);
    staFunc->CopyVector(K2,OK2); 
    max1=(fabs(OK1[m])>fabs(OK2[m]))?fabs(OK1[m]):fabs(OK2[m]);
    max2=(fabs(OK1[m+1])>fabs(OK2[m+1]))?fabs(OK1[m+1]):fabs(OK2[m+1]);  
    if (max1 > max2) {
     alpha = OK1[m];
     beta = OK2[m];
    }
    else {
     alpha = OK1[m+1];
     beta = OK2[m+1];
    }
    staFunc->MultiplyVectorScalar(OK1,-beta,OK1);
    staFunc->AddScaledVector(OK1,OK2,alpha,K1);
    staFunc->NormalizeVector(K1);
    alpha=staFunc->ScalProd(W1,K1);
    if (alpha < 0) {
     for (i=0;i<m+2;i++)
       K1[i]=-K1[i];
    }
    staFunc->CopyVector(K1,W1);

/* adapt V2 */
    staFunc->ClearMatrix(BP1);
    staFunc->CopyMatrixBlock(BP,0,0,BP1,0,0,m,m);
    for (i=0;i<m;i++) {
     BP1[i][m] = W1[i];
     BP1[m][i] = V1[i];
    }    
    staFunc->CopyVectorElements(W2,0,RL,0,m+1);
    staFunc->Decomp(BP1);
    staFunc->Solve(BP1,RL);
    staFunc->NormalizeVector(RL);
    staFunc->CopyVectorElements(V2,0,OLD,0,m+1);
    alpha=staFunc->ScalProd(OLD,RL);
    if (alpha < 0) {
     for (i=0;i<m+1;i++)
       RL[i]=-RL[i];
    }
    staFunc->ClearVector(V2);
    staFunc->CopyVectorElements(RL,0,V2,0,m+1);
/* adapt W2 */
    staFunc->ClearMatrix(BP1);
    staFunc->CopyMatrixBlock(BP,0,0,BP1,0,0,m,m);
    for (i=0;i<m;i++) {
     BP1[i][m] = W1[i];
     BP1[m][i] = V1[i];
    }
    staFunc->TransposeMatrix(BP1,BT1);
    staFunc->Decomp(BT1);
    staFunc->ClearVector(RL);
    staFunc->CopyVectorElements(V2,0,RL,0,m+1);
    staFunc->Solve(BT1,RL);
    staFunc->NormalizeVector(RL);
    staFunc->CopyVectorElements(W2,0,OLD,0,m+1);
    alpha=staFunc->ScalProd(OLD,RL);
    if (alpha < 0) {
     for (i=0;i<m+1;i++)
       RL[i]=-RL[i];
    }  
    staFunc->ClearVector(W2);
    staFunc->CopyVectorElements(RL,0,W2,0,m+1);
  } /* if m >=2 */

/* adapt test function for Double Neimark-Sacker if set */
  if (staData->ActiveSing[4]) { 
   for (i=0;i<m;i++) {
      BP[m][i]=V1[i];
      BP[i][m]=W1[i];
    }
    for (i=0;i<m+2;i++) {
      BP[m+1][i]=V2[i];
      BP[i][m+1]=W2[i];
    }
    staFunc->CopyMatrix(BP,BT);
    Ret=staFunc->Decomp(BT);
    if (Ret) return 1;
    staFunc->ClearVector(VB2);
    VB2[m+1]=1.0;
    staFunc->Solve(BT,VB2);   
    test[4]=VB2[m+1];
  }   
  return 0;
}










